/* FILE: demag_threaded.h            -*-Mode: c++-*-
 *
 * Average H demag field across rectangular cells.  This is a modified
 * version of the simpledemag class, which uses symmetries in the
 * interaction coefficients to reduce memory usage.
 *
 * This is a threaded version of Oxs_Demag
 *
 */

#if OOMMF_THREADS

#ifndef _OXS_DEMAG_THREADED
#define _OXS_DEMAG_THREADED

#include "energy.h"
#include "fft.h"
#include "key.h"
#include "mesh.h"
#include "meshvalue.h"
#include "simstate.h"
#include "threevector.h"
#include "rectangularmesh.h"

#include "oxsthread.h"

/* End includes */

// Support classes
class _Oxs_Demag_Threaded_FFT
  : public Oxs_ThreadRunObj {
public:
  void Set(UINT4m rsize1_x,UINT4m rsize2_x,UINT4m rsize3_x,
           UINT4m csize1_x,UINT4m csize2_x,UINT4m csize3_x,
           UINT4m cstride2_x,UINT4m cstride3_x,
           Oxs_Complex* arr_x,Oxs_FFT3D* fft_x)
  {
    rsize1=rsize1_x;
    rsize2=rsize2_x;
    rsize3=rsize3_x;
    csize1=csize1_x;
    csize2=csize2_x;
    csize3=csize3_x;
    cstride2=cstride2_x;
    cstride3=cstride3_x;
    arr=arr_x;
    fft=fft_x;
  }

  void Cmd(int /* id */,void* data) {
    int direction = *(static_cast<int *>(data));
    if(0==direction) {
      fft->ForwardRealDataNoScale(arr,rsize1,rsize2,rsize3,
                                  csize1,csize2,csize3,
                                  cstride2,cstride3);
    } else {
      fft->InverseRealDataNoScale(arr,rsize1,rsize2,rsize3,
                                  csize1,csize2,csize3,
                                  cstride2,cstride3);
    }
  }

private:
  UINT4m rsize1,rsize2,rsize3;
  UINT4m csize1,csize2,csize3;
  UINT4m cstride2,cstride3;
  Oxs_Complex* arr;
  Oxs_FFT3D* fft;
};

class Oxs_Demag_Threaded:public Oxs_Energy {
private:
#if REPORT_TIME
  mutable Nb_StopWatch ffttime;
  mutable Nb_StopWatch convtime;
  mutable Nb_StopWatch dottime;
#endif // REPORT_TIME

  mutable UINT4m rdimx; // Natural size of real data
  mutable UINT4m rdimy; // Digital Mars compiler wants these as separate
  mutable UINT4m rdimz; //    statements, because of "mutable" keyword.
  mutable UINT4m cdimx; // Full size of complex data
  mutable UINT4m cdimy;
  mutable UINT4m cdimz;
  // 2*cdimx>=rdimx, cdimy>=rdimy, cdimz>=rdimz
  // cdim[xyz] should be powers of 2.
  mutable UINT4m cstridey; // Strides across complex data
  mutable UINT4m cstridez;
  // cstridey>=cdimx, cstridez>=cdimy*cstridey
  // The stride sizes for the real arrays are just double the
  // complex strides, except cstride1 and rstride1 are assumed
  // to be 1.  Total matrix size is effectively cdimz*cstridez
  // Oxs_Complex elements, or twice that many "double" elements.

  mutable UINT4m mesh_id;

  // The A## arrays hold demag coefficients, transformed into
  // frequency domain.  These are held long term.  xcomp,
  // ycomp, and zcomp are used as temporary space, first to hold
  // the transforms of Mx, My, and Mz, then to store Hx, Hy, and
  // Hz.
  mutable UINT4m adimx;
  mutable UINT4m adimy;
  mutable UINT4m adimz;
  mutable UINT4m astridey;
  mutable UINT4m astridez;
  mutable OXS_COMPLEX_REAL_TYPE *A00;
  mutable OXS_COMPLEX_REAL_TYPE *A01;
  mutable OXS_COMPLEX_REAL_TYPE *A02;
  mutable OXS_COMPLEX_REAL_TYPE *A11;
  mutable OXS_COMPLEX_REAL_TYPE *A12;
  mutable OXS_COMPLEX_REAL_TYPE *A22;
  mutable Oxs_Complex *xcomp;
  mutable Oxs_Complex *ycomp;
  mutable Oxs_Complex *zcomp;
#if 0
  mutable Oxs_FFT3D fft; // All transforms are same size, so we need
  /// only one Oxs_FFT3D object.
#else
  mutable Oxs_FFT3D fftx; // Forward/inverse routines aren't reentrant,
  mutable Oxs_FFT3D ffty; // so we need one for thread.
  mutable Oxs_FFT3D fftz;
  mutable _Oxs_Demag_Threaded_FFT fft_thread_x;
  mutable _Oxs_Demag_Threaded_FFT fft_thread_y;
  mutable _Oxs_Demag_Threaded_FFT fft_thread_z;
  mutable Oxs_ThreadTree fft_tree;
#endif

  void (Oxs_Demag_Threaded::*fillcoefs)(const Oxs_Mesh*) const;
  void FillCoefficientArraysFast(const Oxs_Mesh* mesh) const;
  void FillCoefficientArraysStandard(const Oxs_Mesh* mesh) const;
  /// The "standard" variant is simpler but slower, and is retained
  /// mainly for testing and development purposes.

  UINT4m NextPowerOfTwo(UINT4m n) const;  // Helper function
  void ReleaseMemory() const;

protected:
  virtual void GetEnergy(const Oxs_SimState& state,
			 Oxs_EnergyData& oed) const;

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  Oxs_Demag_Threaded(const char* name,     // Child instance id
                       Oxs_Director* newdtr, // App director
                       const char* argstr);  // MIF input block parameters
  virtual ~Oxs_Demag_Threaded();
  virtual BOOL Init();
};


#endif // _OXS_DEMAG_THREADED
#endif // OOMMF_THREADS
