/* FILE: lagrangeevolve.h                 -*-Mode: c++-*-
 *
 * Concrete minimization evolver class, using conjugate gradients
 * applied to augmented Lagrangian, as suggested by Tony Kearsley,
 * 25-Oct-2001.
 */

#ifndef _OXS_LAGRANGEEVOLVE
#define _OXS_LAGRANGEEVOLVE

#include "minevolver.h"
#include "key.h"
#include "output.h"

/* End includes */

class Oxs_LagrangeEvolve:public Oxs_MinEvolver {
private:

  // Scratch space.  This is needed by GetEnergyAndmxHxm as
  // holding space for GetEnergyData calls.  This space
  // is also used for temporary holding by Step.
  Oxs_MeshValue<REAL8m> scratch_energy;
  Oxs_MeshValue<ThreeVector> scratch_field;

  // Cache space.  This should probably be wrapped up in a class.
  // Might also want to add a key with a read lock onto the state.
  UINT4m cache_id;
  REAL8m cache_L;
  Oxs_MeshValue<ThreeVector> cache_x;
  Oxs_MeshValue<REAL8m> cache_energy;
  Oxs_MeshValue<ThreeVector> cache_mxHxm;

  // Extra data for "current" state, i.e., the state
  // with lowest L found so far for current lambda.
  UINT4m current_id;
  REAL8m current_L;
  Oxs_MeshValue<ThreeVector> current_x;
  Oxs_MeshValue<REAL8m> current_energy;
  Oxs_MeshValue<ThreeVector> current_mxHxm;


  // Context initialization on new stage detection
  UINT4m working_stage_number;
  BOOL StageInit();

  // Energy and Lagrangian calculation routines
  REAL8m GetEnergyAndmxHxm(const Oxs_SimState& state,          // Import
			   Oxs_MeshValue<REAL8m> &energy,      // Export
			   Oxs_MeshValue<ThreeVector> &mxHxm); // Export

  void EvaluateLagrangian(const Oxs_SimState&  state,      // Imports...
                          const Oxs_MeshValue<ThreeVector>& x,
                          Oxs_MeshValue<REAL8m>& energy,   // Exports...
                          Oxs_MeshValue<ThreeVector>& mxHxm,
                          REAL8m& rel_L,
                          REAL8m& Lp);

  // Top level, augmented Lagrangian support variables
  REAL8m rho; // Fixed augmentation term, set during initialization
  REAL8m weighted_rho; // rho weighted by |H|
  REAL8m normsq_guard; // Check used when resetting lambda: any
  /// x[i] that is outside 1-norm_guard < |x[i]|^2 < 1+norm_guard
  /// is renormalized to 1.0.  Hopefully the rho term will prevent
  /// this from being invoked very often.
  BOOL lambda_valid;
  Oxs_MeshValue<REAL8m> lambda;
  UINT4m lambda_min_CG_cycle_count;
  REAL8m lambda_reset_L_drop_ratio;
  UINT4m lambda_CG_cycle_count;
  REAL8m lambda_CG_cycle_max_L_drop;
  // For each lambda setting, do at least lambda_min_cg_cycle_count
  // conjugate gradient cycles, and track the maximum drop in L
  // between successive CG base points.  Update lambda the first
  // cycle after lambda_min_cg_cycle_count that fails to drop
  // at least lambda_reset_L_drop_ratio from the max drop.

  // Base point variables
  BOOL base_valid;
  BOOL base_use_conjugate_direction;
  UINT4m base_id;
  REAL8m base_L;
  Oxs_MeshValue<ThreeVector> base_x;
  Oxs_MeshValue<REAL8m> base_energy;
  Oxs_MeshValue<ThreeVector> base_direction;
  REAL8m base_direction_max_mag;
  REAL8m base_dL_normsq;
  REAL8m CG_collapse_guard; // Check against collapsed
  /// conjugate gradient subspaces.  Triggered if
  //  dL*base_direction >= -1 * CG_collapse_guard * |dL|^2.
  // Typically CG_collapse_guard should be on the order of 0.02.

  // Line minimization support
  BOOL bracket_valid;
  REAL8m last_minimum_reduction_ratio;
  REAL8m next_to_last_minimum_reduction_ratio;
  /// The 2 preceding variables, *last_minimum_reduction_ratio, are
  /// cross-call state variables used in TakeLineMinimumStep().
  /// Sense: "Big" (i.e., close to 1) reduction is good.
  REAL8m line_minimum_relwidth; // Line minimum considered
  /// found when relative size of bracketed region is smaller
  /// than this.
  REAL8m minimum_bracket_step;
  REAL8m maximum_bracket_step;
  REAL8m bracket_start_step;
  struct Bracket {
  public:
    REAL8m offset;  // Location is basept + offset*basedir
    REAL8m rel_L;   // Augmented Lagrangian evaluated relative
                   /// to base_state
    REAL8m Lp;      // Derivative of L in direction basedir.
    Bracket() : offset(0.),rel_L(0.),Lp(0.) {}
  };
  Bracket leftpt,midpt,rightpt;

  void FillBracket(REAL8m offset,                          // Imports...
                   const Oxs_SimState* oldstate,
                   Oxs_Key<Oxs_SimState>& new_state_key,   // Exports...
                   Bracket& endpt);

  BOOL FindBracketStep(const Oxs_SimState* oldstate,
		       Oxs_Key<Oxs_SimState>& statekey,
		       BOOL& bracket_error);
  // Takes one step; Returns 1 if minimum is successfully bracketed,
  // otherwise 0.  Also sets bracket_error to 1 on failure to bracket
  // minima inside allowed bracket range.

  BOOL TakeLineMinimumStep(const Oxs_SimState* old_state,     // Import
                           Oxs_Key<Oxs_SimState>& state_key); // Export
  /// Reduces bracket span, and fills cache & state_key with new state.
  /// Returns true if new state has lower L than both previous bracket
  /// points, i.e., if new state should be accepted as a good step.

  void UpdateCurrentData(const Oxs_SimState& state);
  /// Impresses state data into current data block.

  void FillLambdaFromCurrentData(const Oxs_SimState& state);
  /// Sets up lambda from data in current data block.  Also
  /// adjusts current_L to reflect the change in lambda, and
  /// marks cache and base data invalid, since cache_L and
  /// base_L would need to be updated, but aren't here.

  REAL8m Calculate_dL
  (const Oxs_Mesh* mesh,                     // Imports...
   const Oxs_MeshValue<REAL8m>& Ms,
   const REAL8m work_rho,
   const Oxs_MeshValue<ThreeVector>& mxHxm,
   const Oxs_MeshValue<REAL8m> &work_lambda,
   const Oxs_MeshValue<ThreeVector>& x,
   Oxs_MeshValue<ThreeVector>& dL);           // Export
  /// On return, dL is filled, and the return value is
  /// \sum_i |dL[i]|^2.  This routine does not use any
  /// Oxs_LagrangeEvolve member variables.
  /// Note: See also Oxs_LagrangeEvolve::EvaluateLagrangian()

  REAL8m FindMaxMag(const Oxs_MeshValue<ThreeVector>& vec);
  /// Support function; returns max_i |vec[i]|.

  REAL8m VecDot(const Oxs_MeshValue<ThreeVector>& v,
				    const Oxs_MeshValue<ThreeVector>& w);
  /// Support function; returns \sum_i v_i*w_i

  void InterchangeCacheAndCurrentData();
  /// Swaps cache_* and current_* data blocks.

  // Outputs
  UINT4m energy_calc_count;
  UINT4m lambda_cycle_count;
  UINT4m direction_cycle_count;
  UINT4m bracket_cycle_count;
  UINT4m linemin_cycle_count;
  void UpdateDerivedOutputs(const Oxs_SimState& state);
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> max_mxHxm_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> delta_E_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> total_energy_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> energy_calc_count_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> L_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> lambda_cycle_count_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> direction_cycle_count_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> bracket_cycle_count_output;
  Oxs_ScalarOutput<Oxs_LagrangeEvolve> linemin_cycle_count_output;

public:
  virtual const char* ClassName() const; // ClassName() is
  /// automatically generated by the OXS_EXT_REGISTER macro.
  virtual BOOL Init();
  Oxs_LagrangeEvolve(const char* name,     // Child instance id
                     Oxs_Director* newdtr, // App director
                     const char* argstr);  // MIF input block parameters
  virtual ~Oxs_LagrangeEvolve();

  virtual  BOOL
  Step(const Oxs_MinDriver* driver,
       Oxs_ConstKey<Oxs_SimState> current_state,
       const Oxs_DriverStepInfo& step_info,
       Oxs_Key<Oxs_SimState>& next_state);
  // Returns true if step was successful, false if
  // unable to step as requested.
};

#endif // _OXS_LAGRANGEEVOLVE
