/* FILE: modllg_rungekuttaevolve.cc              -*-Mode: c++-*-
 *
 * Concrete evolver class, using Runge-Kutta steps.
 * This version implements a modified LLG equation with |m|
 * control.
 *
 */

#include <float.h>
#include <string>

#include "nb.h"
#include "director.h"
#include "timedriver.h"
#include "simstate.h"
#include "modllg_rungekuttaevolve.h"
#include "key.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy

OC_USE_STRING;

OC_USE_STD_NAMESPACE;  // Specify std namespace, if supported.
/// For some compilers this is needed to get "long double"
/// versions of the basic math library functions, e.g.,
/// long double atan(long double);

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_ModLLG_RungeKuttaEvolve);

/* End includes */

// Constructor
Oxs_ModLLG_RungeKuttaEvolve::Oxs_ModLLG_RungeKuttaEvolve(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_TimeEvolver(name,newdtr,argstr),
    max_step_decrease(0.03125), max_step_increase_limit(4.0),
    max_step_increase_adj_ratio(1.9),
    reject_goal(0.05), reject_ratio(0.05),
    energy_state_id(0),next_timestep(0.),
    rkstep_ptr(NULL)
{
  // Process arguments
  min_timestep=GetRealInitValue("min_timestep",0.);
  max_timestep=GetRealInitValue("max_timestep",1e-10);
  if(max_timestep<=0.0) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
                "Invalid parameter value:"
                " Specified max time step is %g (should be >0.)",
                static_cast<double>(max_timestep));
    throw Oxs_Ext::Error(this,buf);
  }

  allowed_error_rate = GetRealInitValue("error_rate",1.0);
  if(allowed_error_rate>0.0) {
    allowed_error_rate *= PI*1e9/180.; // Convert from deg/ns to rad/s
  }
  allowed_absolute_step_error
    = GetRealInitValue("absolute_step_error",0.2);
  if(allowed_absolute_step_error>0.0) {
    allowed_absolute_step_error *= PI/180.; // Convert from deg to rad
  }
  allowed_relative_step_error
    = GetRealInitValue("relative_step_error",0.01);

  expected_energy_precision =
    GetRealInitValue("energy_precision",1e-10);

  reject_goal = GetRealInitValue("reject_goal",0.05);
  if(reject_goal<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"reject_goal\" value must be non-negative.");
  }

  min_step_headroom = GetRealInitValue("min_step_headroom",0.33);
  if(min_step_headroom<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"min_step_headroom\" value must be bigger than 0.");
  }

  max_step_headroom = GetRealInitValue("max_step_headroom",0.95);
  if(max_step_headroom<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"max_step_headroom\" value must be bigger than 0.");
  }

  if(min_step_headroom>max_step_headroom) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"min_step_headroom\" value must not be larger than"
         " \"max_step_headroom\".");
  }


  alpha = GetRealInitValue("alpha",0.5);

  // User may specify either gamma_G (Gilbert) or
  // gamma_LL (Landau-Lifshitz).  Code uses "gamma"
  // which is LL form.
  if(HasInitValue("gamma_G") && HasInitValue("gamma_LL")) {
    throw Oxs_Ext::Error(this,"Invalid Specify block; "
                         "both gamma_G and gamma_LL specified.");
  } else if(HasInitValue("gamma_G")) {
    gamma = GetRealInitValue("gamma_G")/(1+alpha*alpha);
  } else if(HasInitValue("gamma_LL")) {
    gamma = GetRealInitValue("gamma_LL");
  } else {
    gamma = 2.211e5/(1+alpha*alpha);
  }
  gamma = fabs(gamma); // Force positive

  do_precess = GetIntInitValue("do_precess",1);

  start_dm = GetRealInitValue("start_dm",0.01);
  if(start_dm<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"start_dm\" value must be nonnegative.");
  }
  start_dm *= PI/180.; // Convert from deg to rad

  // |m| adjustment parameters
  resize_norm = RT_INVALID;
  String resize_norm_string = GetStringInitValue("resize_norm","dmdt_ave");
  Oxs_ToLower(resize_norm_string);
  if(resize_norm_string.compare("dmdt_ave")==0) {
    resize_norm = RT_DMDT_AVE;
  } else if(resize_norm_string.compare("dmdt_h")==0) {
    resize_norm = RT_DMDT_H;
  } else if(resize_norm_string.compare("h")==0) {
    resize_norm = RT_H;
  } else if(resize_norm_string.compare("h_ave")==0) {
    resize_norm = RT_H_AVE;
  } else if(resize_norm_string.compare("h_mod")==0) {
    resize_norm = RT_H_MOD;
  } else if(resize_norm_string.compare("norm")==0) {
    resize_norm = RT_NORM;
  } else if(resize_norm_string.compare("none")==0) {
    resize_norm = RT_NONE;
  } else {
    resize_norm = RT_INVALID; // Safety
  }
  if(RT_INVALID == resize_norm) {
    throw Oxs_Ext::Error(this,"Invalid resize_norm setting detected."
			 " Should be one of \"dmdt_ave\" or \"norm\".");
  }
  allowed_mnorm_deviation = GetRealInitValue("norm_error",0.0);

  resize_differential = RT_INVALID;
  String resize_differential_string
    = GetStringInitValue("resize_differential","norm");
  Oxs_ToLower(resize_differential_string);
  if(resize_differential_string.compare("dmdt_ave")==0) {
    resize_differential = RT_DMDT_AVE;
  } else if(resize_differential_string.compare("norm")==0) {
    resize_differential = RT_NORM;
  } else if(resize_differential_string.compare("none")==0) {
    resize_differential = RT_NONE;
  } else {
    resize_differential = RT_INVALID; // Safety
  }
  if(RT_INVALID == resize_differential) {
    throw Oxs_Ext::Error(this,"Invalid resize_differential setting detected."
			 " Should be one of \"dmdt_ave\" or \"norm\".");
  }
  m_mag_adj = GetRealInitValue("mag_adj",1.0);
  m_mag_ie_adj = GetRealInitValue("mag_ie_adj",1.0);
  REAL8m m_mag_ie_period = GetRealInitValue("mag_ie_period",1e-9);

  if(m_mag_ie_period<=0.0) {
    throw Oxs_Ext::Error(this,"Invalid mag_ie_period specified."
			 " Should be integration period, in ns; must be >0.");
  }
  m_mag_ie_freq = 1.0/m_mag_ie_period;

  m_mag_adj *= 1.;               // Default scaling
  m_mag_ie_adj *= m_mag_ie_freq; // Default scaling

  m_mag_width = GetRealInitValue("mag_width",1.0);
  m_mag_width = log(2.0)/(m_mag_width*m_mag_width);

  m_mag_de_adj = GetRealInitValue("mag_de_adj",0.0);


  stage_init_step_control = SISC_AUTO;  // Safety
  String stage_start = GetStringInitValue("stage_start","auto");
  Oxs_ToLower(stage_start);
  if(stage_start.compare("start_dm")==0) {
    stage_init_step_control = SISC_START_DM;
  } else if(stage_start.compare("continuous")==0) {
    stage_init_step_control = SISC_CONTINUOUS;
  } else if(stage_start.compare("auto")==0
            || stage_start.compare("automatic")==0) {
    stage_init_step_control = SISC_AUTO;
  } else {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
                         " \"stage_start\" value must be one of"
                         " start_dm, continuous, or auto.");
  }

  String method = GetStringInitValue("method","rkf54");
  Oxs_ToLower(method); // Do case insensitive match
  if(method.compare("rk2")==0) {
    rkstep_ptr = &Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaStep2;
  } else if(method.compare("rk4")==0) {
    rkstep_ptr = &Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaStep4;
  } else if(method.compare("rkf54m")==0) {
    rkstep_ptr = &Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaFehlbergStep54M;
  } else if(method.compare("rkf54s")==0) {
    rkstep_ptr = &Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaFehlbergStep54S;
  } else if(method.compare("rkf54")==0) {
    rkstep_ptr = &Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaFehlbergStep54;
  } else {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
                         " \"method\" value must be one of"
                         " rk2, rk4, rkf54, rkf54m, or rkf54s.");
  }

  VerifyAllInitArgsUsed();

  // Setup outputs
  max_dm_dt_output.Setup(this,InstanceName(),"Max dm/dt","deg/ns",0,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);
  min_m_output.Setup(this,InstanceName(),"Min |m|","",0,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);
  max_m_output.Setup(this,InstanceName(),"Max |m|","",0,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);
  max_mdev_output.Setup(this,InstanceName(),"Max norm error","",0,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);
  dE_dt_output.Setup(this,InstanceName(),"dE/dt","J/s",0,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);
  delta_E_output.Setup(this,InstanceName(),"Delta E","J",0,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);
  dm_dt_output.Setup(this,InstanceName(),"dm/dt","rad/s",1,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);
  mxH_output.Setup(this,InstanceName(),"mxH","A/m",1,
     &Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs);

  // Setup outputs
  max_dm_dt_output.Register(director,-5);
  min_m_output.Register(director,-5);
  max_m_output.Register(director,-5);
  max_mdev_output.Register(director,-5);
  dE_dt_output.Register(director,-5);
  delta_E_output.Register(director,-5);
  dm_dt_output.Register(director,-5);
  mxH_output.Register(director,-5);

  // dm_dt and mxH output caches are used for intermediate storage,
  // so enable caching.
  dm_dt_output.CacheRequestIncrement(1);
  mxH_output.CacheRequestIncrement(1);

  // Reserve space for temp_state; see Step() method below
  director->ReserveSimulationStateRequest(1);
}

BOOL Oxs_ModLLG_RungeKuttaEvolve::Init()
{
  // Free scratch space allocated by previous problem (if any)
  vtmpA.Release();   vtmpB.Release();
  vtmpC.Release();   vtmpD.Release();

  max_step_increase = max_step_increase_limit;

  // Setup step_headroom and reject_ratio
  step_headroom = max_step_headroom;
  reject_ratio = reject_goal;

  energy_state_id=0;   // Mark as invalid state
  next_timestep=0.;    // Dummy value

  energy = 0.0; // Ensure all cache arrays are reset
  temp_energy = 0.0;
  Heff = ThreeVector(0.0,0.0,0.0);
  tmpHeff = ThreeVector(0.0,0.0,0.0);
  integrated_norm_error = 0.0;
  new_integrated_norm_error = 0.0;
  last_norm_error = 0.0;
  new_norm_error = 0.0;

  return Oxs_TimeEvolver::Init();
}


Oxs_ModLLG_RungeKuttaEvolve::~Oxs_ModLLG_RungeKuttaEvolve()
{}


void Oxs_ModLLG_RungeKuttaEvolve::Calculate_dm_dt
(const Oxs_SimState& state_,
 const Oxs_MeshValue<ThreeVector>& mxH_,
 const Oxs_MeshValue<ThreeVector>* old_dm_dt_ptr_,
 REAL8m pE_pt_, REAL8m timestep_,
 Oxs_MeshValue<ThreeVector>& dm_dt_,
 REAL8m& max_dm_dt_,REAL8m& dE_dt_,REAL8m& min_timestep_,
 REAL8m& min_m_,REAL8m& max_m_)
{ // Imports: state, mxH_, old_dm_dt_, pE_pt_, timestep_
  // Exports: dm_dt_, max_dm_dt_, dE_dt_
  //          min_m_, max_m_
  // NOTE: dm_dt_ is allowed, and in fact is encouraged,
  //   to be the same as mxH_.  In this case, mxH_ is
  //   overwritten by dm_dt on return.
  const Oxs_Mesh& mesh_ = *(state_.mesh);
  const Oxs_MeshValue<REAL8m>& Ms_ = *(state_.Ms);
  const Oxs_MeshValue<ThreeVector>& spin_ = state_.spin;
  const UINT4m size = mesh_.Size(); // Assume import data are compatible
  REAL8m coef1 = 0.;
  if(do_precess) coef1 = -fabs(gamma);
  REAL8m coef2 = fabs(alpha*gamma);
  ThreeVector scratch;

  // Move mxH_ data into dm_dt_.  This is fallback behavior for
  // the case where mxH_ and dm_dt_ are not physically the same
  // storage.  The operator=() member function of Oxs_MeshValue
  // is smart enough to turn this into a NOP if &dm_dt_ == &mxH_.
  dm_dt_ = mxH_;

  // Zero torque data on fixed spins
  UpdateFixedSpinList(&mesh_);
  const UINT4m fixed_count = GetFixedSpinCount();
  for(UINT4m j=0;j<fixed_count;j++) {
    dm_dt_[GetFixedSpin(j)].Set(0.,0.,0.);
  }

  if(!integrated_norm_error.CheckMesh(&mesh_)) {
    integrated_norm_error.AdjustSize(&mesh_);
    integrated_norm_error = 0.0;
  }
  if(!new_integrated_norm_error.CheckMesh(&mesh_)) {
    new_integrated_norm_error.AdjustSize(&mesh_);
    new_integrated_norm_error = 0.0;
  }
  if(!last_norm_error.CheckMesh(&mesh_)) {
    last_norm_error.AdjustSize(&mesh_);
    last_norm_error = 0.0;
  }
  if(!new_norm_error.CheckMesh(&mesh_)) {
    new_norm_error.AdjustSize(&mesh_);
    new_norm_error = 0.0;
  }

#if 0
  REAL8m ie_step_exp = exp(-1*m_mag_ie_freq*timestep_);
#endif

  UINT4m i,max_index=0;
  REAL8m dE_dt_sum=0.0;
  REAL8m max_mxH_sq = 0.0;
  min_m_ = max_m_ = 1.0;
  for(i=0;i<size;i++) {
    if(Ms_[i]==0) {
      dm_dt_[i].Set(0.0,0.0,0.0);
    } else {
      scratch = dm_dt_[i];  // Currently holds mxH_[i]

      // Guard rail
      const double size_guard = 0.01*sqrt(DBL_MAX);
      REAL8 ssize =fabs(scratch.x)+fabs(scratch.y)+fabs(scratch.z);
      if(!Nb_IsFinite(ssize)) {
        // Bad data; punt
        scratch.Set(0.,0.,0.);
        dm_dt_[i] = scratch;
      } else {
        if(ssize>size_guard) {
          ssize = size_guard/ssize;
          scratch *= ssize;
          dm_dt_[i] = scratch;
        }
      }

      dm_dt_[i] *= coef1;   // coef1 == 0 if do_precess if false
      REAL8m mxH_sq = scratch.MagSq();
      if(mxH_sq>max_mxH_sq) {
        max_mxH_sq = mxH_sq;
        max_index = i;
      }
      dE_dt_sum += mxH_sq * Ms_[i] * mesh_.Volume(i);

      REAL8m m_mag = sqrt(spin_[i].MagSq());
      if(m_mag<min_m_) min_m_ = m_mag;
      if(m_mag>max_m_) max_m_ = m_mag;

      scratch ^= spin_[i]; // ((mxH)xm)
      scratch *= coef2;  // |alpha.gamma|((mxH)xm) = -|alpha.gamma|(mx(mxH))
      dm_dt_[i] += scratch;

      // Include |m| adjustment
      if(RT_NONE != resize_differential) {
        REAL8m error = m_mag - 1.0;
#if 0
        REAL8m orig_error= error;
#endif
        if(RT_DMDT_AVE==resize_differential) {
          if(NULL == old_dm_dt_ptr_ || 0.0 == timestep_) {
            scratch = spin_[i];
            scratch.MakeUnit();
            // scratch.Set(0.,0.,0.); // We can't compute rotation
            /// axis, so make no correction.
          } else {
            ThreeVector utmp = (*old_dm_dt_ptr_)[i];
#if 0
if(error>1e-8) {
  fprintf(stderr,"BAD SPIN (%g): %u (%# .8g,%# .8g,%# .8g)\n",
          double(error),i,
          double(spin_[i].x),double(spin_[i].y),double(spin_[i].z));
  fprintf(stderr,"...    dm_dt: (%# .8g,%# .8g,%# .8g)\n",
          double(dm_dt_[i].x),double(dm_dt_[i].y),double(dm_dt_[i].z));
  fprintf(stderr,"...old dm_dt: (%# .8g,%# .8g,%# .8g)\n",
          double(utmp.x),double(utmp.y),double(utmp.z));
}
#endif
            REAL8m odmdtsq = utmp.MagSq();
            utmp ^= dm_dt_[i];
            REAL8m ndmdtsq = dm_dt_[i].MagSq();
            REAL8m umagsq = utmp.MagSq();
            if(umagsq <= odmdtsq*ndmdtsq*REAL8_EPSILON) {
              // dm_dt's too close to extract rotation axis,
              // so just go back towards origin
              scratch = spin_[i];
              scratch.MakeUnit();
              if(fabs(error)>1e-9) {
                fprintf(stderr,"Bad (%g) spin %u skipped: no axis\n",
                        static_cast<double>(error),i); /**/
              }
            } else {
              utmp *= 1.0/sqrt(umagsq);
              REAL8m h = spin_[i] * utmp;
              if(h<0.0) {
                // Redirect u to point towards same halfspace
                // as spin.  For purposes here we don't
                // need the absolute direction of u, only
                // its axis.  This probably isn't necessary,
                // but reduces cases to check, anyway.
                utmp *= -1;
                h *= -1;
              }
#if 0
if(error>1e-8) {
  fprintf(stderr,"...raxis: (%# .8g,%# .8g,%# .8g)\n",
          static_cast<double>(utmp.x),
          static_cast<double>(utmp.y),
          static_cast<double>(utmp.z));
  fprintf(stderr,"...h=1+%g\n",static_cast<double>(h-1));
}
#endif
              if(h>=1.0-REAL8_EPSILON) {
                // Projection falls outside sphere; punt
                // THIS LOOKS LIKE MAJOR PROBLEM!
#if 0
                scratch = spin_[i];
                scratch.MakeUnit();
#else
                // Project towards pole
                scratch = utmp;
                scratch ^= spin_[i];
                scratch ^= utmp;
                scratch.MakeUnit();
                scratch *= spin_[i]*scratch;
                scratch += (h-1)*utmp;
                scratch.MakeUnit();

                // error = sqrt(error*error + 2*(1+error-h));
                // error *= 1 + (h-1)*(h-1)*1e22; // ad hoc
                scratch *= 1 + (h-1)*(h-1)*1e22; // ad hoc

#if 0
if(error>1e-8) {
  fprintf(stderr,"...scratchA: (%# .8g,%# .8g,%# .8g)\n",
          static_cast<double>(scratch.x),
          static_cast<double>(scratch.y),
          static_cast<double>(scratch.z));
}
#endif
                // error = 2.0; // maximum error
                // error *= exp((h-1)*1e9); // EXPERIMENTAL!!!
#endif
#if 0
                if(fabs(error)>1e-9) {
fprintf(stderr,"Bad (%g) spin %u punted: outside sphere\n",
        static_cast<double>(error),i); /**/
                }
#endif
              } else {
                scratch = utmp;       // Should this be u x spin x u
                scratch ^= spin_[i];  // or u x dm_dt ???
                scratch ^= utmp;
                scratch.MakeUnit();

#if 0
if(error>1e-8) {
  fprintf(stderr,"...scratchB: (%# .8g,%# .8g,%# .8g)\n",
          static_cast<double>(scratch.x),
          static_cast<double>(scratch.y),
          static_cast<double>(scratch.z));
}
#endif
                // error = (scratch*spin_[i]) - sqrt(1-h*h);

                // error = (scratch*spin_[i])/sqrt(1-h*h) - 1.0;
                // Limit size
                // if(error>2.0)  error =  2.0;
                // if(error<-2.0) error = -2.0;
              }
            }
          }
        } else {
          // resize_differential == RT_NORM
          scratch = spin_[i];
          if(m_mag>=0.5 || 1.0<m_mag*DBL_MAX) scratch *= 1.0/m_mag;
        }


#if 0
        if(error>0.0) {
          error *= (1. + 1e11*error);
        } else {
          // error *= (1. + 1e11*fabs(error));
        }
#endif


        // Limit error size
        if(error> 1.0) error =  1.0;
        if(error<-0.5) error = -0.5;

#if 0
# if 0
        REAL8m inm = ie_step_exp*integrated_norm_error[i]
          + timestep_*error;
        new_integrated_norm_error[i] = inm;

        error += inm;

        REAL8m sign = -1.0;
        if(error<0.0) {
          error *= -1;
          sign = 1.0;
        }
        REAL8m scale = sign*sqrt(dm_dt_[i].MagSq())
          *sqrt((2+error)*error)*m_mag_adj;

        scale *= 1.0 - exp(-1.0*m_mag_width*error*error);
# else
        REAL8m scale = -1*error*sqrt(dm_dt_[i].MagSq())*m_mag_adj;

# endif
#else
# if 0
        REAL8m inm = ie_step_exp*integrated_norm_error[i]
          + timestep_*error;
        new_integrated_norm_error[i] = inm;
# else
        REAL8m inm = integrated_norm_error[i] + timestep_*error;
        new_integrated_norm_error[i] = inm;
        // Note: Period scaling already divided into m_mag_ie_adj
# endif

        REAL8m dnm = (error - last_norm_error[i]);
        if(timestep_>=1.0 || fabs(dnm) < DBL_MAX * timestep_) {
          dnm /= timestep_;
        } else {
          dnm = 0.0;
        }
        new_norm_error[i] = error;

        REAL8m scale
          = m_mag_adj*error + m_mag_ie_adj*inm + m_mag_de_adj*dnm;

#if 0
fprintf(stderr,"p=%g, i=%g, d=%g\n",
        static_cast<double>(m_mag_adj*error),
        static_cast<double>(m_mag_ie_adj*inm),
        static_cast<double>(m_mag_de_adj*dnm)); /**/
fprintf(stderr," e=%g, lne=%g, diff=%g\n",
        static_cast<double>(error),
        static_cast<double>(last_norm_error[i]),
        static_cast<double>(error-last_norm_error[i])); /**/
#endif

        const double guard = 1e-50; 
        REAL8m dm_dt_size = sqrt(dm_dt_[i].MagSq()); 
        if(fabs(scale)<1.0 || dm_dt_size<guard*DBL_MAX/fabs(scale)) {
          scale *= dm_dt_size;
        } else {
          if(scale<0.0) scale = -guard*DBL_MAX;
          else          scale =  guard*DBL_MAX;
        }
        scale *= -1;

#endif
        scratch *= scale;

#if 0
fprintf(stderr,"error=%g\n..scale=%g\n..adj=%g, ie_adj=%g\n"
        "..inm=%g\n",
        static_cast<double>(error*sign*-1),
        static_cast<double>(scale),
        static_cast<double>(m_mag_adj),
        static_cast<double>(m_mag_ie_adj),
        static_cast<double>(inm)); /**/
fprintf(stderr,"+++ spin = (%#g,%#g,%#g)\n",
        static_cast<double>(spin_[i].x),
        static_cast<double>(spin_[i].y),
        static_cast<double>(spin_[i].z));
fprintf(stderr,"+++ dmdt = (%#g,%#g,%#g)\n",
        static_cast<double>(dm_dt_[i].x),
        static_cast<double>(dm_dt_[i].y),
        static_cast<double>(dm_dt_[i].z));
fprintf(stderr,"+++ adj  = (%#g,%#g,%#g)\n",
        static_cast<double>(scratch.x),
        static_cast<double>(scratch.y),
        static_cast<double>(scratch.z));
#endif

        dm_dt_[i] += scratch; 

#if 0
if(orig_error>1e-8) {
  fprintf(stderr,"...pe=%g, ie=%g, de=%g\n",
          static_cast<double>(m_mag_adj*error),
          static_cast<double>(m_mag_ie_adj*inm),
          static_cast<double>(m_mag_de_adj*dnm)); /**/
  fprintf(stderr,"... err=%g, lne=%g, diff=%g\n",
          static_cast<double>(error),
          static_cast<double>(last_norm_error[i]),
          static_cast<double>(error-last_norm_error[i])); /**/
  fprintf(stderr,"...correction: (%# .8g,%# .8g,%# .8g)\n",
          static_cast<double>(scratch.x),
          static_cast<double>(scratch.y),
          static_cast<double>(scratch.z));
  fprintf(stderr,"...final dmdt: (%# .8g,%# .8g,%# .8g)\n",
          static_cast<double>(dm_dt_[i].x),
          static_cast<double>(dm_dt_[i].y),
          static_cast<double>(dm_dt_[i].z));
}
#endif

      }
    }
  }
  

  REAL8m max_dm_dt_sq = dm_dt_[max_index].MagSq();
  max_dm_dt_ = sqrt(max_dm_dt_sq);
  dE_dt_ = -1 * MU0 * coef2 * dE_dt_sum + pE_pt_;
  /// The first term is (partial E/partial M)*dM/dt, the
  /// second term is (partial E/partial t)*dt/dt.  Note that,
  /// provided Ms_[i]>=0, that by constructions dE_dt_sum above
  /// is always non-negative, so dE_dt_ can only be made positive
  /// by positive pE_pt_.

  // Get bound on smallest stepsize that would actually
  // change spin new_max_dm_dt_index:
  min_timestep_ = DBL_MAX/64.;
  if(max_dm_dt_>1 || REAL8_EPSILON<min_timestep_*max_dm_dt_) {
    min_timestep_ = REAL8_EPSILON/max_dm_dt_;
    // A timestep of size min_timestep will be hopelessly lost
    // in roundoff error.  So increase a bit, based on an empirical
    // fudge factor.  This fudge factor can be tested by running a
    // problem with start_dm = 0.  If the evolver can't climb its
    // way out of the stepsize=0 hole, then this fudge factor is too
    // small.  So far, the most challenging examples have been
    // problems with small cells with nearly aligned spins, e.g., in
    // a remanent state with an external field is applied at t=0.
    // Damping ratio doesn't seem to have much effect, either way.
    if(min_timestep_ < DBL_MIN) {
      // Very large max_dm_dt; punt
      min_timestep_ = DBL_MIN;
    }
    min_timestep_ *= 64;
  } else {
    // Degenerate case: max_dm_dt_ must be exactly or very nearly
    // zero.  Punt.
    min_timestep_ = 1.0;
  }

  return;
}

void Oxs_ModLLG_RungeKuttaEvolve::CheckCache(const Oxs_SimState& cstate)
{
  // Pull cached values out from cstate.
  // If cstate.Id() == energy_state_id, then cstate has been run
  // through either this method or UpdateDerivedOutputs.  Either
  // way, all derived state data should be stored in cstate,
  // except currently the "energy" mesh value array, which is
  // stored independently inside *this.  Eventually it should
  // probably be moved in some fashion into cstate too.
  if(energy_state_id != cstate.Id()) {
    // cached data out-of-date
    UpdateDerivedOutputs(cstate);
  }
  BOOL cache_good = 1;
  REAL8m max_dm_dt,min_m,max_m,max_mdev,dE_dt,delta_E,pE_pt;
  REAL8m timestep_lower_bound;  // Smallest timestep that can actually
  /// change spin with max_dm_dt (due to REAL8_EPSILON restrictions).
  /// The next timestep is based on the error from the last step.  If
  /// there is no last step (either because this is the first step,
  /// or because the last state handled by this routine is different
  /// from the incoming current_state), then timestep is calculated
  /// so that max_dm_dt * timestep = start_dm.

  cache_good &= cstate.GetDerivedData("Max dm/dt",max_dm_dt);
  cache_good &= cstate.GetDerivedData("Min |m|",min_m);
  cache_good &= cstate.GetDerivedData("Max |m|",max_m);
  cache_good &= cstate.GetDerivedData("Max norm error",max_mdev);
  cache_good &= cstate.GetDerivedData("dE/dt",dE_dt);
  cache_good &= cstate.GetDerivedData("Delta E",delta_E);
  cache_good &= cstate.GetDerivedData("pE/pt",pE_pt);
  cache_good &= cstate.GetDerivedData("Timestep lower bound",
                                      timestep_lower_bound);
  cache_good &= (energy_state_id == cstate.Id());
  cache_good &= (dm_dt_output.cache.state_id == cstate.Id());

  if(!cache_good) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_RungeKuttaEvolve::CheckCache: Invalid data cache.");
  }
}


void
Oxs_ModLLG_RungeKuttaEvolve::AdjustState
(REAL8m hstep,
 REAL8m mstep,
 const Oxs_SimState& old_state,
 const Oxs_MeshValue<ThreeVector>& dm_dt,
 Oxs_SimState& new_state,
 REAL8m& norm_error) const
{
  const Oxs_MeshValue<ThreeVector>& old_spin = old_state.spin;
  Oxs_MeshValue<ThreeVector>& new_spin = new_state.spin;

  
  if(!dm_dt.CheckMesh(old_state.mesh)) {
    throw Oxs_Ext::Error(this,
                         "Oxs_ModLLG_RungeKuttaEvolve::AdjustState:"
                         " Import spin and dm_dt are different sizes.");
  }

  new_spin.AdjustSize(old_state.mesh);
  const UINT4m size = old_state.mesh->Size();

  REAL8m min_normsq = DBL_MAX;
  REAL8m max_normsq = 0.0;
  ThreeVector tempspin;
  UINT4m i;
  for(i=0;i<size;++i) {
    tempspin = dm_dt[i];
    tempspin *= mstep;

#ifdef OLDE_CODE
    // For improved accuracy, adjust step vector so that
    // to first order m0 + adjusted_step = v/|v| where
    // v = m0 + step.
    REAL8m adj = 0.5 * tempspin.MagSq();
    tempspin -= adj*old_spin[i];
    tempspin *= 1.0/(1.0+adj);
#endif // OLDE_CODE

    tempspin += old_spin[i];

#ifdef OLDE_CODE
    REAL8m magsq = tempspin.MakeUnit();
#else // OLDE_CODE
    REAL8m magsq = tempspin.MagSq();
#endif // OLDE_CODE

    if(magsq<min_normsq) min_normsq=magsq;
    if(magsq>max_normsq) max_normsq=magsq;

    new_spin[i] = tempspin;
  }
  norm_error = ComputeMaxNormError(sqrt(min_normsq),sqrt(max_normsq));

  // Adjust time and iteration fields in new_state
  new_state.last_timestep=hstep;
  if(old_state.stage_number != new_state.stage_number) {
    // New stage
    new_state.stage_start_time = old_state.stage_start_time
                                + old_state.stage_elapsed_time;
    new_state.stage_elapsed_time = new_state.last_timestep;
  } else {
    new_state.stage_start_time = old_state.stage_start_time;
    new_state.stage_elapsed_time = old_state.stage_elapsed_time
                                  + new_state.last_timestep;
  }

  // Don't touch iteration counts. (?!)  The problem is that one call
  // to Oxs_ModLLG_RungeKuttaEvolve::Step() takes 2 half-steps, and it is the
  // result from these half-steps that are used as the export state.
  // If we increment the iteration count each time through here, then
  // the iteration count goes up by 2 for each call to Step().  So
  // instead, we leave iteration count at whatever value was filled
  // in by the Oxs_ModLLG_RungeKuttaEvolve::NegotiateTimeStep() method.
}


void Oxs_ModLLG_RungeKuttaEvolve::UpdateTimeFields
(const Oxs_SimState& cstate,
 Oxs_SimState& nstate,
 REAL8m stepsize) const
{
  nstate.last_timestep=stepsize;
  if(cstate.stage_number != nstate.stage_number) {
    // New stage
    nstate.stage_start_time = cstate.stage_start_time
                              + cstate.stage_elapsed_time;
    nstate.stage_elapsed_time = stepsize;
  } else {
    nstate.stage_start_time = cstate.stage_start_time;
    nstate.stage_elapsed_time = cstate.stage_elapsed_time
                                + stepsize;
  }
}

void Oxs_ModLLG_RungeKuttaEvolve::NegotiateTimeStep
(const Oxs_TimeDriver* driver,
 const Oxs_SimState&  cstate,
 Oxs_SimState& nstate,
 REAL8m stepsize,
 BOOL use_start_dm,
 BOOL& force_step,
 BOOL& driver_set_step) const
{ // This routine negotiates with driver over the proper step size.
  // As a side-effect, also initializes the nstate data structure,
  // where nstate is the "next state".

  // Pull needed cached values out from cstate.
  REAL8m max_dm_dt;
  if(!cstate.GetDerivedData("Max dm/dt",max_dm_dt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_RungeKuttaEvolve::NegotiateTimeStep: max_dm_dt not cached.");
  }
  REAL8m timestep_lower_bound=0.;  // Smallest timestep that can actually
  /// change spin with max_dm_dt (due to REAL8_EPSILON restrictions).
  if(!cstate.GetDerivedData("Timestep lower bound",
                            timestep_lower_bound)) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_RungeKuttaEvolve::NegotiateTimeStep: "
       " timestep_lower_bound not cached.");
  }
  if(timestep_lower_bound<=0.0) {
    throw Oxs_Ext::Error(this,
       "Oxs_ModLLG_RungeKuttaEvolve::NegotiateTimeStep: "
       " cached timestep_lower_bound value not positive.");
  }

  // The next timestep is based on the error from the last step.  If
  // there is no last step (either because this is the first step,
  // or because the last state handled by this routine is different
  // from the incoming current_state), then timestep is calculated
  // so that max_dm_dt * timestep = start_dm.
  if(use_start_dm || stepsize<=0.0) {
    if(start_dm < sqrt(DBL_MAX/4) * max_dm_dt) {
      stepsize = step_headroom * start_dm / max_dm_dt;
    } else {
      stepsize = sqrt(DBL_MAX/4);
    }
  }

  // Insure step is not outside requested step bounds
  if(stepsize>max_timestep) stepsize = max_timestep;
  if(stepsize<min_timestep) stepsize = min_timestep;
  if(stepsize<timestep_lower_bound) stepsize = timestep_lower_bound;

  // Negotiate with driver over size of next step
  driver->FillState(cstate,nstate);
  UpdateTimeFields(cstate,nstate,stepsize);

  // Update iteration count
  nstate.iteration_count = cstate.iteration_count + 1;
  nstate.stage_iteration_count = cstate.stage_iteration_count + 1;

  // Additional timestep control
  driver->FillStateSupplemental(nstate);

  // Check for forced step
  force_step = 0;
  if(nstate.last_timestep<=min_timestep ||
     nstate.last_timestep<=timestep_lower_bound) {
    // Either driver wants to force stepsize, or else step size can't
    // be reduced any further.
    force_step=1;
  }

  // Is driver responsible for stepsize?
  if(nstate.last_timestep == stepsize) driver_set_step = 0;
  else                                 driver_set_step = 1;
}


BOOL
Oxs_ModLLG_RungeKuttaEvolve::CheckError
(REAL8m global_error_order,
 REAL8m error,
 REAL8m stepsize,
 REAL8m reference_stepsize,
 REAL8m max_dm_dt,
 REAL8m& new_stepsize)
{ // Returns 1 if step is good, 0 if error is too large.
  // Export new_stepsize is set to suggested stepsize
  // for next step.
  //
  // new_stepsize is initially filled with a relative stepsize
  // adjustment ratio, e.g., 1.0 means no change in stepsize.
  // At the end of this routine this ratio is multiplied by
  // stepsize to get the actual absolute stepsize.
  //
  // The import stepsize is the size (in seconds) of the actual
  // step.  The new_stepsize is computed from this based on the
  // various error estimates.  An upper bound is placed on the
  // size of new_stepsize relative to the imports stepsize and
  // reference_stepsize.  reference_stepsize has no effect if
  // it is smaller than max_step_increase*stepsize.  It is
  // usually used only in the case where the stepsize was
  // artificially reduced by, for example, a stage stopping
  // criterion.
  //
  // NOTE: This routine assumes the local error order is
  //     global_error_order + 1.

  BOOL good_step = 1;
  BOOL error_checked=0;

  if(allowed_relative_step_error>=0. || allowed_error_rate>=0.) {
    // Determine tighter rate bound.
    REAL8m rate_error = 0.0;
    if(allowed_relative_step_error<0.) {
      rate_error = allowed_error_rate;
    } else if(allowed_error_rate<0.) {
      rate_error = allowed_relative_step_error * max_dm_dt;
    } else {
      rate_error = allowed_relative_step_error * max_dm_dt;
      if(rate_error>allowed_error_rate) {
        rate_error = allowed_error_rate;
      }
    }
    rate_error *= stepsize;

    // Rate check
    if(error>rate_error) {
      good_step = 0;
      new_stepsize = pow(rate_error/error,1.0/global_error_order);
    } else {
      REAL8m ratio = 0.125*DBL_MAX;
      if(error>=1 || rate_error<ratio*error) {
        REAL8m test_ratio = rate_error/error;
        if(test_ratio<ratio) ratio = test_ratio;
      }
      new_stepsize = pow(ratio,1.0/global_error_order);
    }
    error_checked = 1;
  }

  // Absolute error check
  if(allowed_absolute_step_error>=0.0) {
    REAL8m test_stepsize = 0.0;
    REAL8m local_error_order = global_error_order + 1.0;
    if(error>allowed_absolute_step_error) {
      good_step = 0;
      test_stepsize = pow(allowed_absolute_step_error/error,
                          1.0/local_error_order);
    } else {
      REAL8m ratio = 0.125*DBL_MAX;
      if(error>=1 || allowed_absolute_step_error<ratio*error) {
        REAL8m test_ratio = allowed_absolute_step_error/error;
        if(test_ratio<ratio) ratio = test_ratio;
      }
      test_stepsize = pow(ratio,1.0/local_error_order);
    }
    if(!error_checked || test_stepsize<new_stepsize) {
      new_stepsize = test_stepsize;
    }
    error_checked = 1;
  }

  if(error_checked) {
    new_stepsize *= step_headroom;
    if(new_stepsize<max_step_decrease) {
      new_stepsize = max_step_decrease*stepsize;
    } else {
      new_stepsize *= stepsize;
      REAL8m step_bound = stepsize * max_step_increase;
      const REAL8m refrat = 0.85;  // Ad hoc value
      if(stepsize<reference_stepsize*refrat) {
        step_bound = OC_MIN(step_bound,reference_stepsize);
      } else if(stepsize<reference_stepsize) {
        REAL8m ref_bound = reference_stepsize + (max_step_increase-1)
          *(stepsize-reference_stepsize*refrat)/(1-refrat);
        /// If stepsize = reference_stepsize*refrat,
        ///     then ref_bound = reference_stepsize
        /// If stepsize = reference_stepsize,
        ///     then ref_bound = step_bound
        /// Otherwise, linear interpolation on stepsize.
        step_bound = OC_MIN(step_bound,ref_bound);
      }
      if(new_stepsize>step_bound) new_stepsize = step_bound;
    }
  } else {
    new_stepsize = stepsize;
  }

  return good_step;
}

BOOL Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins
(ResizeType style,
 const Oxs_MeshValue<ThreeVector>& spin1,
 Oxs_MeshValue<ThreeVector>& spin2,
 const Oxs_MeshValue<ThreeVector>& H1,
 const Oxs_MeshValue<ThreeVector>& H2,
 const Oxs_MeshValue<ThreeVector>& dm_dt1,
 const Oxs_MeshValue<ThreeVector>& dm_dt2)
{
  BOOL spins_changed = 1;
  const UINT4m size = spin2.Size();
  if(style == RT_DMDT_AVE) {
    if(dm_dt1.Size() != size || dm_dt2.Size() != size) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins:"
                           " Programming error; wrong dm_dt array size.");
    }
    for(UINT4m i=0;i<size;++i) {
      // First find unit vector of approximate rotation axis,
      // via dm_dt_old x dm_dt_new
      ThreeVector utmp = dm_dt1[i];
      REAL8m cdmdtsq = dm_dt1[i].MagSq();
      utmp ^= dm_dt2[i];
      REAL8m ndmdtsq = dm_dt2[i].MagSq();
      REAL8m umagsq = utmp.MagSq();
      if(umagsq < cdmdtsq*ndmdtsq*REAL8_EPSILON*1024) {
        // dm_dt's too close to extract rotation axis; punt.
        spin2[i].MakeUnit();
      } else {
        utmp *= (1.0/sqrt(umagsq));
        ThreeVector vtmp = utmp;
        vtmp ^= spin2[i];
        REAL8m Esinsq = vtmp.MagSq();
        if(Esinsq < spin2[i].MagSq()*REAL8_EPSILON*4) {
          // m and H are nearly parallel; punt
          spin2[i].MakeUnit();
        } else {
          REAL8m Ecos = spin2[i]*utmp;
          REAL8m adjsq = 1.0 - Ecos*Ecos;
          if(adjsq<=0.0) {
            // Projection of m onto H has norm bigger than one,
            // so we can't adjust norm along proposed direction.
            // Instead, project to H and then resize norm there,
            // or equivalently, just take unit vector parallel
            // to H, with proper sign
            if(Ecos<0.0) utmp *= -1.0;
            spin2[i] = utmp;
          } else {
            vtmp ^= utmp;
            REAL8m B = sqrt(adjsq/Esinsq) - 1.0;
            vtmp *= B;
            spin2[i] += vtmp;
          }
        }
      }
    }
  } else if(style == RT_DMDT_H) {
    // Project along H x dm_dt
    if(H2.Size() != size) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins:"
                           " Programming error; wrong H2 array size.");
    }
    if(dm_dt2.Size() != size) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins:"
                           " Programming error; wrong dm_dt2 array size.");
    }
    for(UINT4m i=0;i<size;++i) {
      ThreeVector utmp = H2[i];
      utmp ^= dm_dt2[i];
      REAL8m umagsq = utmp.MagSq();
      if(umagsq <= H2[i].MagSq() * dm_dt2[i].MagSq() * REAL8_EPSILON) {
        // H2 and dm_dt2 too close to parallel; punt
        spin2[i].MakeUnit();
      } else {
        utmp *= (1.0/sqrt(umagsq));
        REAL8m dot = spin2[i]*utmp;
        REAL8m base = 1 + dot*dot - spin2[i].MagSq(); 
        if(base<0.0) {
          // Not possible to resize spin to unit size along utmp; punt
          spin2[i].MakeUnit();
        } else {
          base  = dot + sqrt(base);
          utmp *= base;
          spin2[i] -= utmp;
        }
      }
    }
  } else if(style == RT_H) {
    // Project along H2xmxH2
    if(H2.Size() != size) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins:"
                           " Programming error; wrong H2 array size.");
    }
    for(UINT4m i=0;i<size;++i) {
      ThreeVector H = H2[i];
      H.MakeUnit();
      ThreeVector vtmp = spin2[i];
      vtmp ^= H;
      REAL8m Ecos = spin2[i] * H;
      REAL8m Esin = sqrt(vtmp.MagSq());
      if(fabs(Ecos)>=1.0 || (fabs(Esin)<1.0 && 1.0>DBL_MAX*fabs(Esin))) {
        // Bad data; just renormalize
        spin2[i].MakeUnit();
      } else {
        vtmp ^= H;
        vtmp *= 1.0/Esin;
        REAL8m B = Esin - sqrt(1-Ecos*Ecos);
        vtmp *= B;
        spin2[i] += vtmp;
      }
    }
  } else if(style == RT_H_AVE) {
    // Average H1 and H2, and project along HxmxH
    if(H1.Size() != size || H2.Size() != size) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins:"
                           " Programming error; wrong H array size.");
    }
    for(UINT4m i=0;i<size;++i) {
      ThreeVector H = H1[i] + H2[i];
      REAL8m Hmagsq = H.MagSq();
      if(Hmagsq <= (H1[i].MagSq()+H2[i].MagSq())*REAL8_EPSILON) {
        // Severe cancellation in H average; punt
        spin2[i].MakeUnit();
      } else {
        H *= (1.0/sqrt(Hmagsq));
        ThreeVector vtmp = spin2[i];
        vtmp ^= H;
        REAL8m Ecos = spin2[i] * H;
        REAL8m Esin = sqrt(vtmp.MagSq());
        if(fabs(Ecos)>=1.0 || (fabs(Esin)<1.0 && 1.0>DBL_MAX*fabs(Esin))) {
          // Bad data; just renormalize
          spin2[i].MakeUnit();
        } else {
          vtmp ^= H;
          vtmp *= 1.0/Esin;
          REAL8m B = Esin - sqrt(1-Ecos*Ecos);
          vtmp *= B;
          spin2[i] += vtmp;
        }
      }
    }
  } else if(style == RT_H_MOD) {
    // Attempt to compute m-invariant portion of H
    if(H1.Size() != size || H2.Size() != size) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins:"
                           " Programming error; wrong H array size.");
    }
    if(spin1.Size() != size) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::ResizeSpins:"
                           " Programming error; wrong spin1 array size.");
    }
    for(UINT4m i=0;i<size;++i) {
      ThreeVector H = H2[i] + H1[i];
      ThreeVector dH = H2[i] - H1[i];
      REAL8m size_check = H1[i].MagSq() + H2[i].MagSq();
      ThreeVector m = spin2[i] + spin1[i];
      ThreeVector dm = spin2[i] - spin1[i];
      REAL8m dm_magsq = dm.MagSq();
      REAL8m wgt = (dH*dm);
      if(dm_magsq<REAL8_EPSILON ||
         (dm_magsq<1.0 && wgt > (0.25 * DBL_MAX) * dm_magsq)) {
        // No data; punt
        spin2[i].MakeUnit();
      } else {
        wgt /= dm_magsq;
        m *= wgt;
        H -= m;
        H *= 0.5;
        REAL8m Hmagsq = H.MagSq();
        if(Hmagsq <= REAL8_EPSILON*size_check) {
          // Severe cancellation in H computation; punt
          spin2[i].MakeUnit();
        } else {
          H *= (1.0/sqrt(Hmagsq));
          ThreeVector vtmp = spin2[i];
          vtmp ^= H;
          REAL8m Esinsq = vtmp.MagSq();
          if(Esinsq < spin2[i].MagSq()*REAL8_EPSILON*4) {
            // m and H are nearly parallel; punt
            spin2[i].MakeUnit();
          } else {
            REAL8m Ecos = spin2[i] * H;
            REAL8m adjsq = 1.0 - Ecos*Ecos;
            if(adjsq<=0.0) {
              // Projection of m onto H has norm bigger than one,
              // so we can't adjust norm along proposed direction.
              // Instead, project to H and then resize norm there,
              // or equivalently, just take unit vector parallel
              // to H, with proper sign
              if(Ecos<0.0) H *= -1.0;
              spin2[i] = H;
            } else {
              vtmp ^= H;
              REAL8m B = 1.0 - sqrt(adjsq/Esinsq);
              vtmp *= B;
              spin2[i] += vtmp;
            }
          }
        }
      }
    }
  } else if(style == RT_NORM) {
    for(UINT4m i=0;i<size;++i) {
      spin2[i].MakeUnit();
    }
  } else {
    // RT_NONE or unsupported
    spins_changed = 0;
  }

#if !defined(NDEBUG)
  if(style != RT_NONE) {
    for(UINT4m i=0;i<size;++i) {
      if(fabs(spin2[i].MagSq()-1.0)> 2 * 64 * REAL8_EPSILON) {
        char buf[1024];
        Oc_Snprintf(buf,sizeof(buf),
                    "Spin renormalize error; |spin[%u]|-1=%g\n",i,
                    static_cast<double>(sqrt(spin2[i].MagSq())-1.0));
        throw Oxs_Ext::Error(this,buf);
      }
    }
  }
#endif

  return spins_changed;
}


REAL8m
Oxs_ModLLG_RungeKuttaEvolve::ComputeMaxNormError
(REAL8m min_norm,
 REAL8m max_norm) const
{ // Convenience function.  Also provides a workaround for some
  // badly broken gcc optimizers.
  // NOTE: This returns a signed quantity!
  REAL8m min_error = min_norm - 1.0;
  REAL8m max_error = max_norm - 1.0;
  REAL8m norm_error = max_error;
  if(fabs(min_error)>fabs(max_error)) norm_error = min_error;
  return norm_error;
}

REAL8m
Oxs_ModLLG_RungeKuttaEvolve::ComputeMaxNormError
(const Oxs_MeshValue<ThreeVector>& spin) const
{
  const UINT4m size = spin.Size();
  if(size<1) return 0.0;
  REAL8m magsq = spin[0].MagSq();
  REAL8m minsq = magsq;
  REAL8m maxsq = magsq;
  for(UINT4m i=1;i<size;++i) {
    magsq = spin[i].MagSq();
    if(magsq<minsq) minsq = magsq;
    if(magsq>maxsq) maxsq = magsq;
  }
  return ComputeMaxNormError(sqrt(minsq),sqrt(maxsq));
}

void Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaStep2
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& current_dm_dt,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,REAL8m& global_error_order,
 REAL8m& norm_error,
 BOOL& new_energy_and_dmdt_computed)
{ // This routine performs a second order Runge-Kutta step, with
  // error estimation.  The form used is the "modified Euler"
  // method due to Collatz (1960).
  //
  // A general RK2 step involves
  //
  //  dm_dt1 = dm_dt(t1,m1)
  //  dm_dt2 = dm_dt(t1+a.h,m1+a.h.dm_dt1)
  //  m2 = m1 + h.( (1-1/(2a)).dm_dt1 + (1/(2a)).dm_dt2 )
  //
  // where 0<a<=1 is a free parameter.  Taking a=1/2 gives
  //
  //  dm_dt1 = dm_dt(t1,m1)
  //  dm_dt2 = dm_dt(t1+h/2,m1+dm_dt1.h/2)
  //  m2 = m1 + dm_dt2*h + O(h^3)
  //
  // This is the "modified Euler" method from Collatz (1960).
  // Setting a=1 yields
  //
  //  dm_dt1 = dm_dt(t1,m1)
  //  dm_dt2 = dm_dt(t1+h,m1+dm_dt1.h)
  //  m2 = m1 + (dm_dt1 + dm_dt2).h/2 + O(h^3)
  //
  // which is the method of Heun (1900).  For details see
  // J. Stoer and R. Bulirsch, "Introduction to Numerical
  // Analysis," Springer 1993, Section 7.2.1, p438.
  //
  // In the code below, we use the "modified Euler" approach,
  // i.e., select a=1/2.

  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  UINT4m i;
  const UINT4m size = cstate->mesh->Size();

  REAL8m pE_pt,max_dm_dt,dE_dt,timestep_lower_bound,dummy_error;
  REAL8m min_m,max_m;

  // Calculate dm_dt2
  AdjustState(stepsize/2,stepsize/2,*cstate,current_dm_dt,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,stepsize/2,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // vtmpA holds dm_dt2
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  nstate->spin = cstate->spin;
  nstate->spin.Accumulate(stepsize,vtmpA);

  // Tweak "last_timestep" field in next_state, and adjust other
  // time fields to protect against rounding errors.
  UpdateTimeFields(*cstate,*nstate,stepsize);

  // Collect norm error info
  norm_error = ComputeMaxNormError(nstate->spin);
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference()); // Lock down

  // To estimate error, compute dm_dt at end state.
  // Error estimate is (h/3)*(end_dm_dt - dm_dt2) + O(h^4)
  GetEnergyDensity(*endstate,temp_energy,&mxH_output.cache.value,
                   &tmpHeff,pE_pt);
  mxH_output.cache.state_id=endstate->Id();
  Calculate_dm_dt(*endstate,mxH_output.cache.value,
                  &current_dm_dt,pE_pt,stepsize,
                  vtmpB,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  REAL8m mdev = ComputeMaxNormError(min_m,max_m);
  if(!endstate->AddDerivedData("Timestep lower bound",
                                timestep_lower_bound) ||
     !endstate->AddDerivedData("Max dm/dt",max_dm_dt) ||
     !endstate->AddDerivedData("Min |m|",min_m) ||
     !endstate->AddDerivedData("Max |m|",max_m) ||
     !endstate->AddDerivedData("Max norm error",mdev) ||
     !endstate->AddDerivedData("pE/pt",pE_pt) ||
     !endstate->AddDerivedData("dE/dt",dE_dt)) {
    throw Oxs_Ext::Error(this,
                         "Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaStep2:"
                         " Programming error; data cache already set.");
  }
  REAL8m max_err_sq = 0.0;
  for(i=0;i<size;++i) {
    ThreeVector tvec = vtmpB[i] - vtmpA[i];
    REAL8m err_sq = tvec.MagSq();
    if(err_sq>max_err_sq) max_err_sq = err_sq;
  }
  error_estimate = sqrt(max_err_sq)*stepsize/3.;
  global_error_order = 2.0;

  // Move end dm_dt data into vtmpA, for use by calling routine.
  // Note that end energy is already in temp_energy, as per
  // contract.
  vtmpA.Swap(vtmpB);
  new_energy_and_dmdt_computed = 1;
}

void Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaStep4
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& current_dm_dt,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,REAL8m& global_error_order,
 REAL8m& norm_error,
 BOOL& new_energy_and_dmdt_computed)
{ // This routine performs two successive "classical" Runge-Kutta
  // steps of size stepsize/2, and stores the resulting magnetization
  // state into the next_state export.  Additionally, a single
  // step of size stepsize is performed, and used for estimating
  // the error.  This error is returned in the export error_estimate;
  // This is the largest error detected cellwise, in radians.  The
  // export global_error_order is always set to 4 by this routine.
  // (The local error order is one better, i.e., 5.)  The norm_error
  // export is set to the cellwise maximum deviation from unit norm
  // across all the spins in the final state, before renormalization.

  // A single RK4 step involves
  //  dm_dt1 = dm_dt(t1,m1)
  //  dm_dt2 = dm_dt(t1+h/2,m1+dm_dt1*h/2)
  //  dm_dt3 = dm_dt(t1+h/2,m1+dm_dt2*h/2)
  //  dm_dt4 = dm_dt(t1+h,m1+dm_dt3*h)
  //  m2 = m1 + dm_dt1*h/6 + dm_dt2*h/3 + dm_dt3*h/3 + dm_dt4*h/6 + O(h^5)
  // To improve accuracy, for each step accumulate dm_dt?, where
  // ?=1,2,3,4, into m2 with proper weights, and add in m1 at the end.

  // To calculate dm_dt, we first fill next_state with the proper
  // spin data (e.g., m1+dm_dt2*h/2).  The utility routine AdjustState
  // is applied for this purpose.  Then next_state is locked down,
  // so that it will have a valid state_id, and GetEnergyDensity
  // is called in order to get mxH and related data.  This is then
  // passed to Calculate_dm_dt.  Note that dm_dt is not calculated
  // for the final state.  This is left for the calling routine,
  // which first examines the error estimates to decide whether
  // or not the step will be accepted.  If the step is rejected,
  // then the energy and dm_dt do not need to be calculated.

  // Scratch space usage:
  //   vtmpA is used to store, successively, dm_dt2, dm_dt3 and
  //      dm_dt4.  In calculating dm_dt?, vtmpA is first filled
  //      with mxH by the GetEnergyDensity() routine.
  //      Calculate_dm_dt() allows import mxH and export dm_dt
  //      to be the same MeshValue structure.
  //   vtmpB is used to accumulate the weighted sums of the dm_dt's,
  //      as explained above.  To effect a small efficiency gain,
  //      dm_dt1 is calculated directly into vtmpB, in the same
  //      manner as the other dm_dt's are filled into vtmpA.
  //   tempState is used to hold the middle state when calculating
  //      the two half steps, and the end state for the single
  //      full size step.

  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  Oxs_SimState* tstate = &(temp_state_key.GetWriteReference());
  nstate->CloneHeader(*tstate);

  UINT4m i;
  const UINT4m size = cstate->mesh->Size();

  REAL8m pE_pt,max_dm_dt,dE_dt,timestep_lower_bound;
  REAL8m dummy_error;
  REAL8m min_m,max_m;

  // Do first half step.  Because dm_dt1 is already calculated,
  // we fill dm_dt2 directly into vtmpB.
  AdjustState(stepsize/4,stepsize/4,*cstate,current_dm_dt,
              temp_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(temp_state_key.GetReadReference(),temp_energy,
                   &vtmpB,&tmpHeff,pE_pt);
  Calculate_dm_dt(temp_state_key.GetReadReference(),vtmpB,
                  &current_dm_dt,pE_pt,0.25*stepsize,
                  vtmpB,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // vtmpB currently holds dm_dt2
  AdjustState(stepsize/4,stepsize/4,*cstate,vtmpB,
              temp_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(temp_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(temp_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,0.25*stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  vtmpB += vtmpA;
  // vtmpA holds dm_dt3, vtmpB holds dm_dt2 + dm_dt3.
  AdjustState(stepsize/2,stepsize/2,*cstate,vtmpA,
              temp_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(temp_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(temp_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,0.5*stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // vtmpA holds dm_dt4
  vtmpA += current_dm_dt;
  vtmpB.Accumulate(0.5,vtmpA); // Add in 0.5*(dm_dt1+dm_dt4)
  tstate = &(temp_state_key.GetWriteReference());
  tstate->spin = cstate->spin;
  tstate->spin.Accumulate(stepsize/6.,vtmpB);
  // Note: state time index set to lasttime + stepsize/2
  // by dm_dt4 calculation above.  Note that "h" here is
  // stepsize/2, so the weights on dm_dt1, dm_dt2, dm_dt3 and
  // dn_dt4 are (stepsize/2)(1/6, 1/3, 1/3, 1/6), respectively.

#ifdef OLDE_CODE
  // Normalize spins in tstate
  for(i=0;i<size;++i) tstate->spin[i].MakeUnit();
#endif // OLDE_CODE

  // At this point, temp_state holds the middle point.
  // Calculate dm_dt for this state, and store in vtmpB.
  tstate = NULL; // Disable non-const access
  const Oxs_SimState* midstate
    = &(temp_state_key.GetReadReference());
  GetEnergyDensity(*midstate,temp_energy,&vtmpB,&tmpHeff,pE_pt);
  Calculate_dm_dt(*midstate,vtmpB,&current_dm_dt,pE_pt,0.5*stepsize,
                  vtmpB,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);

  // Next do second half step.  Store end result in next_state
  AdjustState(stepsize/4,stepsize/4,*midstate,vtmpB,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,0.75*stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  /// Note: The integrated error used inside Calculated_dm_dt only
  /// gets updated after a successful step is taken.  So the reference
  /// state for Calculate_dm_dt is cstate, note midstate, so timestep
  /// should be set to 0.75*stepsize.

  // vtmpB currently holds dm_dt1, vtmpA holds dm_dt2
  vtmpB *= 0.5;
  vtmpB += vtmpA;
  AdjustState(stepsize/4,stepsize/4,*midstate,vtmpA,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,0.75*stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  vtmpB += vtmpA;
  // vtmpA holds dm_dt3, vtmpB holds dm_dt1/2 + dm_dt2 + dm_dt3.
  AdjustState(stepsize/2,stepsize/2,*midstate,vtmpA,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // vtmpA holds dm_dt4
  vtmpB.Accumulate(0.5,vtmpA);
  nstate = &(next_state_key.GetWriteReference());
  nstate->spin = midstate->spin;
  nstate->spin.Accumulate(stepsize/6.,vtmpB);
  midstate = NULL; // We're done using midstate

  // Tweak "last_timestep" field in next_state, and adjust other
  // time fields to protect against rounding errors.
  UpdateTimeFields(*cstate,*nstate,stepsize);

  // Normalize spins in nstate, and collect norm error info
  // Normalize m2, including norm error check
  norm_error = ComputeMaxNormError(nstate->spin);

  nstate = NULL;
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference()); // Lock down

  // Repeat now for full step, storing end result into temp_state
  AdjustState(stepsize/2,stepsize/2,*cstate,current_dm_dt,
              temp_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(temp_state_key.GetReadReference(),temp_energy,
                   &vtmpB,&tmpHeff,pE_pt);
  Calculate_dm_dt(temp_state_key.GetReadReference(),vtmpB,
                  &current_dm_dt,pE_pt,0.5*stepsize,
                  vtmpB,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // vtmpB currently holds dm_dt2
  AdjustState(stepsize/2,stepsize/2,*cstate,vtmpB,
              temp_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(temp_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(temp_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,0.5*stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  vtmpB += vtmpA;
  // vtmpA holds dm_dt3, vtmpB holds dm_dt2 + dm_dt3.
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              temp_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(temp_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(temp_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // vtmpA holds dm_dt4
  vtmpA += current_dm_dt;
  vtmpB.Accumulate(0.5,vtmpA); // Add in 0.5*(dm_dt1+dm_dt4)
  tstate = &(temp_state_key.GetWriteReference());
  tstate->spin = cstate->spin;
  tstate->spin.Accumulate(stepsize/3.,vtmpB);
#ifdef OLDE_CODE
  for(i=0;i<size;++i) tstate->spin[i].MakeUnit(); // Normalize
#endif
  tstate = NULL; // Disable write access
  const Oxs_SimState* teststate
    = &(temp_state_key.GetReadReference()); // Lock down

  // Error is max|next_state.spin-temp_state.spin|/15 + O(stepsize^6)
  REAL8m max_error_sq = 0.0;
  for(i=0;i<size;++i) {
    ThreeVector tvec = endstate->spin[i] - teststate->spin[i];
    REAL8m error_sq = tvec.MagSq();
    if(error_sq>max_error_sq) max_error_sq = error_sq;
    // If we want, we can experiment with adding
    // tvec/15 to endstate->spin.  In theory, this changes
    // the error from O(stepsize^5) to O(stepsize^6)
  }
  error_estimate = sqrt(max_error_sq)/15.;
  global_error_order = 4.0;
  new_energy_and_dmdt_computed = 0;
}

void Oxs_ModLLG_RungeKuttaEvolve::RungeKuttaFehlbergBase54
(RKF_SubType method,
 REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& current_dm_dt,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,REAL8m& global_error_order,
 REAL8m& norm_error,
 BOOL& new_energy_and_dmdt_computed)
{ // Runge-Kutta-Fehlberg routine with combined 4th and 5th
  // order Runge-Kutta steps.  The difference between the
  // two results (4th vs. 5th) is used to estimate the error.
  // The largest detected error detected cellsize is returned
  // in export error_estimate.  The units on this are radians.
  // The export global_error_order is set to 4 by this routine.
  // (The local error order is one better, i.e., 5.)  The norm_error
  // export is set to the cellwise maximum deviation from unit norm
  // across all the spins in the final state, before renormalization.

  // The following coefficients appear in
  //
  //   J. R. Dormand and P. J. Prince, ``A family of embedded
  //   Runge-Kutta formulae,'' J. Comp. Appl. Math., 6, 19--26
  //   (1980).
  //
  // They are also listed in J. Stoer and R. Bulirsch's book,
  // ``Introduction to Numerical Analysis,'' Springer, 2nd edition,
  // Sec. 7.2.5, p 454, but there are a number of errors in the S&B
  // account; the reader is recommended to refer directly to the D&P
  // paper.  A follow-up paper,
  //
  //   J. R. Dormand and P. J. Prince, ``A reconsideration of some 
  //   embedded Runge-Kutta formulae,'' J. Comp. Appl. Math., 15,
  //   203--211 (1986)
  //
  // provides formulae with improved stability and higher order.
  // See also
  //
  //   J. H. Williamson, ``Low-Storage Runge-Kutta Schemes,''
  //   J. Comp. Phys., 35, 48--56 (1980).
  //
  // FORMULAE for RK5(4)7FM:
  //
  //     dm_dt1 = dm_dt(t1,m1)
  //     dm_dt2 = dm_dt(t1+ (1/5)*h, m1+h*k1);
  //     dm_dt3 = dm_dt(t1+(3/10)*h, m1+h*k2);
  //     dm_dt4 = dm_dt(t1+ (4/5)*h, m1+h*k3);
  //     dm_dt5 = dm_dt(t1+ (8/9)*h, m1+h*k4);
  //     dm_dt6 = dm_dt(t1+     1*h, m1+h*k5);
  //     dm_dt7 = dm_dt(t1+     1*h, m1+h*k6);
  //  where
  //     k1 = dm_dt1*1/5
  //     k2 = dm_dt1*3/40       + dm_dt2*9/40
  //     k3 = dm_dt1*44/45      - dm_dt2*56/15      + dm_dt3*32/9
  //     k4 = dm_dt1*19372/6561 - dm_dt2*25360/2187 + dm_dt3*64448/6561
  //               - dm_dt4*212/729
  //     k5 = dm_dt1*9017/3168  - dm_dt2*355/33     + dm_dt3*46732/5247
  //               + dm_dt4*49/176   - dm_dt5*5103/18656
  //     k6 = dm_dt1*35/384     + 0                 + dm_dt3*500/1113
  //               + dm_dt4*125/192  - dm_dt5*2187/6784  + dm_dt6*11/84
  // Then
  //     Da = dm_dt1*35/384     + 0 + dm_dt3*500/1113   + dm_dt4*125/192
  //              - dm_dt5*2187/6784      + dm_dt6*11/84
  //     Db = dm_dt1*5179/57600 + 0 + dm_dt3*7571/16695 + dm_dt4*393/640
  //              - dm_dt5*92097/339200   + dm_dt6*187/2100   + dm_dt7*1/40
  // and
  //     m2a = m1 + h*Da
  //     m2b = m1 + h*Db.
  //
  // where m2a is the 5th order estimate, which is the candidate
  // for the next state, and m2b is the 4th order estimate used
  // only for error estimation/stepsize control.  Note that the
  // 4th order estimate uses more dm/dt evaluations (7) than the
  // 5th order method (6).  This is intentional; the coefficients
  // are selected to minimize error (see D&P paper cited above).
  // The extra calculation costs are minimal, because the 7th dm_dt
  // evaluation is at the candidate next state, so it is re-used
  // on the next step unless the step rejected.
  //
  // The error estimate is
  // 
  //     E = |m2b-m2a| = h*|Db-Da| = C*h^5
  //
  // where C is a constant that can be estimated in terms of E.
  // Note that we don't need to know C in order to adjust the
  // stepsize, because stepsize adjustment involves only the
  // ratio of the old stepsize to the new, so C drops out.

  // Scratch space usage:
  //  The import next_state is used for intermediate state
  // storage for all dm_dt computations.  The final computation
  // is for dm_dt7 at m1+h*k6 = m1+h*Da, which is the candidate
  // next state.  (Da=k6; see FSAL note below.)

  // Four temporary arrays, A-D, are used:
  //
  // Step \  Temp
  // Index \ Array:  A         B         C         D
  // ------+---------------------------------------------
  //   1   |      dm_dt2       -         -         -
  //   2   |      dm_dt2      k2         -         -
  //   3   |      dm_dt2     dm_dt3      -         -
  //   4   |      dm_dt2     dm_dt3     k3         -
  //   5   |      dm_dt2     dm_dt3    dm_dt4      -
  //   6   |      dm_dt2     dm_dt3    dm_dt4     k4
  //   7   |      dm_dt2     dm_dt3    dm_dt4    dm_dt5
  //   8   |        k5       dm_dt3    dm_dt4    dm_dt5
  //   9   |      dm_dt6     dm_dt3    dm_dt4    dm_dt5
  //  10   |      k6(3,6)    dD(3,6)   dm_dt4    dm_dt5
  //  11   |      dm_dt7     dD(3,6)   dm_dt4    dm_dt5
  //  12   |      dm_dt7       dD      dm_dt4    dm_dt5
  //
  // Here dD is Db-Da.  We don't need to compute Db directly.
  // The parenthesized numbers, e.g., k6(3,6) indicates
  // a partially formed value.  The total value k6 depends
  // upon dm_dt1, dm_dt3, dm_dt4, dm_dt5, and dm_dt6.  But
  // if we form k6 directly at step 11 in array A, then we
  // lose dm_dt6 which is needed to compute dD.  Rather than
  // use an additional array, we accumulate partial results
  // into arrays A and B for k6 and dD as indicated.
  //   Note that Da = k6.  Further, note that dm_dt7 is
  // dm_dt for the next state candidate.  (This is the 'F',
  // short for 'FSAL' ("first same as last"?) in the method
  // name, RK5(4)7FM.  The 'M' means minimized error norm,
  // 7 is the number of stages, and 5(4) is the main/subsidiary
  // integration formula order.  See the D&P 1986 paper for
  // details and additional references.)

  // Coefficient arrays, a, b, dc, defined by:
  //
  //   dm_dtN = dm_dt(t1+aN*h,m1+h*kN)
  //       kN = \sum_{M=1}^{M=N} dm_dtM*bNM
  //  Db - Da = \sum dm_dtM*dcM
  //
  REAL8m a1,a2,a3,a4; // a5 and a6 are 1.0
  REAL8m b11,b21,b22,b31,b32,b33,b41,b42,b43,b44,
    b51,b52,b53,b54,b55,b61,b63,b64,b65,b66; // b62 is 0.0
  REAL8m dc1,dc3,dc4,dc5,dc6,dc7;  // c[k] = b6k, and c^[2]=c[2]=0.0,
  /// where c are the coeffs for Da, c^ for Db, and dcM = c^[M]-c[M].

  switch(method) {
  case RK547FC:
    /////////////////////////////////////////////////////////////////
    // Coefficients for Dormand & Prince RK5(4)7FC
    a1 = 1./5.;
    a2 = 3./10.;
    a3 = 6./13.;
    a4 = 2./3.;
    // a5 and a6 are 1.0

    b11 =      1./5.;
  
    b21 =      3./40.;
    b22 =      9./40.;
  
    b31 =    264./2197.;
    b32 =    -90./2197.;
    b33 =    840./2197.;
  
    b41 =    932./3645.;
    b42 =    -14./27.;
    b43 =   3256./5103.;
    b44 =   7436./25515.;
  
    b51 =   -367./513.;
    b52 =     30./19.;
    b53 =   9940./5643.;
    b54 = -29575./8208.;
    b55 =   6615./3344.;
  
    b61 =     35./432.;
    b63 =   8500./14553.;
    b64 = -28561./84672.;
    b65 =    405./704.;
    b66 =     19./196.;
    // b62 is 0.0

    // Coefs for error calculation (c^[k] - c[k]).
    // Note that c[k] = b6k, and c^[2]=c[2]=0.0
    dc1 =     11./108.    - b61;
    dc3 =   6250./14553.  - b63;
    dc4 =  -2197./21168.  - b64;
    dc5 =     81./176.    - b65;
    dc6 =    171./1960.   - b66;
    dc7 =      1./40.;
    break;

  case RK547FM:
    /////////////////////////////////////////////////////////////////
    // Coefficients for Dormand & Prince RK5(4)7FM
    a1 = 1./5.;
    a2 = 3./10.;
    a3 = 4./5.;
    a4 = 8./9.;
    // a5 and a6 are 1.0

    b11 =      1./5.;
  
    b21 =      3./40.;
    b22 =      9./40.;
  
    b31 =     44./45.;
    b32 =    -56./15.;
    b33 =     32./9.;
  
    b41 =  19372./6561.;
    b42 = -25360./2187.;
    b43 =  64448./6561.;
    b44 =   -212./729.;
  
    b51 =   9017./3168.;
    b52 =   -355./33.;
    b53 =  46732./5247.;
    b54 =     49./176.;
    b55 =  -5103./18656.;
  
    b61 =     35./384.;
    b63 =    500./1113.;
    b64 =    125./192.;
    b65 =  -2187./6784.;
    b66 =     11./84.;
    // b62 is 0.0

    // Coefs for error calculation (c^[k] - c[k]).
    // Note that c[k] = b6k, and c^[2]=c[2]=0.0
    dc1 =   5179./57600.  - b61;
    dc3 =   7571./16695.  - b63;
    dc4 =    393./640.    - b64;
    dc5 = -92097./339200. - b65;
    dc6 =    187./2100.   - b66;
    dc7 =      1./40.;
    break;
  case RK547FS:
    /////////////////////////////////////////////////////////////////
    // Coefficients for Dormand & Prince RK5(4)7FS
    a1 = 2./9.;
    a2 = 1./3.;
    a3 = 5./9.;
    a4 = 2./3.;
    // a5 and a6 are 1.0

    b11 =      2./9.;
  
    b21 =      1./12.;
    b22 =      1./4.;
  
    b31 =     55./324.;
    b32 =    -25./108.;
    b33 =     50./81.;
  
    b41 =     83./330.;
    b42 =    -13./22.;
    b43 =     61./66.;
    b44 =      9./110.;
  
    b51 =    -19./28.;
    b52 =      9./4.;
    b53 =      1./7.;
    b54 =    -27./7.;
    b55 =     22./7.;
  
    b61 =     19./200.;
    b63 =      3./5.;
    b64 =   -243./400.;
    b65 =     33./40.;
    b66 =      7./80.;
    // b62 is 0.0

    // Coefs for error calculation (c^[k] - c[k]).
    // Note that c[k] = b6k, and c^[2]=c[2]=0.0
    dc1 =    431./5000.  - b61;
    dc3 =    333./500.   - b63;
    dc4 =  -7857./10000. - b64;
    dc5 =    957./1000.  - b65;
    dc6 =    193./2000.  - b66;
    dc7 =     -1./50.;
    break;
  default:
    throw Oxs_Ext::Error(this,
                 "Oxs_ModLLG_RungeKuttaEvolve::RungeKuttaFehlbergBase54:"
                 " Programming error; Invalid sub-type.");
  }

#ifdef CODE_CHECKS
  // COEFFICIENT CHECKS ////////////////////////////////////////
  // Try to catch some simple typing errors
#define EPS (6*REAL8_EPSILON)
  if(fabs(b11-a1)>EPS ||
     fabs(b21+b22-a2)>EPS ||
     fabs(b31+b32+b33-a3)>EPS ||
     fabs(b41+b42+b43+b44-a4)>EPS ||
     fabs(b51+b52+b53+b54+b55-1.0)>EPS ||
     fabs(b61+b63+b64+b65+b66-1.0)>EPS) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),
                "Coefficient check failed:\n"
                "1: %g\n2: %g\n3: %g\n4: %g\n5: %g\n6: %g",
                static_cast<double>(b11-a1),
                static_cast<double>(b21+b22-a2),
                static_cast<double>(b31+b32+b33-a3),
                static_cast<double>(b41+b42+b43+b44-a4),
                static_cast<double>(b51+b52+b53+b54+b55-1.0),
                static_cast<double>(b61+b63+b64+b65+b66-1.0));
    throw Oxs_Ext::Error(this,buf);
  }
#endif // CODE_CHECKS

  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  REAL8m pE_pt,max_dm_dt,dE_dt,timestep_lower_bound;
  REAL8m dummy_error;
  REAL8m min_m,max_m;

  // Step 1
  AdjustState(stepsize*a1,stepsize*b11,*cstate,current_dm_dt,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,stepsize*a1,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);

  // Steps 2 and 3
  vtmpB = current_dm_dt;
  vtmpB *= b21;
  vtmpB.Accumulate(b22,vtmpA);
  AdjustState(stepsize*a2,stepsize,*cstate,vtmpB,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpB,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpB,
                  &current_dm_dt,pE_pt,stepsize*a2,
                  vtmpB,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);

  // Steps 4 and 5
  vtmpC = current_dm_dt;
  vtmpC *= b31;
  vtmpC.Accumulate(b32,vtmpA);
  vtmpC.Accumulate(b33,vtmpB);
  AdjustState(stepsize*a3,stepsize,*cstate,vtmpC,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpC,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpC,
                  &current_dm_dt,pE_pt,stepsize*a3,
                  vtmpC,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);

  // Steps 6 and 7
  vtmpD = current_dm_dt;
  vtmpD *= b41;
  vtmpD.Accumulate(b42,vtmpA);
  vtmpD.Accumulate(b43,vtmpB);
  vtmpD.Accumulate(b44,vtmpC);
  AdjustState(stepsize*a4,stepsize,*cstate,vtmpD,
              next_state_key.GetWriteReference(),dummy_error);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpD,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpD,
                  &current_dm_dt,pE_pt,stepsize*a4,
                  vtmpD,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // Array holdings: A=dm_dt2   B=dm_dt3   C=dm_dt4   D=dm_dt5

  // Step 8 and 9
  vtmpA *= b52;
  vtmpA.Accumulate(b51,current_dm_dt);
  vtmpA.Accumulate(b53,vtmpB);
  vtmpA.Accumulate(b54,vtmpC);
  vtmpA.Accumulate(b55,vtmpD);
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              next_state_key.GetWriteReference(),dummy_error); // a5=1.0
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&tmpHeff,pE_pt);
  Calculate_dm_dt(next_state_key.GetReadReference(),vtmpA,
                  &current_dm_dt,pE_pt,stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  // Array holdings: A=dm_dt6   B=dm_dt3   C=dm_dt4   D=dm_dt5

  // Step 10
  UINT4m i;
  const UINT4m size = cstate->mesh->Size();
  for(i=0;i<size;i++) {
    ThreeVector dm_dt3 = vtmpB[i];
    ThreeVector dm_dt6 = vtmpA[i];
    vtmpA[i] = b63 * dm_dt3  +  b66 * dm_dt6;  // k6(3,6)
    vtmpB[i] = dc3 * dm_dt3  +  dc6 * dm_dt6;  // dD(3,6)
  }
  // Array holdings: A=k6(3,6)   B=dD(3,6)   C=dm_dt4   D=dm_dt5

  // Step 11
  vtmpA.Accumulate(b61,current_dm_dt);   // Note: b62=0.0
  vtmpA.Accumulate(b64,vtmpC);
  vtmpA.Accumulate(b65,vtmpD);
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              next_state_key.GetWriteReference(),norm_error); // a6=1.0
  const Oxs_SimState& endstate
    = next_state_key.GetReadReference(); // Candidate next state
  GetEnergyDensity(endstate,temp_energy,&mxH_output.cache.value,
                   &tmpHeff,pE_pt);
  mxH_output.cache.state_id=endstate.Id();
  Calculate_dm_dt(endstate,mxH_output.cache.value,
                  &current_dm_dt,pE_pt,stepsize,
                  vtmpA,max_dm_dt,dE_dt,timestep_lower_bound,
                  min_m,max_m);
  REAL8m mdev = ComputeMaxNormError(min_m,max_m);

  if(!endstate.AddDerivedData("Timestep lower bound",
                              timestep_lower_bound) ||
     !endstate.AddDerivedData("Max dm/dt",max_dm_dt) ||
     !endstate.AddDerivedData("Min |m|",min_m) ||
     !endstate.AddDerivedData("Max |m|",max_m) ||
     !endstate.AddDerivedData("Max norm error",mdev) ||
     !endstate.AddDerivedData("pE/pt",pE_pt) ||
     !endstate.AddDerivedData("dE/dt",dE_dt)) {
    throw Oxs_Ext::Error(this,
          "Oxs_ModLLG_RungeKuttaEvolve::RungeKuttaFehlbergBase54:"
          " Programming error; data cache already set.");
  }
  // Array holdings: A=dm_dt7   B=dD(3,6)   C=dm_dt4   D=dm_dt5

  // Step 12
  vtmpB.Accumulate(dc1,current_dm_dt);
  vtmpB.Accumulate(dc4,vtmpC);
  vtmpB.Accumulate(dc5,vtmpD);
  vtmpB.Accumulate(dc7,vtmpA);
  // Array holdings: A=dm_dt7   B=dD   C=dm_dt4   D=dm_dt5

  // next_state holds candidate next state, normalized and
  // with proper time field settings; see Step 11.  Note that
  // Step 11 also set norm_error.

  // Error estimate is max|m2a-m2b| = h*max|dD|
  REAL8m max_dD_sq=0.0;
  for(i=0;i<size;i++) {
    REAL8m magsq = vtmpB[i].MagSq();
    if(magsq>max_dD_sq) max_dD_sq = magsq;
  }
  error_estimate = stepsize * sqrt(max_dD_sq);
  global_error_order = 4.0;
  new_energy_and_dmdt_computed = 1;
}

void Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaFehlbergStep54
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& current_dm_dt,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,REAL8m& global_error_order,
 REAL8m& norm_error,
 BOOL& new_energy_and_dmdt_computed)
{
  RungeKuttaFehlbergBase54(RK547FC,stepsize,
     current_state_key,current_dm_dt,next_state_key,
     error_estimate,global_error_order,norm_error,
     new_energy_and_dmdt_computed);
}

void Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaFehlbergStep54M
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& current_dm_dt,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,REAL8m& global_error_order,
 REAL8m& norm_error,
 BOOL& new_energy_and_dmdt_computed)
{
  RungeKuttaFehlbergBase54(RK547FM,stepsize,
     current_state_key,current_dm_dt,next_state_key,
     error_estimate,global_error_order,norm_error,
     new_energy_and_dmdt_computed);
}

void Oxs_ModLLG_RungeKuttaEvolve::TakeRungeKuttaFehlbergStep54S
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& current_dm_dt,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,REAL8m& global_error_order,
 REAL8m& norm_error,
 BOOL& new_energy_and_dmdt_computed)
{
  RungeKuttaFehlbergBase54(RK547FS,stepsize,
     current_state_key,current_dm_dt,next_state_key,
     error_estimate,global_error_order,norm_error,
     new_energy_and_dmdt_computed);
}

REAL8m Oxs_ModLLG_RungeKuttaEvolve::MaxDiff
(const Oxs_MeshValue<ThreeVector>& vecA,
 const Oxs_MeshValue<ThreeVector>& vecB)
{
  UINT4m size = vecA.Size();
  if(vecB.Size()!=size) {
    throw Oxs_Ext::Error(this,
                 "Oxs_ModLLG_RungeKuttaEvolve::MaxDiff:"
                 " Import MeshValues incompatible (different lengths).");
  }
  REAL8m max_magsq = 0.0;
  for(UINT4m i=0;i<size;i++) {
    ThreeVector vtemp = vecB[i] - vecA[i];
    REAL8m magsq = vtemp.MagSq();
    if(magsq>max_magsq) max_magsq = magsq;
  }
  return sqrt(max_magsq);
}

void Oxs_ModLLG_RungeKuttaEvolve::AdjustStepHeadroom(INT4m step_reject)
{ // step_reject should be 0 or 1, reflecting whether the current
  // step was rejected or not.  This routine updates reject_ratio
  // and adjusts step_headroom appropriately.

  // First adjust reject_ratio, weighing mostly the last
  // thirty or so results.
  reject_ratio = (31*reject_ratio + step_reject)/32.;

  // Adjust step_headroom
  if(reject_ratio>reject_goal && step_reject>0) {
    // Reject ratio too high and getting worse
    step_headroom *= 0.925;
  }
  if(reject_ratio<reject_goal && step_reject<1) {
    // Reject ratio too small and getting smaller
    step_headroom *= 1.075;
  }

  if(step_headroom>max_step_headroom) step_headroom=max_step_headroom;
  if(step_headroom<min_step_headroom) step_headroom=min_step_headroom;
}

BOOL
Oxs_ModLLG_RungeKuttaEvolve::Step(const Oxs_TimeDriver* driver,
                      Oxs_ConstKey<Oxs_SimState> current_state_key,
                      const Oxs_DriverStepInfo& step_info,
                      Oxs_Key<Oxs_SimState>& next_state_key)
{
  const REAL8m bad_energy_cut_ratio = 0.75;
  const REAL8m bad_energy_step_increase = 1.3;

  const REAL8m previous_next_timestep = next_timestep;

  const Oxs_SimState& cstate = current_state_key.GetReadReference();

  CheckCache(cstate);

  // Note if start_dm is being used
  BOOL start_dm_active=0;
  if(next_timestep<=0.0 ||
     (cstate.stage_iteration_count<1
      && step_info.current_attempt_count==0)) {
    if(cstate.stage_number==0
       || stage_init_step_control == SISC_START_DM) {
      start_dm_active = 1;
    } else if(stage_init_step_control == SISC_CONTINUOUS) {
      start_dm_active = 0;  // Safety
    } else if(stage_init_step_control == SISC_AUTO) {
      // Automatic detection based on energy values across
      // stage boundary.  For now just assume true:
      start_dm_active = 0;
    } else {
      throw Oxs_Ext::Error(this,
           "Oxs_ModLLG_RungeKuttaEvolve::Step; Programming error:"
           " unrecognized stage_init_step_control value");
    }
  }

  // Negotiate timestep, and also initialize both next_state and
  // temp_state structures.
  Oxs_SimState* work_state = &(next_state_key.GetWriteReference());
  BOOL force_step=0,driver_set_step=0;
  NegotiateTimeStep(driver,cstate,*work_state,next_timestep,
                    start_dm_active,force_step,driver_set_step);
  REAL8m stepsize = work_state->last_timestep;
  work_state=NULL; // Safety: disable pointer

  // Step
  REAL8m error_estimate,norm_error;
  REAL8m global_error_order;
  BOOL new_energy_and_dmdt_computed;
  BOOL reject_step=0;
  (this->*rkstep_ptr)(stepsize,current_state_key,
                      dm_dt_output.cache.value,
                      next_state_key,
                      error_estimate,global_error_order,norm_error,
                      new_energy_and_dmdt_computed);

  // Resize?
  if(RT_NONE != resize_norm && fabs(norm_error)>allowed_mnorm_deviation) {
    if(RT_NORM != resize_norm && !new_energy_and_dmdt_computed) {
      REAL8m new_pE_pt,new_max_dm_dt,new_dE_dt,new_timestep_lower_bound;
      REAL8m new_min_m,new_max_m;
      GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                       &vtmpA,&tmpHeff,new_pE_pt);
      Calculate_dm_dt(next_state_key.GetReadReference(),vtmpA,
                      &dm_dt_output.cache.value,new_pE_pt,stepsize,
                      vtmpA,new_max_dm_dt,
                      new_dE_dt,new_timestep_lower_bound,
                      new_min_m,new_max_m);
    }
    work_state = &(next_state_key.GetWriteReference());
    work_state->ClearDerivedData();
    ResizeSpins(resize_norm,
                cstate.spin,work_state->spin,
                Heff,tmpHeff,
                dm_dt_output.cache.value,vtmpA);
    work_state=NULL; // Safety: disable pointer
    norm_error = 0.0;  // Lie just a little
    new_energy_and_dmdt_computed = 0;
  }

  const Oxs_SimState& nstate = next_state_key.GetReadReference();

  REAL8m max_dm_dt;
  cstate.GetDerivedData("Max dm/dt",max_dm_dt);
  REAL8m reference_stepsize = stepsize;
  if(driver_set_step) reference_stepsize = previous_next_timestep;
  BOOL good_step = CheckError(global_error_order,error_estimate,
                              stepsize,reference_stepsize,
                              max_dm_dt,next_timestep);
  /// Note: Might want to use average or larger of max_dm_dt
  /// and new_max_dm_dt (computed below.)

  if(!good_step && !force_step) {
    // Bad step; The only justfication to do energy and dm_dt
    // computation would be to get an energy-based stepsize
    // adjustment estimate, which we will need to try if 
    // next_timestep is larger than cut applied by energy
    // rejection code (i.e., bad_energy_cut_ratio).
    if(next_timestep<=stepsize*bad_energy_cut_ratio) {
      AdjustStepHeadroom(1);
      return 0; // Don't bother with energy calculation
    }
    reject_step=1; // Otherwise, mark step rejected and see what energy
    /// info suggests for next stepsize
  }

  if(start_dm_active && !force_step) {
    // Check that no spin has moved by more than start_dm
    REAL8m diff = MaxDiff(cstate.spin,nstate.spin);
    if(diff>start_dm) {
      next_timestep = step_headroom * stepsize * (start_dm/diff);
      if(next_timestep<=stepsize*bad_energy_cut_ratio) {
        AdjustStepHeadroom(1);
        return 0; // Don't bother with energy calculation
      }
      reject_step=1; // Otherwise, mark step rejected and see what energy
      /// info suggests for next stepsize
    }
  }

#ifdef OLDE_CODE
  if(fabs(norm_error)>0.0005) {
    fprintf(stderr,
            "Iteration %u passed error check; norm_error=%8.5f\n",
            nstate.iteration_count,static_cast<double>(norm_error));
  } /**/
#endif // OLDE_CODE

  // Energy timestep control:
  //   The relationship between energy error and stepsize appears to be
  // highly nonlinear, so that estimating appropriate stepsize from energy
  // increase if difficult.  Perhaps it is possible to include energy
  // interpolation into RK step routines, but for the present we just
  // reduce the step by a fixed ratio if we detect energy increase beyond
  // that which can be attributed to numerical errors.  Of course, this
  // doesn't take into account the expected energy decrease (which depends
  // on the damping ratio alpha), which is another reason to try to build
  // it into the high order RK step routines.
  REAL8m pE_pt,new_pE_pt=0.;
  cstate.GetDerivedData("pE/pt",pE_pt);
  if(new_energy_and_dmdt_computed) {
    nstate.GetDerivedData("pE/pt",new_pE_pt);
  } else {
    GetEnergyDensity(nstate,temp_energy,
                     &mxH_output.cache.value,
                     &tmpHeff,new_pE_pt);
    mxH_output.cache.state_id=nstate.Id();
    if(!nstate.AddDerivedData("pE/pt",new_pE_pt)) {
      throw Oxs_Ext::Error(this,
           "Oxs_ModLLG_RungeKuttaEvolve::Step:"
           " Programming error; data cache (pE/pt) already set.");
    }
  }

  REAL8m dE=0.0;
  REAL8m var_dE=0.0;
  REAL8m total_E=0.0;
  UINT4m i;
  const UINT4m size = nstate.mesh->Size();
  for(i=0;i<size;++i) {
    REAL8m vol = nstate.mesh->Volume(i);
    REAL8m e = energy[i];
    total_E += e * vol;
    REAL8m new_e = temp_energy[i];
    dE += (new_e - e) * vol;
    var_dE += (new_e*new_e + e*e)*vol*vol;
  }
  if(!nstate.AddDerivedData("Delta E",dE)) {
    throw Oxs_Ext::Error(this,
         "Oxs_ModLLG_RungeKuttaEvolve::Step:"
         " Programming error; data cache (Delta E) already set.");
  }


  if(expected_energy_precision>=0.) {
    var_dE *= expected_energy_precision * expected_energy_precision;
    /// Variance, assuming error in each energy[i] term is independent,
    /// uniformly distributed, 0-mean, with range
    ///        +/- expected_energy_precision*energy[i].
    /// It would probably be better to get an error estimate directly
    /// from each energy term.
    REAL8m E_numerror = OC_MAX(fabs(total_E)*expected_energy_precision,
                               2*sqrt(var_dE));
    REAL8m pE_pt_max = OC_MAX(pE_pt,new_pE_pt); // Might want to
    /// change this from a constant to a linear function in timestep.

    REAL8m reject_dE = 2*E_numerror + pE_pt_max * stepsize;
    if(dE>reject_dE) {
      REAL8m teststep = bad_energy_cut_ratio*stepsize;
      if(teststep<next_timestep) {
        next_timestep=teststep;
        max_step_increase = bad_energy_step_increase;
        // Damp the enthusiasm of the RK stepper routine for
        // stepsize growth, for a step or two.
      }
      if(!force_step) reject_step=1;
    }
  }

  if(!force_step && reject_step) {
    AdjustStepHeadroom(1);
    return 0;
  }

  // Otherwise, we are accepting the new step.

  // Calculate dm_dt at new point, and fill in cache.
  if(new_energy_and_dmdt_computed) {
    dm_dt_output.cache.value.Swap(vtmpA);
  } else {
    REAL8m new_max_dm_dt,new_dE_dt,new_timestep_lower_bound;
    REAL8m new_min_m,new_max_m;
    if(mxH_output.cache.state_id != nstate.Id()) { // Safety
    throw Oxs_Ext::Error(this,
         "Oxs_ModLLG_RungeKuttaEvolve::Step:"
         " Programming error; mxH_output cache improperly filled.");
    }
    Calculate_dm_dt(nstate,
                    mxH_output.cache.value,&dm_dt_output.cache.value,
                    new_pE_pt,stepsize,vtmpA,new_max_dm_dt,
                    new_dE_dt,new_timestep_lower_bound,
                    new_min_m,new_max_m);
    dm_dt_output.cache.value.Swap(vtmpA);
    REAL8m new_mdev = ComputeMaxNormError(new_min_m,new_max_m);
    if(!nstate.AddDerivedData("Timestep lower bound",
                              new_timestep_lower_bound) ||
       !nstate.AddDerivedData("Max dm/dt",new_max_dm_dt) ||
       !nstate.AddDerivedData("Min |m|",new_min_m) ||
       !nstate.AddDerivedData("Max |m|",new_max_m) ||
       !nstate.AddDerivedData("Max norm error",new_mdev) ||
       !nstate.AddDerivedData("dE/dt",new_dE_dt)) {
      throw Oxs_Ext::Error(this,
                           "Oxs_ModLLG_RungeKuttaEvolve::Step:"
                           " Programming error; data cache already set.");
    }
  }
  dm_dt_output.cache.state_id = nstate.Id();

  energy_state_id = nstate.Id();
  energy.Swap(temp_energy);
  Heff.Swap(tmpHeff);

  if(RT_NONE != resize_differential) {
    integrated_norm_error.Swap(new_integrated_norm_error);
    last_norm_error.Swap(new_norm_error);
  }

  AdjustStepHeadroom(0);
  if(!force_step && max_step_increase<max_step_increase_limit) {
    max_step_increase *= max_step_increase_adj_ratio;
  }
  if(max_step_increase>max_step_increase_limit) {
    max_step_increase = max_step_increase_limit;
  }

  return 1; // Accept step
}

void Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs(const Oxs_SimState& state)
{ // This routine fills all the Oxs_ModLLG_RungeKuttaEvolve Oxs_ScalarOutput's to
  // the appropriate value based on the import "state", and any of
  // Oxs_VectorOutput's that have CacheRequest enabled are filled.
  // It also makes sure all the expected WOO objects in state are
  // filled.
  max_dm_dt_output.cache.state_id
    = min_m_output.cache.state_id
    = max_m_output.cache.state_id
    = max_mdev_output.cache.state_id
    = dE_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = delta_E_output.cache.state_id
    = 0;  // Mark change in progress

  REAL8m dummy_value;
  if(!state.GetDerivedData("Max dm/dt",max_dm_dt_output.cache.value) ||
     !state.GetDerivedData("Min |m|",min_m_output.cache.value) ||
     !state.GetDerivedData("Max |m|",max_m_output.cache.value) ||
     !state.GetDerivedData("Max norm error",max_mdev_output.cache.value) ||
     !state.GetDerivedData("dE/dt",dE_dt_output.cache.value) ||
     !state.GetDerivedData("Delta E",delta_E_output.cache.value) ||
     !state.GetDerivedData("pE/pt",dummy_value) ||
     !state.GetDerivedData("Timestep lower bound",dummy_value) ||
     (dm_dt_output.GetCacheRequestCount()>0
      && dm_dt_output.cache.state_id != state.Id()) ||
     (mxH_output.GetCacheRequestCount()>0
      && mxH_output.cache.state_id != state.Id())) {

    // Missing at least some data, so calculate from scratch

    // Calculate H and mxH outputs
    Oxs_MeshValue<ThreeVector>& mxH = mxH_output.cache.value;
    REAL8m pE_pt;
    GetEnergyDensity(state,energy,&mxH,&Heff,pE_pt);
    energy_state_id=state.Id();
    mxH_output.cache.state_id=state.Id();
    if(!state.GetDerivedData("pE/pt",dummy_value)) {
      state.AddDerivedData("pE/pt",pE_pt);
    }

    // Calculate dm/dt, Max dm/dt and dE/dt
    Oxs_MeshValue<ThreeVector>& dm_dt
      = dm_dt_output.cache.value;
    dm_dt_output.cache.state_id=0;
    REAL8m timestep_lower_bound;
    Calculate_dm_dt(state,mxH,NULL,pE_pt,0.0,dm_dt,
                    max_dm_dt_output.cache.value,
                    dE_dt_output.cache.value,timestep_lower_bound,
                    min_m_output.cache.value,max_m_output.cache.value);
    dm_dt_output.cache.state_id=state.Id();
    if(!state.GetDerivedData("Max dm/dt",dummy_value)) {
      state.AddDerivedData("Max dm/dt",max_dm_dt_output.cache.value);
    }
    if(!state.GetDerivedData("Min |m|",dummy_value)) {
      state.AddDerivedData("Min |m|",min_m_output.cache.value);
    }
    if(!state.GetDerivedData("Max |m|",dummy_value)) {
      state.AddDerivedData("Max |m|",max_m_output.cache.value);
    }
    if(!state.GetDerivedData("Max norm error",dummy_value)) {
      state.AddDerivedData("Max norm error",
                           ComputeMaxNormError(min_m_output.cache.value,
                                               max_m_output.cache.value));
    }
    if(!state.GetDerivedData("dE/dt",dummy_value)) {
      state.AddDerivedData("dE/dt",dE_dt_output.cache.value);
    }
    if(!state.GetDerivedData("Timestep lower bound",dummy_value)) {
      state.AddDerivedData("Timestep lower bound",
                           timestep_lower_bound);
    }

    if(!state.GetDerivedData("Delta E",dummy_value)) {
      if(state.previous_state_id!=0 && state.stage_iteration_count>0) {
        // Strictly speaking, we should be able to create dE for
        // stage_iteration_count==0 for stages>0, but as a practical
        // matter we can't at present.  Should give this more thought.
        // -mjd, 27-July-2001
        throw Oxs_Ext::Error(this,
           "Oxs_ModLLG_RungeKuttaEvolve::UpdateDerivedOutputs:"
           " Can't derive Delta E from single state.");
      }
      state.AddDerivedData("Delta E",0.0);
      dummy_value = 0.;
    }
    delta_E_output.cache.value=dummy_value;
  }

  max_dm_dt_output.cache.value*=(180e-9/PI);
  /// Convert from radians/second to deg/ns

  max_dm_dt_output.cache.state_id
    = min_m_output.cache.state_id
    = max_m_output.cache.state_id
    = max_mdev_output.cache.state_id
    = dE_dt_output.cache.state_id
    = delta_E_output.cache.state_id
    = state.Id();
}
