/* FILE: semianalyticevolve.cc                 -*-Mode: c++-*-
 *
 * Concrete evolver class, using semi-analytic steps.
 *
 */

#include <float.h>
#include <string>

#include "nb.h"
#include "director.h"
#include "timedriver.h"
#include "simstate.h"
#include "semianalyticevolve.h"
#include "key.h"
#include "energy.h"             // Needed to make MSVC++ 5 happy

OC_USE_STRING;

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_SemiAnalyticEvolve);

/* End includes */

//#define DEEP_DEBUG
#ifdef DEEP_DEBUG
#if 0
# define CHECK_INDEX 1096
# define CHECK_START  2775
# define CHECK_STOP   2779
#endif
#if 0
# define CHECK_INDEX 1840
# define CHECK_START  1870
# define CHECK_STOP   1876
#endif
#if 0
# define CHECK_INDEX 1794
# define CHECK_START  3302
# define CHECK_STOP   3308
#endif
#if 0
# define CHECK_INDEX 1794
# define CHECK_START  602
# define CHECK_STOP   608
#endif
#if 1
# define CHECK_INDEX 1794
# define CHECK_START  3975
# define CHECK_STOP   4025
#endif
#if 0
# define CHECK_INDEX  1348
# define CHECK_START  270
# define CHECK_STOP   300
#endif
#if 0
# define CHECK_INDEX  1348
# define CHECK_START  110
# define CHECK_STOP   150
#endif
#if 0
# define CHECK_INDEX  715
# define CHECK_START  1
# define CHECK_STOP   5
#endif
#endif // DEEP_DEBUG

// Constructor
Oxs_SemiAnalyticEvolve::Oxs_SemiAnalyticEvolve
( const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  const char* argstr)   // MIF input block parameters
  : Oxs_TimeEvolver(name,newdtr,argstr),
    mesh_id(0),
    max_step_decrease(0.03125), max_step_increase_limit(4.0),
    max_step_increase_adj_ratio(1.9),
    reject_goal(0.05), reject_ratio(0.05),
    debug_step_error(0),debug_step_error_start(-1),debug_step_error_stop(-1),
    debug_step_error_iteration_count(-2),
    debug_step_error_iteration_subcount(-2),
    energy_state_id(0),next_timestep(0.),old_timestep_cached(0.),
    mxH_old_cached_is_valid(0),sastep_ptr(NULL)
{
  // Process arguments
  min_timestep=GetRealInitValue("min_timestep",0.);
  max_timestep=GetRealInitValue("max_timestep",1e-10);
  if(max_timestep<=0.0) {
    char buf[4096];
    Oc_Snprintf(buf,sizeof(buf),
                "Invalid parameter value:"
                " Specified max time step is %g (should be >0.)",
                static_cast<double>(max_timestep));
    throw Oxs_Ext::Error(this,buf);
  }

  allowed_error_rate = GetRealInitValue("error_rate",1.0);
  if(allowed_error_rate>0.0) {
    allowed_error_rate *= PI*1e9/180.; // Convert from deg/ns to rad/s
  }
  allowed_absolute_step_error
    = GetRealInitValue("absolute_step_error",0.2);
  if(allowed_absolute_step_error>0.0) {
    allowed_absolute_step_error *= PI/180.; // Convert from deg to rad
  }
  allowed_relative_step_error
    = GetRealInitValue("relative_step_error",0.01);

  expected_energy_precision =
    GetRealInitValue("energy_precision",1e-10);

  reject_goal = GetRealInitValue("reject_goal",0.05);
  if(reject_goal<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"reject_goal\" value must be non-negative.");
  }

  min_step_headroom = GetRealInitValue("min_step_headroom",0.33);
  if(min_step_headroom<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"min_step_headroom\" value must be bigger than 0.");
  }

  max_step_headroom = GetRealInitValue("max_step_headroom",0.95);
  if(max_step_headroom<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"max_step_headroom\" value must be bigger than 0.");
  }

  if(min_step_headroom>max_step_headroom) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"min_step_headroom\" value must not be larger than"
         " \"max_step_headroom\".");
  }


  if(HasInitValue("alpha")) {
    OXS_GET_INIT_EXT_OBJECT("alpha",Oxs_ScalarField,alpha_init);
  } else {
    alpha_init.SetAsOwner(dynamic_cast<Oxs_ScalarField *>
                          (MakeNew("Oxs_UniformScalarField",director,
                                   "value 0.5")));
  }

  // User may specify either gamma_G (Gilbert) or
  // gamma_LL (Landau-Lifshitz).
  gamma_style = GS_INVALID;
  if(HasInitValue("gamma_G") && HasInitValue("gamma_LL")) {
    throw Oxs_Ext::Error(this,"Invalid Specify block; "
                         "both gamma_G and gamma_LL specified.");
  } else if(HasInitValue("gamma_G")) {
    gamma_style = GS_G;
    OXS_GET_INIT_EXT_OBJECT("gamma_G",Oxs_ScalarField,gamma_init);
  } else if(HasInitValue("gamma_LL")) {
    gamma_style = GS_LL;
    OXS_GET_INIT_EXT_OBJECT("gamma_LL",Oxs_ScalarField,gamma_init);
  } else {
    gamma_style = GS_G;
    gamma_init.SetAsOwner(dynamic_cast<Oxs_ScalarField *>
                          (MakeNew("Oxs_UniformScalarField",director,
                                   "value -2.211e5")));
  }
  allow_signed_gamma = GetIntInitValue("allow_signed_gamma",0);
  do_precess = GetIntInitValue("do_precess",1);

  start_dm = GetRealInitValue("start_dm",0.01);
  if(start_dm<0.) {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
         " \"start_dm\" value must be nonnegative.");
  }
  start_dm *= PI/180.; // Convert from deg to rad

  stage_init_step_control = SISC_AUTO;  // Safety
  String stage_start = GetStringInitValue("stage_start","auto");
  Oxs_ToLower(stage_start);
  if(stage_start.compare("start_dm")==0) {
    stage_init_step_control = SISC_START_DM;
  } else if(stage_start.compare("continuous")==0) {
    stage_init_step_control = SISC_CONTINUOUS;
  } else if(stage_start.compare("auto")==0
            || stage_start.compare("automatic")==0) {
    stage_init_step_control = SISC_AUTO;
  } else {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
                         " \"stage_start\" value must be one of"
                         " start_dm, continuous, or auto.");
  }

  String method = GetStringInitValue("method","saac2");
  Oxs_ToLower(method); // Do case insensitive match
  if(method.compare("sa1")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticStep1;
  } else if(method.compare("sa2")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticStep2;
  } else if(method.compare("saac1")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep1;
  } else if(method.compare("saac2")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2;
  } else if(method.compare("saac2b")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2b;
  } else if(method.compare("saac2c")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2c;
  } else if(method.compare("sat2")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticTorqueStep2;
  } else if(method.compare("sat2b")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeSemiAnalyticTorqueStep2b;
  } else if(method.compare("rkf54m")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeRungeKuttaFehlbergStep54M;
  } else if(method.compare("rkf54s")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeRungeKuttaFehlbergStep54S;
  } else if(method.compare("rkf54")==0) {
    sastep_ptr = &Oxs_SemiAnalyticEvolve::TakeRungeKuttaFehlbergStep54;
  } else {
    throw Oxs_Ext::Error(this,"Invalid initialization detected:"
                         " \"method\" value must be one of"
                         " sa1, sa2, saac1, saac2,"
                         " saac2b, saac2c, sat2, sat2b,"
                         " rkf54, rkf54m, rkf54s");
  }

  // Setup outputs
  max_dm_dt_output.Setup(this,InstanceName(),"Max dm/dt","deg/ns",0,
     &Oxs_SemiAnalyticEvolve::UpdateDerivedOutputs);
  dE_dt_output.Setup(this,InstanceName(),"dE/dt","J/s",0,
     &Oxs_SemiAnalyticEvolve::UpdateDerivedOutputs);
  delta_E_output.Setup(this,InstanceName(),"Delta E","J",0,
     &Oxs_SemiAnalyticEvolve::UpdateDerivedOutputs);
  mxH_output.Setup(this,InstanceName(),"mxH","A/m",1,
     &Oxs_SemiAnalyticEvolve::UpdateDerivedOutputs);

  max_dm_dt_output.Register(director,-5);
  dE_dt_output.Register(director,-5);
  delta_E_output.Register(director,-5);
  mxH_output.Register(director,-5);

  // mxH output cache is used for intermediate storage,
  // so enable caching.
  mxH_output.CacheRequestIncrement(1);

  // Debug output
  if(HasInitValue("debug_step_error_basename")) {
    debug_step_error = 1;
    debug_step_error_basename
      = GetStringInitValue("debug_step_error_basename");
    debug_step_error_start = GetIntInitValue("debug_step_error_start");
    debug_step_error_stop  = GetIntInitValue("debug_step_error_stop");
  } else {
    debug_step_error = 0;  // Safety
  }

  VerifyAllInitArgsUsed();

  // Reserve space for temp_state; see Step() method below
  director->ReserveSimulationStateRequest(1);
}

BOOL Oxs_SemiAnalyticEvolve::Init()
{
  Oxs_TimeEvolver::Init();

  // Free scratch space allocated by previous problem (if any)
  vtmpA.Release();   vtmpB.Release();
  vtmpC.Release();   vtmpD.Release();
  vtmpE.Release();   vtmpF.Release();
  rkf54_data.Release();

  // Free memory used by LLG gamma and alpha
  mesh_id = 0;
  alpha.Release();
  gamma.Release();

  max_step_increase = max_step_increase_limit;

  // Setup step_headroom and reject_ratio
  step_headroom = max_step_headroom;
  reject_ratio = reject_goal;

  // Free cache space and initialize cache settings
  energy.Release();  H_cached.Release();  mxH_old_cached.Release();
  energy_state_id=0;      // Mark as invalid state
  next_timestep=0.;       // Dummy value
  old_timestep_cached=0.; // Dummy value
  mxH_old_cached_is_valid = 0;

  return 1;
}


Oxs_SemiAnalyticEvolve::~Oxs_SemiAnalyticEvolve()
{}

void Oxs_SemiAnalyticEvolve::UpdateMeshArrays(const Oxs_Mesh* mesh)
{
  mesh_id = 0; // Mark update in progress
  const UINT4m size = mesh->Size();
  UINT4m i;

  alpha_init->FillMeshValue(mesh,alpha);
  gamma_init->FillMeshValue(mesh,gamma);

  if(gamma_style == GS_G) { // Convert to LL form
    for(i=0;i<size;++i) {
      REAL8m cell_alpha = alpha[i];
      gamma[i] /= (1+cell_alpha*cell_alpha);
    }
  }
  if(!allow_signed_gamma) {
    for(i=0;i<size;++i) gamma[i] = -1*fabs(gamma[i]);
  }

  mesh_id = mesh->Id();
}

void Oxs_SemiAnalyticEvolve::Compute_dm_dt
(const ThreeVector& m,
 const ThreeVector& mxH,
 const REAL8m import_gamma,
 const REAL8m import_alpha,
 ThreeVector& dm_dt) const
{
  REAL8m coef1 = import_gamma;
  REAL8m coef2 = -1*import_alpha*coef1;
  if(!do_precess) coef1 = 0.0;

  ThreeVector scratch = mxH;
  dm_dt = mxH;
  dm_dt *= coef1;   // coef1 == 0 if do_precess if false
  scratch ^= m; // ((mxH)xm)
  scratch *= coef2;  // -alpha.gamma((mxH)xm) = alpha.gamma(mx(mxH))
  dm_dt += scratch;
}

void Oxs_SemiAnalyticEvolve::ComputeMiscValues
(const Oxs_SimState& state,
 const Oxs_MeshValue<ThreeVector>& mxH,
 REAL8m pE_pt,
 REAL8m& max_dm_dt,REAL8m& dE_dt,REAL8m& min_tstep) const
{
  // NOTE: This code assumes mxH == 0 at fixed spins.
  const Oxs_Mesh* mesh = state.mesh;
  const Oxs_MeshValue<REAL8m>& Ms = *(state.Ms);
  const UINT4m size = mesh->Size(); // Assume import data are compatible

#ifndef NDEBUG
  // Consistency checks
  if(mesh_id != mesh->Id()
     || !gamma.CheckMesh(mesh)
     || !alpha.CheckMesh(mesh)) {
    throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::ComputeMiscValues:"
           " mesh_id, gamma or alpha not properly initialized.");
  }
#endif // NDEBUG

  REAL8m max_dm_dt_sq = 0.0;
  REAL8m dE_dt_sum=0.0;
  for(UINT4m i=0;i<size;i++) {
    if(Ms[i]==0) continue;

    REAL8m mxH_sq = mxH[i].MagSq();
    REAL8m coef1 = gamma[i];
    REAL8m coef2 = -1*alpha[i]*coef1;
    if(!do_precess) coef1 = 0.0;
    REAL8m dm_dt_sq =  mxH_sq * (coef1*coef1 + coef2*coef2);

    if(dm_dt_sq>max_dm_dt_sq) {
      max_dm_dt_sq = dm_dt_sq;
    }

    dE_dt_sum += mxH_sq * Ms[i] * mesh->Volume(i) * coef2;
  }
  max_dm_dt = sqrt(max_dm_dt_sq);

  dE_dt = -1 * MU0 * dE_dt_sum + pE_pt;
  /// The first term is (partial E/partial M)*dM/dt, the
  /// second term is (partial E/partial t)*dt/dt.  Note that,
  /// provided Ms_[i]>=0, that by constructions dE_dt_sum above
  /// is always non-negative, so dE_dt can only be made positive
  /// by positive pE_pt.

  // Get bound on smallest stepsize that would actually
  // change spin new_max_dm_dt_index:
  min_tstep = DBL_MAX/64.;
  if(max_dm_dt>1 || REAL8_EPSILON<min_tstep*max_dm_dt) {
    min_tstep = REAL8_EPSILON/max_dm_dt;
    // A timestep of size min_tstep will be hopelessly lost
    // in roundoff error.  So increase a bit, based on an empirical
    // fudge factor.  This fudge factor can be tested by running a
    // problem with start_dm = 0.  If the evolver can't climb its
    // way out of the stepsize=0 hole, then this fudge factor is too
    // small.  So far, the most challenging examples have been
    // problems with small cells with nearly aligned spins, e.g., in
    // a remanent state with an external field is applied at t=0.
    // Damping ratio doesn't seem to have much effect, either way.
    min_tstep *= 64;
  } else {
    // Degenerate case: max_dm_dt must be exactly or very nearly
    // zero.  Punt.
    min_tstep = 1.0;
  }

  return;
}

void Oxs_SemiAnalyticEvolve::CheckCache(const Oxs_SimState& cstate)
{
  // Pull cached values out from cstate.
  // If cstate.Id() == energy_state_id, then cstate has been run
  // through either this method or UpdateDerivedOutputs.  Either
  // way, all derived state data should be stored in cstate,
  // except currently the "energy" mesh value array, which is
  // stored independently inside *this.  Eventually that should
  // probably be moved in some fashion into cstate too.
  if(energy_state_id != cstate.Id()) {
    // cached data out-of-date
    UpdateDerivedOutputs(cstate);
  }
  BOOL cache_good = 1;
  REAL8m max_dm_dt,dE_dt,delta_E,pE_pt;
  REAL8m timestep_lower_bound;  // Smallest timestep that can actually
  /// change spin with max_dm_dt (due to REAL8_EPSILON restrictions).
  /// The next timestep is based on the error from the last step.  If
  /// there is no last step (either because this is the first step,
  /// or because the last state handled by this routine is different
  /// from the incoming current_state), then timestep is calculated
  /// so that max_dm_dt * timestep = start_dm.

  cache_good &= cstate.GetDerivedData("Max dm/dt",max_dm_dt);
  cache_good &= cstate.GetDerivedData("dE/dt",dE_dt);
  cache_good &= cstate.GetDerivedData("Delta E",delta_E);
  cache_good &= cstate.GetDerivedData("pE/pt",pE_pt);
  cache_good &= cstate.GetDerivedData("Timestep lower bound",
                                      timestep_lower_bound);
  cache_good &= (energy_state_id == cstate.Id());

  if(!cache_good) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::CheckCache: Invalid data cache.");
  }
}

void
Oxs_SemiAnalyticEvolve::ZeroFixedSpinFields
(Oxs_MeshValue<ThreeVector>* mxH,
 Oxs_MeshValue<ThreeVector>* H) const
{
  const UINT4m fixed_count = GetFixedSpinCount();
  for(UINT4m j=0;j<fixed_count;j++) {
    UINT4m index = GetFixedSpin(j);
    if(mxH) (*mxH)[index].Set(0.,0.,0.);
    if(H)     (*H)[index].Set(0.,0.,0.);
  }
}

void
Oxs_SemiAnalyticEvolve::AdjustState
(REAL8m hstep,
 REAL8m mstep,
 const Oxs_SimState& old_state,
 const Oxs_MeshValue<ThreeVector>& H,
 Oxs_SimState& new_state) const
{
  // Advance spins about field H using LLG dynamics for the timestep
  // mstep (in seconds).  For details, see NOTES III, 25-Nov-2002, p15.
  //
  // Export new state has time index from old_state + hstep,
  // and spins from old state advanced by time mstep.
  //
  // NOTE: Round-off error for extremely small mstep is not as small
  //       as is perhaps possible.  There is an explicit timestep==0.0
  //       check to handle one case, but this is a more general issue.
  //
  // NOTE: H should be zeroed at fixed spin sites before being passed to
  //       this function.  (In other words, this function makes no fixed
  //       spin consideration.)



#ifndef NDEBUG
  // Consistency checks
  if(!H.CheckMesh(old_state.mesh)) {
    throw Oxs_Ext::Error(this,
                         "Oxs_SemiAnalyticEvolve::AdjustState:"
                         " Import spin and H are different sizes.");
  }
  if(mesh_id != old_state.mesh->Id()
     || !gamma.CheckMesh(old_state.mesh)
     || !alpha.CheckMesh(old_state.mesh)) {
    throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::AdjustState:"
           " mesh_id, gamma or alpha not properly initialized.");
  }
#endif // NDEBUG

  const Oxs_MeshValue<ThreeVector>& old_spin = old_state.spin;
  Oxs_MeshValue<ThreeVector>& new_spin = new_state.spin;

  new_spin.AdjustSize(old_state.mesh);
  const UINT4m size = old_state.mesh->Size();

  UINT4m i;
  for(i=0;i<size;++i) {

    const ThreeVector& m = old_spin[i];

    // Assume m is either a unit vector or at least pretty close.
    // If |H| is too small to move m by at least REAL8_EPSILON (say
    // REAL8_EPSILON/16 just to be generous) even in the case where
    // H and m are orthogonal, then we can just skip the rest and
    // copy m over to new_spin.
    REAL8m Hnormsq = H[i].MagSq();
#ifdef DEEP_DEBUG
    if(Hnormsq<=0.0) {
      new_spin[i] = m;
      continue;
    }
#else // DEEP_DEBUG
    if(gamma[i]*gamma[i]*mstep*mstep*Hnormsq*(1+alpha[i]*alpha[i])
       < REAL8_EPSILON*REAL8_EPSILON/(16*16)) {
      new_spin[i] = m;
      continue;
    }
#endif // DEEP_DEBUG
    REAL8m Hnorm = sqrt(Hnormsq);

    // Introduce spherical coordinate system built off of a rotated
    // Cartesian coordinate system having +z direction parallel to h, x
    // parallel to m minus the projection of m onto H, and y parallel to
    // H x m.
    ThreeVector ez = H[i];
    ez *= 1.0/Hnorm;

    REAL8m dot = m*ez;

    ThreeVector ex(m.x - dot*ez.x,
                   m.y - dot*ez.y,
                   m.z - dot*ez.z);
    REAL8m ex_normsq = ex.MagSq();
    if(ex_normsq == 0.0) { new_spin[i] = m; continue; }
    ex *= 1.0/sqrt(ex_normsq);

    ThreeVector ey = ez;
    ey ^= ex;
    REAL8m ey_normsq = ey.MagSq();
    if(ey_normsq == 0.0) { new_spin[i] = m; continue; }
    ey *= 1.0/sqrt(ey_normsq);

    // The damping term is
    //   2.atan(tan(theta0/2).e^(alpha.gamma.|H|.step)
    //   = 2.atan(tan(acos(m*ez)/2).e^(alpha.gamma.|H|.step)
    // Using the relations
    //   tan(a/2) = (1-cos(a))/sin(a) = sin(a)/(1+cos(a))
    // we can rewrite this as
    //   2.atan(((1-m*ez)/|m x ez|).e^(alpha.gamma.|H|.step)
    // = 2.atan((|m x ez|/(1+m*ez)).e^(alpha.gamma.|H|.step)
    //                 (See Gradshteyn and Ryzhik, 1.317.5.)
    // This involves computing the cross product m x ez, but
    // removes the acos computation, and if we choose between
    // the last two formulae depending on whether m*ez is <0
    // or >0 (respectively), then should be more accurate to
    // boot.  For example, an expansion for acos(x) for x
    // near 1 is
    //
    //   acos(x) = sqrt(2*(1-x))*(1+(1-x)/12+3.(1-x)^2/160+...)
    //
    // Since we have x = dot = m*ez, the lead term is
    // sqrt(2*(1-m*ez)), and 1-m*ez is noisy for m and ez
    // close to parallel.
#if 0 // DEEP_DEBUG
    REAL8m phi0   = 0.0;
    REAL8m phi1 = -1*gamma[i]*Hnorm*mstep + phi0;
    REAL8m theta0 = acos(dot);
    REAL8m theta1 = 2*atan(tan(theta0/2.)
                     *exp(alpha[i]*gamma[i]*Hnorm*mstep));
#else
    REAL8m phi0   = 0.0;
    REAL8m phi1 = -1*gamma[i]*Hnorm*mstep + phi0;
    ThreeVector tmp = m;
    tmp ^= ez;
    REAL8m sin_mez = sqrt(tmp.MagSq());
    REAL8m foo = exp(alpha[i]*gamma[i]*Hnorm*mstep);
    if(dot>=0.0) {
      foo *= sin_mez/(1+dot);
    } else {
      foo *= (1-dot);
      foo /= sin_mez;   // sin_mez will be close to zero
      /// if dot is close to -1.
    }
#if 0
    REAL8m theta1 = 2*atan(foo);
#endif
#endif

    // Convert to rotated Cartesian coordinates
#if 0
    REAL8m z1 = cos(theta1);
    REAL8m sintheta1 = sin(theta1);  // Not very accurate for theta1 =~ PI
#else
    REAL8m invfoo = 1.0/foo;
    REAL8m z1 = 1.0;
    if(foo != 0.0) {
      if(invfoo == 0.0) z1 = -1.0;
      else              z1 = (invfoo - foo)/(invfoo + foo);
    }
    REAL8m sintheta1 = 2./(invfoo + foo);
    // Note: foo + invfoo can't overflow, because for example
    // if foo>1e8 then invfoo<1e-8 so foo + invfoo = foo [fp]
    // (assuming eps =~ 1e-16), and similarly for invfoo>1e8.
#endif
    REAL8m x1 = cos(phi1)*sintheta1;
    REAL8m y1 = sin(phi1)*sintheta1;

    // Convert back to original Cartesian coordinates.
    new_spin[i].Set(x1*ex.x+y1*ey.x+z1*ez.x,
                    x1*ex.y+y1*ey.y+z1*ez.y,
                    x1*ex.z+y1*ey.z+z1*ez.z);
  }

  // Adjust time and iteration fields in new_state
  new_state.last_timestep=hstep;
  if(old_state.stage_number != new_state.stage_number) {
    // New stage
    new_state.stage_start_time = old_state.stage_start_time
                                + old_state.stage_elapsed_time;
    new_state.stage_elapsed_time = new_state.last_timestep;
  } else {
    new_state.stage_start_time = old_state.stage_start_time;
    new_state.stage_elapsed_time = old_state.stage_elapsed_time
                                  + new_state.last_timestep;
  }

  // Don't touch iteration counts. (?!)  The problem is that one call to
  // the Oxs_SemiAnalyticEvolve::...Step2() routines take 2 half-steps,
  // and it is the result from these half-steps that are used as the
  // export state.  If we increment the iteration count each time
  // through here, then the iteration count goes up by 2 for each call
  // to ...Step2().  So instead, we leave iteration count at whatever
  // value was filled in by the
  // Oxs_SemiAnalyticEvolve::NegotiateTimeStep() method.
}

void
Oxs_SemiAnalyticEvolve::AdjustState
(REAL8m hstep,
 REAL8m mstep,
 const Oxs_SimState& old_state,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_SimState& new_state) const
{
  // Advance spins about field H using LLG dynamics for the timestep
  // mstep (in seconds).  For details, see NOTES III, 25-Nov-2002, p15.
  //
  // Export new state has time index from old_state + hstep,
  // and spins from old state advanced by time mstep.
  //
  // NOTE: Round-off error for extremely small mstep is not as small
  //       as is perhaps possible.  There is an explicit timestep==0.0
  //       check to handle one case, but this is a more general issue.
  //
  // NOTE: H should be zeroed at fixed spin sites before being passed to
  //       this function.  (In other words, this function makes no fixed
  //       spin consideration.)
  // NOTE: This is the same as the preceding AdjustState, except this
  //       routine uses an additional import (mxH) in an attempt to
  //       improve numerics.

#ifndef NDEBUG
  // Consistency checks
  if(!H.CheckMesh(old_state.mesh)) {
    throw Oxs_Ext::Error(this,
                         "Oxs_SemiAnalyticEvolve::AdjustState:"
                         " Import spin and H are different sizes.");
  }
  if(!mxH.CheckMesh(old_state.mesh)) {
    throw Oxs_Ext::Error(this,
                         "Oxs_SemiAnalyticEvolve::AdjustState:"
                         " Import spin and mxH are different sizes.");
  }
  if(mesh_id != old_state.mesh->Id()
     || !gamma.CheckMesh(old_state.mesh)
     || !alpha.CheckMesh(old_state.mesh)) {
    throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::AdjustState:"
           " mesh_id, gamma or alpha not properly initialized.");
  }
#endif // NDEBUG

  const Oxs_MeshValue<ThreeVector>& old_spin = old_state.spin;
  Oxs_MeshValue<ThreeVector>& new_spin = new_state.spin;

  new_spin.AdjustSize(old_state.mesh);
  const UINT4m size = old_state.mesh->Size();

  UINT4m i;
  for(i=0;i<size;++i) {

    const ThreeVector& m = old_spin[i];

#ifdef DEEP_DEBUG
    int zero_movement = 0;

    if(CHECK_START <= old_state.iteration_count
       && old_state.iteration_count < CHECK_STOP
       && CHECK_INDEX == i) {
      fprintf(stdout,"..AS m0[%4u] = (%18.15f,%18.15f,%18.15f)\n",
              (unsigned int)i,m.x,m.y,m.z);
      fprintf(stdout,"............H = (%18.15f,%18.15f,%18.15f)\n",
              H[i].x,H[i].y,H[i].z);
      fprintf(stdout,"..........mxH = (%18.15f,%18.15f,%18.15f)\n",
              mxH[i].x,mxH[i].y,mxH[i].z);
      fprintf(stdout,"..........m*H =  %18.15f\n",m*H[i]);
    }
#endif // DEEP_DEBUG

    ThreeVector ey = mxH[i];
    if(ey.MagSq() == 0.0) {
      // No torque
#ifdef DEEP_DEBUG
      if(CHECK_START <= old_state.iteration_count
         && old_state.iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        ++zero_movement;
        fprintf(stdout,"ZEROED MOVEMENT (a) at index %d\n",i); /**/
      } else {
        new_spin[i] = m;
        continue;
      }
#else
      new_spin[i] = m;
      continue;
#endif
    }
    ey *= -1;
    ey.MakeUnit();

    ThreeVector tH = H[i];
    // tH is suppose to be orthogonal to mxH.  Make sure.
    tH.Accum(-1*(tH * ey),ey);

    // Assume m is either a unit vector or at least pretty close.
    // If |H| is too small to move m by at least REAL8_EPSILON (say
    // REAL8_EPSILON/16 just to be generous) even in the case where
    // H and m are orthogonal, then we can just skip the rest and
    // copy m over to new_spin.
    REAL8m Hnormsq = tH.MagSq();
    if(gamma[i]*gamma[i]*mstep*mstep*Hnormsq*(1+alpha[i]*alpha[i])
       < REAL8_EPSILON*REAL8_EPSILON/(16*16)) {
#ifdef DEEP_DEBUG
      if(CHECK_START <= old_state.iteration_count
         && old_state.iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        ++zero_movement;
        fprintf(stdout,"ZEROED MOVEMENT (b) at index %d\n",i); /**/
      } else {
        new_spin[i] = m;
        continue;
      }
#else
      new_spin[i] = m;
      continue;
#endif
    }
    REAL8m Hnorm = sqrt(Hnormsq);

    // Introduce spherical coordinate system built off of a rotated
    // Cartesian coordinate system having +z direction parallel to h, x
    // parallel to m minus the projection of m onto H, and y parallel to
    // H x m.
    ThreeVector ez = tH;
    ez.MakeUnit();

    REAL8m dotyz = ey*ez;
#if 0
    if(fabs(dotyz)>128*REAL8_EPSILON) {
      // Bad data???  Punt: fall back to great circle trajectory.
      ez = mxH[i];
      ez ^= m;
      Hnorm = sqrt(ez.MagSq());
      ez.MakeUnit();
      dotyz = ey*ez;
    }
#endif
    if(dotyz!=0.0) {
      // Sharpen coordinates
      ey.Accum(-dotyz,ez);
      ey.MakeUnit();
    }

    ThreeVector ex = ey;
    ex ^= ez;
    ex.MakeUnit();  // Safety
    
    REAL8m dot = m*ez;

#if 0
    if(dot<-0.5) {
      fprintf(stdout,"Iteration %5d, index %5d, dot = %5.2f\n",
              (int)old_state.iteration_count,(int)i,dot);
    }
#endif
#if 0
    if(dot <= -1.0 || dot >= 1.0) { // Protect against roundoff when
      /// m and H are parallel or anti-parallel.  Either way, spin
      /// does not move.
      /// BAD BAD BAD!  The new code below handles round-off in dot.
      /// Moreover (and more importantly), just because dot is close
      /// to +/-1 does ***not*** mean that the spin doesn't move.
      /// For example, suppose the angle between m and ez is 1e-9
      /// radians.  Then m*ez =~ 1 + (1e-9)^2/2 = 1 [fp].  But the
      /// orthogonal term is on the order of sin(1e-9) =~ 1e-9, which
      /// is relatively large (~sqrt(eps)) compared to |m|.
      /// ADVICE: Junk this code.
#ifdef DEEP_DEBUG
      if(CHECK_START <= old_state.iteration_count
         && old_state.iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        ++zero_movement;
        fprintf(stdout,"ZEROED MOVEMENT (c) at index %d\n",i); /**/
      } else {
        new_spin[i] = m;
        continue;
      }
#else
      new_spin[i] = m;
      continue;
#endif
    }
#endif

    // The damping term is
    //   2.atan(tan(theta0/2).e^(alpha.gamma.|H|.step)
    //   = 2.atan(tan(acos(m*ez)/2).e^(alpha.gamma.|H|.step)
    // Using the relations
    //   tan(a/2) = (1-cos(a))/sin(a) = sin(a)/(1+cos(a))
    // we can rewrite this as
    //   2.atan(((1-m*ez)/|m x ez|).e^(alpha.gamma.|H|.step)
    // = 2.atan((|m x ez|/(1+m*ez)).e^(alpha.gamma.|H|.step)
    //                 (See Gradshteyn and Ryzhik, 1.317.5.)
    // This involves computing the cross product m x ez, but
    // removes the acos computation, and if we choose between
    // the last two formulae depending on whether m*ez is <0
    // or >0 (respectively), then should be more accurate to
    // boot.  For example, an expansion for acos(x) for x
    // near 1 is
    //
    //   acos(x) = sqrt(2*(1-x))*(1+(1-x)/12+3.(1-x)^2/160+...)
    //
    // Since we have x = dot = m*ez, the lead term is
    // sqrt(2*(1-m*ez)), and 1-m*ez is noisy for m and ez
    // close to parallel.
#if 0
    REAL8m phi0   = 0.0;
    REAL8m phi1 = -1*gamma[i]*Hnorm*mstep + phi0;
    REAL8m theta0 = acos(dot);
    REAL8m theta1 = 2*atan(tan(theta0/2.)
                     *exp(alpha[i]*gamma[i]*Hnorm*mstep));
#else
    REAL8m phi0   = 0.0;
    REAL8m phi1 = -1*gamma[i]*Hnorm*mstep + phi0;
    ThreeVector tmp = m;
    tmp ^= ez;
    REAL8m sin_mez = sqrt(tmp.MagSq());
    REAL8m foo = exp(alpha[i]*gamma[i]*Hnorm*mstep);
    if(dot>=0.0) {
      foo *= sin_mez/(1+dot);
    } else {
      foo *= (1-dot);
      foo /= sin_mez;   // sin_mez will be close to zero
      /// if dot is close to -1.
    }
#if 0
    REAL8m theta1 = 2*atan(foo);
#endif
#endif

    // Convert to rotated Cartesian coordinates
#if 0
    REAL8m z1 = cos(theta1);
    REAL8m sintheta1 = sin(theta1);  // Not very accurate for theta1 =~ PI
#else
    REAL8m invfoo = 1.0/foo;
    REAL8m z1 = 1.0;
    if(foo != 0.0) {
      if(invfoo == 0.0) z1 = -1.0;
      else              z1 = (invfoo - foo)/(invfoo + foo);
    }
    REAL8m sintheta1 = 2./(invfoo + foo);
    // Note: foo + invfoo can't overflow, because for example
    // if foo>1e8 then invfoo<1e-8 so foo + invfoo = foo [fp]
    // (assuming eps =~ 1e-16), and similarly for invfoo>1e8.
#endif
    REAL8m x1 = cos(phi1)*sintheta1;
    REAL8m y1 = sin(phi1)*sintheta1;

    // Convert back to original Cartesian coordinates.
    new_spin[i].Set(x1*ex.x+y1*ey.x+z1*ez.x,
                    x1*ex.y+y1*ey.y+z1*ez.y,
                    x1*ex.z+y1*ey.z+z1*ez.z);

#ifdef DEEP_DEBUG
    if(CHECK_START <= old_state.iteration_count
       && old_state.iteration_count < CHECK_STOP
       && CHECK_INDEX == i) {
      fprintf(stdout,".......... mb = (%18.15f,%18.15f,%18.15f)\n",
              new_spin[i].x,new_spin[i].y,new_spin[i].z);
    }
#endif // DEEP_DEBUG

#ifdef DEEP_DEBUG
    if(zero_movement>0) {
      if(CHECK_START <= old_state.iteration_count
         && old_state.iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        // Compute m -> m~ -> m
        ThreeVector zd;
        REAL8m tx = m*ex;
        REAL8m ty = m*ey;
        REAL8m tz = m*ez;
        zd.Set(tx*ex.x+ty*ey.x+tz*ez.x,
               tx*ex.y+ty*ey.y+tz*ez.y,
               tx*ex.z+ty*ey.z+tz*ez.z);
        zd -= m;
        fprintf(stdout,"..xfrmed diff = (%18.15f,%18.15f,%18.15f) = %#.5g\n",
                zd.x,zd.y,zd.z,sqrt(zd.MagSq()));
        zd = new_spin[i];
        zd -= m;
        fprintf(stdout,"..zeroed diff = (%18.15f,%18.15f,%18.15f) = %#.5g\n",
                zd.x,zd.y,zd.z,sqrt(zd.MagSq()));
        fprintf(stdout,"..sin_mez = %#.5g\n",sin_mez);
        fprintf(stdout,"...theta1 = %#.5g\n",2*atan(foo));
      }
      new_spin[i] = m;
    }
#if 0
    if(CHECK_START <= old_state.iteration_count
       && old_state.iteration_count < CHECK_STOP
       && CHECK_INDEX == i) {
      fprintf(stdout,"... ex*ex=%# 10.3e, ey*ey=%# 10.3e, ez*ez=%# 10.3e\n",
              ex.MagSq(),ey.MagSq(),ez.MagSq());
      fprintf(stdout,"... ex*ey=%# 10.3e, ex*ez=%# 10.3e, ey*ez=%# 10.3e\n",
              ex*ey,ex*ez,ey*ez);
    }
#endif

#endif // DEEP_DEBUG

  }

  // Adjust time and iteration fields in new_state
  new_state.last_timestep=hstep;
  if(old_state.stage_number != new_state.stage_number) {
    // New stage
    new_state.stage_start_time = old_state.stage_start_time
                                + old_state.stage_elapsed_time;
    new_state.stage_elapsed_time = new_state.last_timestep;
  } else {
    new_state.stage_start_time = old_state.stage_start_time;
    new_state.stage_elapsed_time = old_state.stage_elapsed_time
                                  + new_state.last_timestep;
  }

  // Don't touch iteration counts. (?!)  The problem is that one call to
  // the Oxs_SemiAnalyticEvolve::...Step2() routines take 2 half-steps,
  // and it is the result from these half-steps that are used as the
  // export state.  If we increment the iteration count each time
  // through here, then the iteration count goes up by 2 for each call
  // to ...Step2().  So instead, we leave iteration count at whatever
  // value was filled in by the
  // Oxs_SemiAnalyticEvolve::NegotiateTimeStep() method.
}

void
Oxs_SemiAnalyticEvolve::AxisProject
(const Oxs_MeshValue<ThreeVector>& m,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& axis,
 Oxs_MeshValue<ThreeVector>& Hproj) const
{ // Export Hproj is a projection of H onto axis (i.e., Hproj || axis)
  // such that m x H == m x Hproj
  const UINT4m size = m.Size();
  if(size != H.Size()) {
    throw Oxs_Ext::Error(this,
                         "Oxs_SemiAnalyticEvolve::AxisProject:"
                         " Import m and H are different sizes.");
  }
  if(size != axis.Size()) {
    throw Oxs_Ext::Error(this,
                         "Oxs_SemiAnalyticEvolve::AxisProject:"
                         " Import m and axis are different sizes.");
  }
  Hproj.AdjustSize(m);

#if 1
  ThreeVector tmpa,tmpb;
  for(UINT4m i=0;i<size;++i) {
    // This code assumes that axis lies in plane defined by m and H
#if 1
    tmpa = m[i];    tmpa ^= H[i];
    tmpb = m[i];    tmpb ^= axis[i];  REAL8m tmpb_magsq = tmpb.MagSq();
    REAL8m dot = tmpa*tmpb;
    if(tmpb_magsq<1.0 && fabs(dot)*REAL8_EPSILON>=tmpb_magsq) {
      // Either m and axis are nearly parallel, or else |axis|
      // is very small.  The latter can happen if the vectors
      // being crossed to yield axis are nearly parallel, or,
      // of course, if either of the crossed vectors is nearly
      // zero.  First try: punt, by falling back to H.
      // NOTE: Might want to include a check against |u|
      Hproj[i] = H[i];
    } else {
      REAL8m mult = dot/tmpb_magsq;
      Hproj[i] = axis[i];
      Hproj[i] *= mult;
    }
#else
    tmpa = m[i];    tmpa ^= H[i];     REAL8m tmpa_magsq = tmpa.MagSq();
    tmpb = m[i];    tmpb ^= axis[i];  REAL8m tmpb_magsq = tmpb.MagSq();
    if(tmpb_magsq<1.0 && tmpa_magsq*REAL8_EPSILON>=tmpb_magsq) {
      // Either m and axis are nearly parallel, or else |axis|
      // is very small.  The latter can happen if the vectors
      // being crossed to yield axis are nearly parallel, or,
      // of course, if either of the crossed vectors is nearly
      // zero.  First try: punt, by falling back to H.
      Hproj[i] = H[i];
    } else {
      REAL8m mult = sqrt(tmpa_magsq/tmpb_magsq);
      if(tmpa*tmpb<0) {
        mult *= -1;
      }
      Hproj[i] = axis[i];
      Hproj[i] *= mult;
    }
#endif
  }
#else
  for(UINT4m i=0;i<size;++i) {
    // This code works even if axis is not in plane defined by m and H,
    // in which case it projects to the projection of axis onto the
    // m-H plane.  Also, as compared to the alternative block above,
    // this block does not require a square root operation.  I have
    // not checked for any floating point precision issues.
    const REAL8m m_taxicabnorm = m[i].TaxicabNorm();
    if(m_taxicabnorm == 0.0) { // Safety
      Hproj[i] = H[i];
      continue;
    }
    REAL8m mm = m[i].MagSq();
    REAL8m mH = m[i]*H[i];
    REAL8m mu = m[i]*axis[i];
    REAL8m Hu = H[i]*axis[i];
    REAL8m uu = axis[i].MagSq();

    REAL8m multnum = mu*Hu - mH*uu;
    REAL8m multden = mm*uu - mu*mu;

    // Compute rough adjustment limit. Use taxicab norm, because it is
    // cheaper to compute than L2 norm.  (Also less prone to overflow
    // than simple computation of L2 by sqrt(H*H).)
#if 1
    const REAL8m LIMIT = 1 + 10000*H[i].TaxicabNorm()/m_taxicabnorm;
#else
    const REAL8m LIMIT = 0.0;
#endif
    if(multden<=0.0) {
      // Either m and u are parallel, or else bad numerics.
      Hproj[i] = m[i];
      if(multnum>0) {
        Hproj[i] *= LIMIT;
      } else if(multnum<0) {
        Hproj[i] *= -1*LIMIT;
      } else {
        // Punt;
        Hproj[i].Set(0,0,0);
      }
      Hproj[i] += H[i];
      continue;
    }

    // Limit adjustment.  This mainly provides some protection in the
    // case where u is nearly (or exactly) parallel to m.
    Hproj[i] = m[i];
    if(fabs(multnum) > LIMIT*multden) {
      if(multnum>0) {
        Hproj[i] *=  LIMIT;
      } else {
        Hproj[i] *= -LIMIT;
      }
    } else {
      Hproj[i] *= multnum/multden;
    }
    Hproj[i] += H[i];
  }
#endif

}

void
Oxs_SemiAnalyticEvolve::AxisCorrectedAdjustState
(REAL8m hstep,
 REAL8m mstep,
 const Oxs_SimState& current_state,
 const Oxs_MeshValue<ThreeVector>& mxH_old,
 const Oxs_MeshValue<ThreeVector>& mxH_current,
 Oxs_SimState& new_state) const
{ // Same as ::AdjustState, but takes two torques as input
  // and produces the axis-corrected adjustment, which is
  // analytic LLG evolution about the axis mxH_old x mxH.

#ifndef NDEBUG
  // Consistency checks
  if(!mxH_current.CheckMesh(current_state.mesh)) {
    throw Oxs_Ext::Error(this,
                         "Oxs_SemiAnalyticEvolve::AdjustState:"
                         " Import spin and H are different sizes.");
  }
  if(mesh_id != current_state.mesh->Id()
     || !gamma.CheckMesh(current_state.mesh)
     || !alpha.CheckMesh(current_state.mesh)) {
    throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::AdjustState:"
           " mesh_id, gamma or alpha not properly initialized.");
  }
#endif // NDEBUG

  const Oxs_MeshValue<ThreeVector>& current_spin = current_state.spin;
  Oxs_MeshValue<ThreeVector>& new_spin = new_state.spin;

  new_spin.AdjustSize(current_state.mesh);
  const UINT4m size = current_state.mesh->Size();

  UINT4m i;
  for(i=0;i<size;++i) {

    const ThreeVector& m = current_spin[i];

    REAL8m torque_magsq = mxH_current[i].MagSq();
    if(torque_magsq == 0.0) {
      // Zero torque, so no spin movement
      new_spin[i] = m;
      continue;
    }

    ThreeVector H = mxH_old[i];
    H ^= mxH_current[i];  // Rotation axis direction

    if(H.MagSq()<REAL8_EPSILON) {
      // Old and current torques are or are nearly parallel.
      // Punt.  Set axis to direction perpendicular to m and
      // torque, so spin follows great circle trajectory.
      H = m;
      H ^= mxH_current[i];
    }

    ThreeVector test = m;  test ^= H;
    REAL8m normb_sq = test.MagSq();

    if(normb_sq<1.0 && torque_magsq*REAL8_EPSILON > normb_sq) {
      // spin lies along rotation axis, so no spin movement
      new_spin[i] = m;
      continue;
    }

    H *= sqrt(torque_magsq/normb_sq);
    /// This value of H lies along direction of Torque_old x Torque,
    /// and has magnitude adjusted so it produces same torque on
    /// m as original Torque.

    // The remainder is the same as ::AdjustState
    REAL8m Hnormsq = H.MagSq();
    if(Hnormsq == 0.0) {
      new_spin[i] = m;
      continue;
    }
    REAL8m Hnorm = sqrt(Hnormsq);

    // Introduce spherical coordinate system built off of a rotated
    // Cartesian coordinate system having +z direction parallel to h, x
    // parallel to m minus the projection of m onto H, and y parallel to
    // H x m.
    ThreeVector ez = H;
    ez *= 1.0/Hnorm;

    REAL8m dot = m*ez;
    if(dot <= -1 || dot >= 1) { // Protect against roundoff when
      // m and H are parallel or anti-parallel.  Either way,
      // spin does not move.
      new_spin[i] = m;
      continue;
    }

    ThreeVector ex(m.x - dot*ez.x,
                   m.y - dot*ez.y,
                   m.z - dot*ez.z);
    REAL8m ex_normsq = ex.MagSq();
    if(ex_normsq == 0.0) { new_spin[i] = m; continue; }
    ex *= 1.0/sqrt(ex_normsq);

    ThreeVector ey = ez;
    ey ^= ex;
    REAL8m ey_normsq = ey.MagSq();
    if(ey_normsq == 0.0) { new_spin[i] = m; continue; }
    ey *= 1.0/sqrt(ey_normsq);

    REAL8m theta0 = acos(dot);
    REAL8m phi0   = 0.0;

    REAL8m phi1 = -1*gamma[i]*Hnorm*mstep + phi0;
    REAL8m theta1 = 2*atan(tan(theta0/2.)
                     *exp(alpha[i]*gamma[i]*Hnorm*mstep));

    // Convert to rotated Cartesian coordinates
    REAL8m z1 = cos(theta1);
    REAL8m sintheta1 = sin(theta1);
    REAL8m x1 = cos(phi1)*sintheta1;
    REAL8m y1 = sin(phi1)*sintheta1;

    // Convert back to original Cartesian coordinates.
    new_spin[i].Set(x1*ex.x+y1*ey.x+z1*ez.x,
                    x1*ex.y+y1*ey.y+z1*ez.y,
                    x1*ex.z+y1*ey.z+z1*ez.z);
  }

  // Adjust time and iteration fields in new_state
  new_state.last_timestep=hstep;
  if(current_state.stage_number != new_state.stage_number) {
    // New stage
    new_state.stage_start_time = current_state.stage_start_time
                                + current_state.stage_elapsed_time;
    new_state.stage_elapsed_time = new_state.last_timestep;
  } else {
    new_state.stage_start_time = current_state.stage_start_time;
    new_state.stage_elapsed_time = current_state.stage_elapsed_time
                                  + new_state.last_timestep;
  }

  // Don't touch iteration counts. (?!)  The problem is that one call to
  // the Oxs_SemiAnalyticEvolve::...Step2() routines take 2 half-steps,
  // and it is the result from these half-steps that are used as the
  // export state.  If we increment the iteration count each time
  // through here, then the iteration count goes up by 2 for each call
  // to ...Step2().  So instead, we leave iteration count at whatever
  // value was filled in by the
  // Oxs_SemiAnalyticEvolve::NegotiateTimeStep() method.
}


void Oxs_SemiAnalyticEvolve::UpdateTimeFields
(const Oxs_SimState& cstate,
 Oxs_SimState& nstate,
 REAL8m stepsize) const
{
  nstate.last_timestep=stepsize;
  if(cstate.stage_number != nstate.stage_number) {
    // New stage
    nstate.stage_start_time = cstate.stage_start_time
                              + cstate.stage_elapsed_time;
    nstate.stage_elapsed_time = stepsize;
  } else {
    nstate.stage_start_time = cstate.stage_start_time;
    nstate.stage_elapsed_time = cstate.stage_elapsed_time
                                + stepsize;
  }
}

void Oxs_SemiAnalyticEvolve::NegotiateTimeStep
(const Oxs_TimeDriver* driver,
 const Oxs_SimState&  cstate,
 Oxs_SimState& nstate,
 REAL8m stepsize,
 BOOL use_start_dm,
 BOOL& force_step,
 BOOL& driver_set_step) const
{ // This routine negotiates with driver over the proper step size.
  // As a side-effect, also initializes the nstate data structure,
  // where nstate is the "next state".

  // Pull needed cached values out from cstate.
  REAL8m max_dm_dt;
  if(!cstate.GetDerivedData("Max dm/dt",max_dm_dt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::NegotiateTimeStep: max_dm_dt not cached.");
  }
  REAL8m timestep_lower_bound=0.;  // Smallest timestep that can actually
  /// change spin with max_dm_dt (due to REAL8_EPSILON restrictions).
  if(!cstate.GetDerivedData("Timestep lower bound",
                            timestep_lower_bound)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::NegotiateTimeStep: "
       " timestep_lower_bound not cached.");
  }
  if(timestep_lower_bound<=0.0) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::NegotiateTimeStep: "
       " cached timestep_lower_bound value not positive.");
  }

  // The next timestep is based on the error from the last step.  If
  // there is no last step (either because this is the first step,
  // or because the last state handled by this routine is different
  // from the incoming current_state), then timestep is calculated
  // so that max_dm_dt * timestep = start_dm.
  if(use_start_dm || stepsize<=0.0) {
    if(start_dm < sqrt(DBL_MAX/4) * max_dm_dt) {
      stepsize = step_headroom * start_dm / max_dm_dt;
    } else {
      stepsize = sqrt(DBL_MAX/4);
    }
  }

  // Insure step is not outside requested step bounds
  if(stepsize>max_timestep) stepsize = max_timestep;
  if(stepsize<min_timestep) stepsize = min_timestep;
  if(stepsize<timestep_lower_bound) stepsize = timestep_lower_bound;

  // Negotiate with driver over size of next step
  driver->FillState(cstate,nstate);
  UpdateTimeFields(cstate,nstate,stepsize);

  // Update iteration count
  nstate.iteration_count = cstate.iteration_count + 1;
  nstate.stage_iteration_count = cstate.stage_iteration_count + 1;

  // Additional timestep control
  driver->FillStateSupplemental(nstate);

  // Check for forced step
  force_step = 0;
  if(nstate.last_timestep<=min_timestep ||
     nstate.last_timestep<=timestep_lower_bound) {
    // Either driver wants to force stepsize, or else step size can't
    // be reduced any further.
    force_step=1;
  }

  // Is driver responsible for stepsize?
  if(nstate.last_timestep == stepsize) driver_set_step = 0;
  else                                 driver_set_step = 1;
}


BOOL
Oxs_SemiAnalyticEvolve::CheckError
(REAL8m global_error_order,
 REAL8m error,
 REAL8m stepsize,
 REAL8m reference_stepsize,
 REAL8m max_dm_dt,
 REAL8m& new_stepsize)
{ // Returns 1 if step is good, 0 if error is too large.
  // Export new_stepsize is set to suggested stepsize
  // for next step.
  //
  // new_stepsize is initially filled with a relative stepsize
  // adjustment ratio, e.g., 1.0 means no change in stepsize.
  // At the end of this routine this ratio is multiplied by
  // stepsize to get the actual absolute stepsize.
  //
  // The import stepsize is the size (in seconds) of the actual
  // step.  The new_stepsize is computed from this based on the
  // various error estimates.  An upper bound is placed on the
  // size of new_stepsize relative to the imports stepsize and
  // reference_stepsize.  reference_stepsize has no effect if
  // it is smaller than max_step_increase*stepsize.  It is
  // usually used only in the case where the stepsize was
  // artificially reduced by, for example, a stage stopping
  // criterion.
  //
  // NOTE: This routine assumes the local error order is
  //     global_error_order + 1.

  BOOL good_step = 1;
  BOOL error_checked=0;

  if(allowed_relative_step_error>=0. || allowed_error_rate>=0.) {
    // Determine tighter rate bound.
    REAL8m rate_error = 0.0;
    if(allowed_relative_step_error<0.) {
      rate_error = allowed_error_rate;
    } else if(allowed_error_rate<0.) {
      rate_error = allowed_relative_step_error * max_dm_dt;
    } else {
      rate_error = allowed_relative_step_error * max_dm_dt;
      if(rate_error>allowed_error_rate) {
        rate_error = allowed_error_rate;
      }
    }
    rate_error *= stepsize;

    // Rate check
    if(error>rate_error) {
      good_step = 0;
      new_stepsize = pow(rate_error/error,1.0/global_error_order);
    } else {
      REAL8m ratio = 0.125*DBL_MAX;
      if(error>=1 || rate_error<ratio*error) {
        REAL8m test_ratio = rate_error/error;
        if(test_ratio<ratio) ratio = test_ratio;
      }
      new_stepsize = pow(ratio,1.0/global_error_order);
    }
    error_checked = 1;
  }

  // Absolute error check
  if(allowed_absolute_step_error>=0.0) {
    REAL8m test_stepsize = 0.0;
    REAL8m local_error_order = global_error_order + 1.0;
    if(error>allowed_absolute_step_error) {
      good_step = 0;
      test_stepsize = pow(allowed_absolute_step_error/error,
                          1.0/local_error_order);
    } else {
      REAL8m ratio = 0.125*DBL_MAX;
      if(error>=1 || allowed_absolute_step_error<ratio*error) {
        REAL8m test_ratio = allowed_absolute_step_error/error;
        if(test_ratio<ratio) ratio = test_ratio;
      }
      test_stepsize = pow(ratio,1.0/local_error_order);
    }
    if(!error_checked || test_stepsize<new_stepsize) {
      new_stepsize = test_stepsize;
    }
    error_checked = 1;
  }

  if(error_checked) {
    new_stepsize *= step_headroom;
    if(new_stepsize<max_step_decrease) {
      new_stepsize = max_step_decrease*stepsize;
    } else {
      new_stepsize *= stepsize;
      REAL8m step_bound = stepsize * max_step_increase;
      const REAL8m refrat = 0.85;  // Ad hoc value
      if(stepsize<reference_stepsize*refrat) {
        step_bound = OC_MIN(step_bound,reference_stepsize);
      } else if(stepsize<reference_stepsize) {
        REAL8m ref_bound = reference_stepsize + (max_step_increase-1)
          *(stepsize-reference_stepsize*refrat)/(1-refrat);
        /// If stepsize = reference_stepsize*refrat,
        ///     then ref_bound = reference_stepsize
        /// If stepsize = reference_stepsize,
        ///     then ref_bound = step_bound
        /// Otherwise, linear interpolation on stepsize.
        step_bound = OC_MIN(step_bound,ref_bound);
      }
      if(new_stepsize>step_bound) new_stepsize = step_bound;
    }
  } else {
    new_stepsize = stepsize;
  }

  return good_step;
}

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticStep1
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& /* mxH */,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Takes a simple semi-analytic step

  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  Oxs_SimState* tstate = &(temp_state_key.GetWriteReference());
  nstate->CloneHeader(*tstate);

  UINT4m i;
  const UINT4m size = cstate->mesh->Size();

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  // Move ahead two half steps
  AdjustState(stepsize/2,stepsize/2,*cstate,
              H,temp_state_key.GetWriteReference());
  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,NULL,&vtmpB,pE_pt);
  ZeroFixedSpinFields(NULL,&vtmpB);
  AdjustState(stepsize/2,stepsize/2,temp_state_key.GetReadReference(),
              vtmpB,*nstate);
  UpdateTimeFields(*cstate,*nstate,stepsize);
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference()); // Lock down
  nstate = NULL; // Safety
  GetEnergyDensity(*endstate,temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticStep1:"
       " Programming error; data cache (pE/pt) already set.");
  }

  // Move ahead one full step
  AdjustState(stepsize,stepsize,*cstate,
              H,temp_state_key.GetWriteReference());
  const Oxs_SimState* teststate
    = &(temp_state_key.GetReadReference()); // Release write lock


  // Error estimate is max|endstate.spin-teststate.spin| + O(stepsize^3)
  REAL8m max_error_sq = 0.0;
  for(i=0;i<size;++i) {
    ThreeVector tvec = endstate->spin[i] - teststate->spin[i];
    REAL8m error_sq = tvec.MagSq();
    if(error_sq>max_error_sq) max_error_sq = error_sq;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 1.0;
  new_energy_and_field_computed = 1;

  // Note that end energy is already in temp_energy, mxH is
  // in vtmpA, and H is in vtmpB as per contract.
}


void
Oxs_SemiAnalyticEvolve::SemiAnalyticSubstep2
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
 const Oxs_MeshValue<ThreeVector>& H,          // Field at cstate
 const Oxs_SimState* base_state,               // Used for updating state
 REAL8m effective_stepsize,                   /// time fields
 Oxs_Key<Oxs_SimState>& next_state_key         // Final state
)
{// If base_state is null, then time fields are not updated.
 // If base_state is not null, then the time fields in next_state
 // are updated from base_state with a step size of effective_stepsize
 // NOTE: Uses temp_energy and vtmpF as scratch space.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  REAL8m pE_pt;  // Dummy value; not used.

  AdjustState(stepsize,stepsize,*cstate,H,
              next_state_key.GetWriteReference());
  GetEnergyDensity(next_state_key.GetReadReference(),
                   temp_energy,NULL,&vtmpF,pE_pt);
  ZeroFixedSpinFields(NULL,&vtmpF); // vtmpF holds H value at point B

  // vtmpF holds H at point B

  vtmpF += H;
  vtmpF *= 0.5;

  // Take step with new effective H
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  AdjustState(stepsize,stepsize,*cstate,vtmpF,*nstate);
  if(base_state != NULL) {
    UpdateTimeFields(*base_state,*nstate,effective_stepsize);
  }
  next_state_key.GetReadReference(); // Release write lock
}

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticStep2
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& /* mxH */,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{
  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  Oxs_SimState* tstate = &(temp_state_key.GetWriteReference());
  nstate->CloneHeader(*tstate);
  next_state_key.GetReadReference();
  nstate = NULL; // Safety

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  // Take two half-steps
  SemiAnalyticSubstep2(stepsize/2,current_state_key,H,NULL,0,
                       temp_state_key);

  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,NULL,&vtmpE,pE_pt);
  ZeroFixedSpinFields(NULL,&vtmpE); // vtmpE holds H value at midpoint

  // Take second half-step, starting at temp_state
  SemiAnalyticSubstep2(stepsize/2,temp_state_key,vtmpE,
                       cstate,stepsize,next_state_key);
  const Oxs_SimState* endstate = &(next_state_key.GetReadReference());

  // Redo as a single step
  SemiAnalyticSubstep2(stepsize,current_state_key,H,NULL,0,
                       temp_state_key);
  const Oxs_SimState* teststate = &(temp_state_key.GetReadReference());

  // Error estimate is max|endstate.spin-teststate.spin| + O(stepsize^4) (?)
  REAL8m max_error_sq = 0.0;
  const UINT4m size = cstate->mesh->Size();
  for(UINT4m i=0;i<size;++i) {
    ThreeVector tvec = endstate->spin[i] - teststate->spin[i];
    REAL8m error_sq = tvec.MagSq();
    if(error_sq>max_error_sq) max_error_sq = error_sq;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 2.0;
  new_energy_and_field_computed = 1;

  GetEnergyDensity(*endstate,temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticStep2:"
       " Programming error; data cache (pE/pt) already set.");
  }

  // Note that end energy is in temp_energy, mxH is in vtmpA, and H is
  // in vtmpB as per contract.
}

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep1
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Takes a simple semi-analytic step
  // NOTE: It is the responsibility of the calling routine to
  //       handle the setting and updating of mxH_old_cached.

  // Do we have valid old torque?
  if(!mxH_old_cached_is_valid) {
    // No.  Fall back to uncorrected Step1
    TakeSemiAnalyticStep1(stepsize,current_state_key,H,mxH,next_state_key,
                          error_estimate,global_error_order,
                          new_energy_and_field_computed);
    return;
  }

  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  Oxs_SimState* tstate = &(temp_state_key.GetWriteReference());
  nstate->CloneHeader(*tstate);

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  // Move ahead two half steps
  AxisCorrectedAdjustState(stepsize/2,stepsize/2,*cstate,
                           mxH_old_cached,mxH,
                           temp_state_key.GetWriteReference());
  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,&vtmpA,NULL,pE_pt);
  ZeroFixedSpinFields(&vtmpA,NULL);

  AxisCorrectedAdjustState(stepsize/2,stepsize/2,
                           temp_state_key.GetReadReference(),
                           mxH,vtmpA,*nstate);
  UpdateTimeFields(*cstate,*nstate,stepsize);
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference()); // Lock down
  nstate = NULL; // Safety
  GetEnergyDensity(*endstate,temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep1:"
       " Programming error; data cache (pE/pt) already set.");
  }

  // Move ahead one full step
  AxisCorrectedAdjustState(stepsize,stepsize,*cstate,
              mxH_old_cached,mxH,
              temp_state_key.GetWriteReference());
  const Oxs_SimState* teststate
    = &(temp_state_key.GetReadReference()); // Release write lock


  // Error estimate is max|endstate.spin-teststate.spin| + O(stepsize^3)
  REAL8m max_error_sq = 0.0;
  const UINT4m size = cstate->mesh->Size();
  for(UINT4m i=0;i<size;++i) {
    ThreeVector tvec = endstate->spin[i] - teststate->spin[i];
    REAL8m error_sq = tvec.MagSq();
    if(error_sq>max_error_sq) max_error_sq = error_sq;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 1.0;
  new_energy_and_field_computed = 1;

  // Note that end energy is already in temp_energy, mxH is
  // in vtmpA, and H is in vtmpB as per contract.
}

void
Oxs_SemiAnalyticEvolve::SemiAnalyticAxisCorrectedSubstep2
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
 const Oxs_MeshValue<ThreeVector>& H,          // Field at cstate
 const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
 const Oxs_MeshValue<ThreeVector>& Torque_alt, // Alternative torque,
                                    /// used to compute rotation axis
 const Oxs_SimState* base_state,               // Used for updating state
 REAL8m effective_stepsize,                   /// time fields
 Oxs_Key<Oxs_SimState>& next_state_key         // Final state
)
{ // Computes one semi-analytic predictor-corrector step with axis
  // correction, of size stepsize, with no error control.  This routine is
  // intended as a support function for ::SemiAnalyticAxisCorrectedStep2.
  // In addition to modifying next_state_key, this routine also uses
  // as scratch space vtmpA, vtmpB and temp_energy.  Energy is
  // computed for the predictor step, but *not* at the corrected step,
  // so if that is desired the calling routine should call
  // GetEnergyDensity() on next_state upon return.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& cspin = cstate->spin;

  const UINT4m size = cstate->mesh->Size();
  const REAL8m SINCHECKSQ = 1e-4;

  UINT4m i=0;
  REAL8m pE_pt;  // Dummy value; not used.

  ThreeVector axis,tmpa,tmpb;
  vtmpA.AdjustSize(cstate->mesh);   // Axis 1
  for(i=0;i<size;++i) {
    REAL8m checksq = SINCHECKSQ*Torque_0[i].MagSq()*Torque_alt[i].MagSq();
    axis = Torque_0[i];
    axis ^= Torque_alt[i];
    if(axis.MagSq() < checksq) {
      // Torques are too close to get good axis.
      // Fall back to great-circle axis
      vtmpA[i] = Torque_0[i];
      vtmpA[i] ^= cspin[i];
    } else {
      // Project field onto torque-derived axis
      tmpa = cspin[i]; tmpa ^= H[i];
      tmpb = cspin[i]; tmpb ^= axis;  REAL8m tmpb_magsq = tmpb.MagSq();
      REAL8m dot = tmpa*tmpb;
      if(tmpb_magsq<1.0 && fabs(dot)*REAL8_EPSILON>=tmpb_magsq) {
        vtmpA[i] = Torque_0[i];
        vtmpA[i] ^= cspin[i];
      } else {
        REAL8m mult = dot/tmpb_magsq;
        vtmpA[i] = axis;
        vtmpA[i] *= mult;
      }
    }

  }
  // vtmpA is axis projected value of H at point A

  // Move ahead one step
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              next_state_key.GetWriteReference());
  const Oxs_SimState* tstate = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;
  GetEnergyDensity(*tstate,temp_energy,&vtmpA,&vtmpB,pE_pt);

  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  /// vtmpA holds mxH at point B
  /// vtmpB holds H at point B

  // Update axis and compute average projected field
  for(i=0;i<size;++i) {
    REAL8m checksq = SINCHECKSQ*vtmpA[i].MagSq()*Torque_0[i].MagSq();
    axis = vtmpA[i];
    axis ^= Torque_0[i];
    if(axis.MagSq() < checksq) {
      // Torques are too close to get good axis.
      // Fall back to great-circle axes
      tmpa = Torque_0[i];
      tmpa ^= cspin[i];   // Projected H at point A
      tmpb = vtmpA[i];
      tmpb ^= tspin[i];   // Projected H at point B
      tmpa += tmpb;
      tmpa *= 0.5;
      vtmpA[i] = tmpa;
    } else {
      // Project field onto torque-derived axis
      tmpa = cspin[i]; tmpa ^= H[i];
      tmpb = cspin[i]; tmpb ^= axis;
      REAL8m tmpb_magsq_A = tmpb.MagSq();
      REAL8m dot_A = tmpa*tmpb;

      tmpa = tspin[i]; tmpa ^= vtmpB[i];
      tmpb = tspin[i]; tmpb ^= axis;
      REAL8m tmpb_magsq_B = tmpb.MagSq();
      REAL8m dot_B = tmpa*tmpb;

      if((tmpb_magsq_A<1.0 && fabs(dot_A)*REAL8_EPSILON>=tmpb_magsq_A) ||
         (tmpb_magsq_B<1.0 && fabs(dot_B)*REAL8_EPSILON>=tmpb_magsq_B)) {
      tmpa = Torque_0[i];
      tmpa ^= cspin[i];   // Projected H at point A
      tmpb = vtmpA[i];
      tmpb ^= tspin[i];   // Projected H at point B
      tmpa += tmpb;
      tmpa *= 0.5;
      vtmpA[i] = tmpa;
      } else {
        vtmpA[i] = axis;
        vtmpA[i] *= 0.5*(dot_A/tmpb_magsq_A + dot_B/tmpb_magsq_B);
      }
    }
  }
  tstate = NULL;  // Safety.
  /// NOTE: Neither tstate nor tspin should be used beyond this point.

  // Take step with new effective H
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  AdjustState(stepsize,stepsize,*cstate,vtmpA,*nstate);
  if(base_state != NULL) {
    UpdateTimeFields(*base_state,*nstate,effective_stepsize);
  }
  next_state_key.GetReadReference(); // Release write lock
}

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Takes a simple semi-analytic step
  // NOTE: It is the responsibility of the calling routine to
  //       handle the setting and updating of mxH_old_cached.

  // Do we have valid old torque?
  if(!mxH_old_cached_is_valid) {
    // No.  Fall back to Torque Step2
    TakeSemiAnalyticTorqueStep2(stepsize,current_state_key,
                                H,mxH,next_state_key,
                                error_estimate,global_error_order,
                                new_energy_and_field_computed);
    return;
  }

  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  next_state_key.GetReadReference().CloneHeader
    (temp_state_key.GetWriteReference());
  temp_state_key.GetReadReference(); // Release write lock

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  // Move ahead two half steps

  // First half-step
  SemiAnalyticAxisCorrectedSubstep2(stepsize/2,
        current_state_key,H,mxH,mxH_old_cached,NULL,0,temp_state_key);
  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,&vtmpD,&vtmpE,pE_pt);
  ZeroFixedSpinFields(&vtmpD,&vtmpE);
  /// vtmpD holds mxH value at midpoint, vtmpE holds H value at midpoint

  // Second half-step
  SemiAnalyticAxisCorrectedSubstep2(stepsize/2,
      temp_state_key,vtmpE,vtmpD,mxH,cstate,stepsize,next_state_key);

  // For comparison, move forward one full step
  SemiAnalyticAxisCorrectedSubstep2(stepsize,
      current_state_key,H,mxH,mxH_old_cached,NULL,0,temp_state_key);

  // Auxiliary pointers to endpoints
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& endspin = endstate->spin;

  const Oxs_SimState* tstate
    = &(temp_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;

  // Error estimate is max|endspin-tspin| + O(stepsize^4) (?)
  REAL8m max_error_sq = 0.0;
  const UINT4m size = cstate->mesh->Size();
  for(UINT4m i=0;i<size;++i) {
    ThreeVector tvec = endspin[i] - tspin[i];
    REAL8m error_sq = tvec.MagSq();
    if(error_sq>max_error_sq) max_error_sq = error_sq;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 2.0;

  // Compute energy and field at *real* endpoint
  GetEnergyDensity(*endstate,
                   temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2:"
       " Programming error; data cache (pE/pt) already set.");
  }

  new_energy_and_field_computed = 1;

  // Note that end energy is in temp_energy, mxH is in vtmpA, and H is
  // in vtmpB as per contract.
}

#define TRY1
#ifdef TRY1
void
Oxs_SemiAnalyticEvolve::SemiAnalyticAxisCorrectedSubstep2b
(REAL8m old_stepsize,             // timestep between prev and cstate
 const Oxs_MeshValue<ThreeVector>& m_old,      // Spin at prev state
 const Oxs_MeshValue<ThreeVector>& Torque_old, // Torque at prev state
 REAL8m stepsize,                              // current timestep
 Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
#ifdef DEEP_DEBUG
 const Oxs_MeshValue<ThreeVector>& H_0,        // Field at cstate
#else
 const Oxs_MeshValue<ThreeVector>&,            // Field at cstate
#endif
 const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
 const Oxs_SimState* base_state,               // Used for updating state
 REAL8m effective_stepsize,                   ///    time fields
 Oxs_Key<Oxs_SimState>& next_state_key         // Final state
)
{ // Computes one semi-analytic predictor-corrector step with axis
  // correction, of size stepsize, with no error control.  This routine is
  // intended as a support function for ::SemiAnalyticAxisCorrectedStep2b.
  // In addition to modifying next_state_key, this routine also uses
  // as scratch space vtmpA, vtmpB and temp_energy.  Energy is
  // computed for the predictor step, but *not* at the corrected step,
  // so if that is desired the calling routine should call
  // GetEnergyDensity() on next_state upon return.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& cspin = cstate->spin;


  const UINT4m size = cstate->mesh->Size();
#if 0
  const REAL8m SINCHECKSQ = 1e-4;
  const REAL8m SINCHECKSQ_PRE = SINCHECKSQ;
#else
  const REAL8m SINCHECKSQ_PRE = 3e-14;
  const REAL8m SINCHECKSQ = 3e-14;
  // 1e-1 is about 18    degrees
  // 1e-2 is about  5    degrees
  // 1e-3 is about  1.8  degrees
  // 1e-4 is about  0.6  degrees
  // 1e-5 is about  0.18 degrees
  // 1e-6 is about  0.06 degrees
  // 1e-8 is about  6e-3 degrees
  // 3e-14 is about 1e-5 degrees
#endif

  UINT4m i=0;
  REAL8m pE_pt;  // Dummy value; not used.

  ThreeVector axis,tmpa,tmpb;
  vtmpA.AdjustSize(cstate->mesh);   // Axis 1
  REAL8m lambda = 0.0;
  if(old_stepsize != 0.0) {
    lambda = -0.5*stepsize/old_stepsize;
  }
  for(i=0;i<size;++i) {
    REAL8m checksq
      = SINCHECKSQ_PRE*Torque_0[i].MagSq()*Torque_old[i].MagSq();
    axis = Torque_old[i];
    axis ^= Torque_0[i];
    if(axis.MagSq() <= checksq) {
      // Torques are too close to get good axis.
      // Fall back to great-circle path
#ifdef DEEP_DEBUG
      if(CHECK_START <= cstate->iteration_count
         && cstate->iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        fprintf(stdout,"Substep pt A\n");
      }
#endif // DEEP_DEBUG
      tmpa = Torque_old[i];
      tmpa ^= m_old[i];
      tmpa *= lambda;
      tmpb = Torque_0[i];
      tmpb ^= cspin[i];
#if 1
      vtmpB[i] = tmpb;
#endif
      tmpb *= (1-lambda);
      tmpa += tmpb;
      vtmpA[i] = tmpa;
    } else {
      // Tighten axis
      tmpa = Torque_0[i];
      REAL8m dott = (axis * tmpa);
      if(fabs(dott)>4*REAL8_EPSILON) {
        tmpa *= dott/tmpa.MagSq();
        axis -= tmpa;
      }

      // Project field onto torque-derived axis
      tmpa = m_old[i]; tmpa ^= axis;  REAL8m tmpa_magsq = tmpa.MagSq();
      REAL8m dota = Torque_old[i]*tmpa;

      tmpb = cspin[i]; tmpb ^= axis;  REAL8m tmpb_magsq = tmpb.MagSq();
      REAL8m dotb = Torque_0[i]*tmpb;

      if(axis.MagSq() == 0.0 || fabs(lambda)>2 ||
         (tmpa_magsq<1.0 && 32*fabs(dota)*REAL8_EPSILON>=tmpa_magsq) ||
         (tmpb_magsq<1.0 && 32*fabs(dotb)*REAL8_EPSILON>=tmpb_magsq)) {
        // Bad data; fall back to torque-based great-circle path.
#ifdef DEEP_DEBUG
        if(CHECK_START <= cstate->iteration_count
           && cstate->iteration_count < CHECK_STOP
           && CHECK_INDEX == i) {
          fprintf(stdout,"Substep pt B\n");
        }
#endif // DEEP_DEBUG
        tmpa = Torque_old[i];
        tmpa ^= m_old[i];
        tmpa *= lambda;
        tmpb = Torque_0[i];
        tmpb ^= cspin[i];
#if 1
        vtmpB[i] = tmpb;
#endif
        tmpb *= (1-lambda);
        tmpa += tmpb;
        vtmpA[i] = tmpa;
      } else {
#ifdef DEEP_DEBUG
        if(CHECK_START <= cstate->iteration_count
           && cstate->iteration_count < CHECK_STOP
           && CHECK_INDEX == i) {
          fprintf(stdout,"Substep pt C\n");
        }
#endif // DEEP_DEBUG
        REAL8m wgta = dota/tmpa_magsq;
        REAL8m wgtb = dotb/tmpb_magsq;
        vtmpA[i] = axis;
        vtmpA[i] *= lambda*wgta + (1-lambda)*wgtb;
#if 1
        vtmpB[i] = axis;
        vtmpB[i] *= wgtb;
#endif
      }
    }

  }
  // vtmpA is axis projected value of H at point A

  // Move ahead one step
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              next_state_key.GetWriteReference());
  const Oxs_SimState* tstate = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;
  GetEnergyDensity(*tstate,temp_energy,&vtmpA,NULL,pE_pt);

  ZeroFixedSpinFields(&vtmpA,NULL);
  /// vtmpA holds mxH at point B

#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    fprintf(stdout,"Substep m0  = (%18.15f,%18.15f,%18.15f)\n",
            cstate->spin[CHECK_INDEX].x,cstate->spin[CHECK_INDEX].y,
            cstate->spin[CHECK_INDEX].z);
    fprintf(stdout,"Substep m1a = (%18.15f,%18.15f,%18.15f)\n",
            tspin[CHECK_INDEX].x,tspin[CHECK_INDEX].y,
            tspin[CHECK_INDEX].z);
    fprintf(stdout,"Substep  H0 = (%18.15g,%18.15g,%18.15g)\n",
            H_0[CHECK_INDEX].x,H_0[CHECK_INDEX].y,
            H_0[CHECK_INDEX].z);
    fprintf(stdout,"Substep Hef = (%18.15g,%18.15g,%18.15g)\n",
            vtmpA[CHECK_INDEX].x,vtmpA[CHECK_INDEX].y,
            vtmpA[CHECK_INDEX].z);
    ThreeVector foo = vtmpA[CHECK_INDEX];
    foo.MakeUnit();
    fprintf(stdout,"Substep u   = (%18.15f,%18.15f,%18.15f)\n",
            foo.x,foo.y,foo.z);
#if 0
    REAL8m footmp = foo*cstate->spin[CHECK_INDEX];
    fprintf(stdout,"Substep u*m   = %# .15g => %# .5g degrees)\n",
            footmp,acos(footmp)*180./PI);
#endif
  }
#endif // DEEP_DEBUG

  // Update axis and compute average projected field
  for(i=0;i<size;++i) {
    REAL8m checksq = SINCHECKSQ*vtmpA[i].MagSq()*Torque_0[i].MagSq();
    axis = Torque_0[i];
    axis ^= vtmpA[i];
    if(axis.MagSq() <= checksq) {
      // Torques are too close to get good axis.
      // Fall back to great-circle axes
#ifdef DEEP_DEBUG
      if(CHECK_START <= cstate->iteration_count
         && cstate->iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        fprintf(stdout,"Substep pt D\n");
      }
#endif // DEEP_DEBUG
      tmpa = Torque_0[i];
      tmpa ^= cspin[i];   // Projected H at point A
      tmpb = vtmpA[i];
      tmpb ^= tspin[i];   // Projected H at point B
      tmpa += tmpb;
      tmpa *= 0.5;
      vtmpA[i] = tmpa;
    } else {
      // Tighten axis
      tmpa = Torque_0[i];
      REAL8m dott = (axis * tmpa);
      if(fabs(dott)>4*REAL8_EPSILON) {
        tmpa *= dott/tmpa.MagSq();
        axis -= tmpa;
      }

      // Project field onto torque-derived axis
      tmpb = cspin[i]; tmpb ^= axis;
      REAL8m tmpb_magsq_A = tmpb.MagSq();
      REAL8m dot_A = Torque_0[i]*tmpb;

      tmpb = tspin[i]; tmpb ^= axis;
      REAL8m tmpb_magsq_B = tmpb.MagSq();
      REAL8m dot_B = vtmpA[i]*tmpb;

      // Axis direction check
      tmpa = Torque_0[i];
      tmpa ^= cspin[i];
      REAL8m axisdirchecka = tmpa*axis;

      tmpb = vtmpA[i];
      tmpb ^= tspin[i];
      REAL8m axisdircheckb = tmpb*axis;

      if(axis.MagSq() == 0.0 ||
         (tmpb_magsq_A<1.0 && 32*fabs(dot_A)*REAL8_EPSILON>=tmpb_magsq_A) ||
         (tmpb_magsq_B<1.0 && 32*fabs(dot_B)*REAL8_EPSILON>=tmpb_magsq_B) ||
         axisdirchecka<=0.0 || axisdircheckb<=0.0) {
        // Bad data; fall back to torque-based great-circle path.
#ifdef DEEP_DEBUG
        if(CHECK_START <= cstate->iteration_count
           && cstate->iteration_count < CHECK_STOP
           && CHECK_INDEX == i) {
        fprintf(stdout,"Substep pt E\n");
        }
#endif // DEEP_DEBUG
        tmpa = Torque_0[i];
        tmpa ^= cspin[i];   // Projected H at point A
        tmpb = vtmpA[i];
        tmpb ^= tspin[i];   // Projected H at point B
        tmpa += tmpb;
        tmpa *= 0.5;
        vtmpA[i] = tmpa;
      } else {
        // Project fields onto common axis.
#if 0
        vtmpA[i] = axis;
        vtmpA[i] *= 0.5*(dot_A/tmpb_magsq_A + dot_B/tmpb_magsq_B);
#else
        tmpa = cspin[i];
        tmpa ^= axis;
        REAL8m norma = sqrt(Torque_0[i].MagSq()/tmpa.MagSq());

        tmpb = tspin[i];
        tmpb ^= axis;
        REAL8m normb = sqrt(vtmpA[i].MagSq()/tmpb.MagSq());

#ifdef DEEP_DEBUG
        if(CHECK_START <= cstate->iteration_count
           && cstate->iteration_count < CHECK_STOP
           && CHECK_INDEX == i) {
          fprintf(stdout,"Substep pt F\n");
          fprintf(stdout,"-->Told = (%# .15g,%# .15g,%# .15g)\n",
                  Torque_old[i].x,Torque_old[i].y,Torque_old[i].z);
          fprintf(stdout,"-->  T0 = (%# .15g,%# .15g,%# .15g)\n",
                  Torque_0[i].x,Torque_0[i].y,Torque_0[i].z);
          fprintf(stdout,"-->  T1 = (%# .15g,%# .15g,%# .15g)\n",
                  vtmpA[i].x,vtmpA[i].y,vtmpA[i].z);
          fprintf(stdout,"--> axis = (%# .15g,%# .15g,%# .15g)\n",
                  axis.x,axis.y,axis.z);
          fprintf(stdout,"-->cspin = (%18.15f,%18.15f,%18.15f)\n",
                  cspin[i].x,cspin[i].y,cspin[i].z);
          fprintf(stdout,"-->tspin = (%18.15f,%18.15f,%18.15f)\n",
                  tspin[i].x,tspin[i].y,tspin[i].z);
          fprintf(stdout,"-->norma=%#.15g, normb=%#.15g\n",
                  norma,normb);
          fprintf(stdout,"-->ave1: %# .15g\n-->ave2: %# .15g\n",
                  0.5*(norma+normb),
                  0.5*(dot_A/tmpb_magsq_A + dot_B/tmpb_magsq_B));
        }
#endif // DEEP_DEBUG

        vtmpA[i] = axis;
        vtmpA[i] *= 0.5*(norma+normb);
#endif
      }
    }
  }
  tstate = NULL;  // Safety.
  /// NOTE: Neither tstate nor tspin should be used beyond this point.

  // Take step with new effective H
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  AdjustState(stepsize,stepsize,*cstate,vtmpA,*nstate);
#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    fprintf(stdout,"Substep m1b = (%18.15f,%18.15f,%18.15f)\n",
            nstate->spin[CHECK_INDEX].x,nstate->spin[CHECK_INDEX].y,
            nstate->spin[CHECK_INDEX].z);
    fprintf(stdout,"Substep Hef = (%18.15g,%18.15g,%18.15g)\n",
            vtmpA[CHECK_INDEX].x,vtmpA[CHECK_INDEX].y,
            vtmpA[CHECK_INDEX].z);
    ThreeVector foo = vtmpA[CHECK_INDEX];
    foo.MakeUnit();
    fprintf(stdout,"Substep u   = (%18.15f,%18.15f,%18.15f)\n",
            foo.x,foo.y,foo.z);
#if 1
    REAL8m footmp = foo*cstate->spin[CHECK_INDEX];
    fprintf(stdout,"Substep u*m   = %# .15g => %# .5g degrees)\n",
            footmp,acos(footmp)*180./PI);
#endif
    foo = cstate->spin[CHECK_INDEX];
    foo ^= vtmpA[CHECK_INDEX];
    fprintf(stdout,"Substep mxH = (%18.15g,%18.15g,%18.15g)\n",
            foo.x,foo.y,foo.z);
  }
#endif // DEEP_DEBUG
  if(base_state != NULL) {
    UpdateTimeFields(*base_state,*nstate,effective_stepsize);
  }
  next_state_key.GetReadReference(); // Release write lock
}
#endif // TRY1

#ifdef TRY2
void
Oxs_SemiAnalyticEvolve::SemiAnalyticAxisCorrectedSubstep2b
(REAL8m old_stepsize,             // timestep between prev and cstate
 const Oxs_MeshValue<ThreeVector>& m_old,      // Spin at prev state
 const Oxs_MeshValue<ThreeVector>& Torque_old, // Torque at prev state
 REAL8m stepsize,                              // current timestep
 Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
 const Oxs_MeshValue<ThreeVector>&,            // Field at cstate
 const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
 const Oxs_SimState* base_state,               // Used for updating state
 REAL8m effective_stepsize,                   ///    time fields
 Oxs_Key<Oxs_SimState>& next_state_key         // Final state
)
{ // Computes one semi-analytic predictor-corrector step with axis
  // correction, of size stepsize, with no error control.  This routine is
  // intended as a support function for ::SemiAnalyticAxisCorrectedStep2b.
  // In addition to modifying next_state_key, this routine also uses
  // as scratch space vtmpA, vtmpB and temp_energy.  Energy is
  // computed for the predictor step, but *not* at the corrected step,
  // so if that is desired the calling routine should call
  // GetEnergyDensity() on next_state upon return.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& cspin = cstate->spin;


  const UINT4m size = cstate->mesh->Size();

  UINT4m i=0;
  REAL8m pE_pt;  // Dummy value; not used.

  ThreeVector axis,axis_torque,tmpa,tmpb;
  vtmpA.AdjustSize(cstate->mesh);   // Axis 1
  REAL8m lambda = 0.0;
  if(old_stepsize != 0.0) {
    lambda = -0.5*stepsize/old_stepsize;
  }
  for(i=0;i<size;++i) {
    axis = Torque_old[i];
    axis ^= Torque_0[i];

    const double slip = 1e-4;
    axis_torque = Torque_0[i]; axis_torque ^= cspin[i];
    axis_torque *= slip*sqrt(Torque_0[i].MagSq());

    axis += axis_torque;

    // Project field onto torque-derived axis
    tmpa = m_old[i]; tmpa ^= axis;  REAL8m tmpa_magsq = tmpa.MagSq();
    REAL8m dota = Torque_old[i]*tmpa;

    tmpb = cspin[i]; tmpb ^= axis;  REAL8m tmpb_magsq = tmpb.MagSq();
    REAL8m dotb = Torque_0[i]*tmpb;

    if(dota*(4*REAL8_EPSILON)>tmpa_magsq) {
      // Safety
      vtmpA[i] = Torque_0[i];
      vtmpA[i] ^= cspin[i];
    } else {
      REAL8m wgta = dota/tmpa_magsq;
      REAL8m wgtb = dotb/tmpb_magsq;
      vtmpA[i] = axis;
      vtmpA[i] *= lambda*wgta + (1-lambda)*wgtb;
    }
  }
  // vtmpA is axis projected value of H at point A

  // Move ahead one step
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              next_state_key.GetWriteReference());
  const Oxs_SimState* tstate = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;
  GetEnergyDensity(*tstate,temp_energy,&vtmpA,NULL,pE_pt);

  ZeroFixedSpinFields(&vtmpA,NULL);
  /// vtmpA holds mxH at point B

  // Update axis and compute average projected field
  for(i=0;i<size;++i) {
    axis = Torque_0[i];
    axis ^= vtmpA[i];

    const double slip = 1e-4;
    axis_torque = Torque_0[i]; axis_torque ^= cspin[i];
    axis_torque *= slip*sqrt(vtmpA[i].MagSq());

    axis += axis_torque;

    // Project field onto torque-derived axis
    tmpa = cspin[i]; tmpa ^= axis;  REAL8m tmpa_magsq = tmpa.MagSq();
    REAL8m dota = Torque_0[i]*tmpa;

    tmpb = tspin[i]; tmpb ^= axis;  REAL8m tmpb_magsq = tmpb.MagSq();
    REAL8m dotb = vtmpA[i]*tmpb;

    if(dotb*(4*REAL8_EPSILON)>tmpb_magsq) {
      // Safety
      vtmpA[i] = Torque_0[i];
      vtmpA[i] ^= cspin[i];
    } else {
      REAL8m wgta = dota/tmpa_magsq;
      REAL8m wgtb = dotb/tmpb_magsq;
      vtmpA[i] = axis;
      vtmpA[i] *= 0.5*(wgta + wgtb);
    }
  }
  tstate = NULL;  // Safety.
  /// NOTE: Neither tstate nor tspin should be used beyond this point.

  // Take step with new effective H
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  AdjustState(stepsize,stepsize,*cstate,vtmpA,*nstate);
  if(base_state != NULL) {
    UpdateTimeFields(*base_state,*nstate,effective_stepsize);
  }
  next_state_key.GetReadReference(); // Release write lock
}
#endif // TRY2

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2b
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Takes a simple semi-analytic step
  // NOTE: It is the responsibility of the calling routine to
  //       handle the setting and updating of mxH_old_cached.

  // Do we have valid old torque?
  if(!mxH_old_cached_is_valid) {
    // No.  Fall back to Torque Step2b
    TakeSemiAnalyticTorqueStep2b(stepsize,current_state_key,
                                 H,mxH,next_state_key,
                                 error_estimate,global_error_order,
                                 new_energy_and_field_computed);
    return;
  }

  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  next_state_key.GetReadReference().CloneHeader
    (temp_state_key.GetWriteReference());
  temp_state_key.GetReadReference(); // Release write lock

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  // Move ahead two half steps

  // First half-step
  SemiAnalyticAxisCorrectedSubstep2b(
               old_timestep_cached,m_old_cached,mxH_old_cached,
               stepsize/2,current_state_key,H,mxH,NULL,0,temp_state_key);
  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,&vtmpD,&vtmpE,pE_pt);
  ZeroFixedSpinFields(&vtmpD,&vtmpE);
  /// vtmpD holds mxH value at midpoint, vtmpE holds H value at midpoint

  // Second half-step
  SemiAnalyticAxisCorrectedSubstep2b(stepsize/2,cstate->spin,mxH,
      stepsize/2,temp_state_key,vtmpE,vtmpD,cstate,stepsize,next_state_key);

#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    fprintf(stdout,"--- Iteration %5d ---\n",cstate->iteration_count);
    fprintf(stdout,"Timestep: %g\n",stepsize);
    fprintf(stdout,"m0     = (%18.15f,%18.15f,%18.15f)\n",
            cstate->spin[CHECK_INDEX].x,cstate->spin[CHECK_INDEX].y,
            cstate->spin[CHECK_INDEX].z);
    const Oxs_SimState* checkstate = &(temp_state_key.GetReadReference());
    fprintf(stdout,"m_mid  = (%18.15f,%18.15f,%18.15f)\n",
            checkstate->spin[CHECK_INDEX].x,checkstate->spin[CHECK_INDEX].y,
            checkstate->spin[CHECK_INDEX].z);
    checkstate = &(next_state_key.GetReadReference());
    fprintf(stdout,"m_end  = (%18.15f,%18.15f,%18.15f)\n",
            checkstate->spin[CHECK_INDEX].x,checkstate->spin[CHECK_INDEX].y,
            checkstate->spin[CHECK_INDEX].z);
  }
#endif // DEEP_DEBUG

  // For comparison, move forward one full step
  SemiAnalyticAxisCorrectedSubstep2b(
               old_timestep_cached,m_old_cached,mxH_old_cached,
               stepsize,current_state_key,H,mxH,NULL,0,temp_state_key);

#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    const Oxs_SimState* checkstate = &(temp_state_key.GetReadReference());
    fprintf(stdout,"m_comp = (%18.15f,%18.15f,%18.15f)\n",
            checkstate->spin[CHECK_INDEX].x,checkstate->spin[CHECK_INDEX].y,
            checkstate->spin[CHECK_INDEX].z);
    fprintf(stdout,"mxH*t  = (%10.7g,%10.7g,%10.7g)\n",
            mxH[CHECK_INDEX].x*stepsize,
            mxH[CHECK_INDEX].y*stepsize,
            mxH[CHECK_INDEX].z*stepsize);
    fflush(stdout);
    
  }
#endif // DEEP_DEBUG

  // Auxiliary pointers to endpoints
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& endspin = endstate->spin;

  const Oxs_SimState* tstate
    = &(temp_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;

  // Error estimate is max|endspin-tspin| + O(stepsize^4) (?)
  REAL8m max_error_sq = 0.0;
  const UINT4m size = cstate->mesh->Size();
  if(!debug_step_error
     || (int)cstate->iteration_count < debug_step_error_start
     || (int)cstate->iteration_count >= debug_step_error_stop) {
    for(UINT4m i=0;i<size;++i) {
      ThreeVector tvec = tspin[i] - endspin[i];
      REAL8m error_sq = tvec.MagSq();
      if(error_sq>max_error_sq) max_error_sq = error_sq;
#ifdef DEEP_DEBUG
      if(error_sq>=1.0) {
        fprintf(stdout,"****** Iteration %6d, Index %5d, error=%g ******\n",
                (int)cstate->iteration_count,i,sqrt(error_sq));
      }
#endif // DEEP_DEBUG
    }
  } else {
    vtmpF.AdjustSize(cstate->mesh);
    for(UINT4m i=0;i<size;++i) {
      ThreeVector tvec = tspin[i] - endspin[i];
      REAL8m error_sq = tvec.MagSq();
      if(error_sq>max_error_sq) max_error_sq = error_sq;
#ifdef DEEP_DEBUG
      if(error_sq>1.0) {
        fprintf(stdout,"****** Index %5d, error=%g ******\n",i,sqrt(error_sq));
      }
#endif // DEEP_DEBUG
      vtmpF[i] = tvec;
    }
    // Dump vtmpF to disk
    if(debug_step_error_iteration_count != (int)cstate->iteration_count) {
      debug_step_error_iteration_count = (int)cstate->iteration_count;
      debug_step_error_iteration_subcount = 0;
    } else {
      ++debug_step_error_iteration_subcount;
    }
    const int namebufsize = debug_step_error_basename.length()+256;
    char* namebuf;
    namebuf = new char[namebufsize];
    Oc_Snprintf(namebuf,namebufsize,
                "%s-%06d-%02d.ovf",
                debug_step_error_basename.c_str(),
                debug_step_error_iteration_count,
                debug_step_error_iteration_subcount);
    String title = "";
    String desc = "";
    String units = "";
    String datatype = "binary";
    String precision = "8";
    cstate->mesh->WriteOvf(namebuf,1,
                           title.c_str(),
                           desc.c_str(),
                           units.c_str(),
                           cstate->mesh->NaturalOutputType(),
                           datatype.c_str(),
                           precision.c_str(),
                           &vtmpF,NULL);
    delete[] namebuf;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 2.0;

  // Compute energy and field at *real* endpoint
  GetEnergyDensity(*endstate,
                   temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2b:"
       " Programming error; data cache (pE/pt) already set.");
  }

  new_energy_and_field_computed = 1;

  // Note that end energy is in temp_energy, mxH is in vtmpA, and H is
  // in vtmpB as per contract.
}

void
Oxs_SemiAnalyticEvolve::ComputeAxisCorrectedH
(REAL8m stepsize,REAL8m lambda,
 REAL8m gamma,REAL8m rotation_limit,
 REAL8m stability_adj,
 const ThreeVector& m0, const ThreeVector& T0,
 const ThreeVector& m1, const ThreeVector& T1,
 ThreeVector& acH,ThreeVector& acT0,ThreeVector& acT1)
{
  ThreeVector tmpv1 = m0;
  ThreeVector tmpv2 = m1;

  ThreeVector tmpr1 = T0;  tmpr1 ^= tmpv1;
  ThreeVector tmpr2 = T1;  tmpr2 ^= tmpv2;

  ThreeVector tmpv1xv2 = tmpv1; tmpv1xv2 ^= tmpv2;

  ThreeVector tmpv1xv2xdr = tmpr1;
  tmpv1xv2xdr -= tmpr2;
  tmpv1xv2xdr ^= tmpv1xv2;

  // Compute A = (m1*m1)(m2*m2)-(m1*m2)^2
  REAL8m A = tmpv1xv2.MagSq();
  REAL8m B = tmpv1xv2xdr*tmpv2;
  REAL8m C = tmpv1xv2xdr*tmpv1;

  A += stability_adj;  // Adds stability

  REAL8m s=0,t=0;

  // Don't allow s or t to increase adjusted H beyond rotation limit.
  // (Assume here that |m|=1, and ignore contribution from mxHxm.)
  REAL8m step_check = fabs(gamma)*stepsize/(2*PI*rotation_limit);

  if(step_check>0.0) {
    step_check = 1.0/step_check;
    if(B != 0.0) {
      if(fabs(B)<A*step_check) {
        s = B/A;
      } else {
        s = step_check;
        if(B<0) s *= -1;
      }
    }
    if(C != 0.0) {
      if(fabs(C)<A*step_check) {
        t = C/A;
      } else {
        t = step_check;
        if(C<0) t *= -1;
      }
    }
  }

  tmpr1 -= tmpr2;
  tmpr1 *= lambda;
  tmpr1 += tmpr2;

  acT0 = tmpv1;
  acT0 ^= tmpr1;
  acT0.Accum(t*lambda,tmpv1xv2);   // acT0 = m0 x Hadjusted

  acT1 = tmpv2;
  acT1 ^= tmpr1;
  acT1.Accum(-s*lambda,tmpv1xv2);  // acT1 = m1 x Hadjusted

  tmpv1 *= s;
  tmpv2 *= t;
  tmpv1 -= tmpv2;
  tmpv1 *= lambda;
  tmpv1 += tmpv2;

  tmpr1 += tmpv1;

  acH = tmpr1;
}

void
Oxs_SemiAnalyticEvolve::SemiAnalyticAxisCorrectedSubstep2c
(REAL8m old_stepsize,             // timestep between prev and cstate
 const Oxs_MeshValue<ThreeVector>& m_old,      // Spin at prev state
 const Oxs_MeshValue<ThreeVector>& Torque_old, // Torque at prev state
 REAL8m stepsize,                              // current timestep
 Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
#ifdef DEEP_DEBUG
 const Oxs_MeshValue<ThreeVector>& /* H_0 */,        // Field at cstate
#else
 const Oxs_MeshValue<ThreeVector>&,            // Field at cstate
#endif
 const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
 const Oxs_SimState* base_state,               // Used for updating state
 REAL8m effective_stepsize,                   ///    time fields
 Oxs_Key<Oxs_SimState>& next_state_key,        // Final state
 REAL8m rotation_limit       // Allowed rotation, in fraction
 /// of a complete revolution, e.g., 0.5 => max half rotation.
)
{ // Computes one semi-analytic predictor-corrector step with axis
  // correction, of size stepsize, with no error control.  This routine is
  // intended as a support function for ::SemiAnalyticAxisCorrectedStep2c.
  // In addition to modifying next_state_key, this routine also uses
  // as scratch space vtmpA, vtmpB and temp_energy.  Energy is
  // computed for the predictor step, but *not* at the corrected step,
  // so if that is desired the calling routine should call
  // GetEnergyDensity() on next_state upon return.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& cspin = cstate->spin;


#ifndef NDEBUG
  // Consistency checks
  if(!gamma.CheckMesh(cstate->mesh)) {
    throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::SemiAnalyticAxisCorrectedSubstep2c:"
           " gamma not properly initialized.");
  }
#endif // NDEBUG


  const UINT4m size = cstate->mesh->Size();

  UINT4m i=0;
  REAL8m pE_pt;  // Dummy value; not used.

  ThreeVector tmpr1,tmpr2,tmpv1,tmpv2,tmpv1xv2,tmpv1xv2xdr;
  ThreeVector tmpa,tmpb;
  ThreeVector acH,acT0,acT1;
  vtmpA.AdjustSize(cstate->mesh);
  vtmpB.AdjustSize(cstate->mesh);
  REAL8m lambda = 0.0;
  if(old_stepsize != 0.0) {
    lambda = -0.5*stepsize/old_stepsize;
  }
  for(i=0;i<size;++i) {
    ComputeAxisCorrectedH(stepsize,lambda,gamma[i],
                          rotation_limit,1e-4,
                          m_old[i],Torque_old[i],
                          cspin[i],Torque_0[i],
                          acH,acT0,acT1);
    vtmpA[i] = acH;
    vtmpB[i] = acT1;
  }
  // vtmpA is axis projected value of H at point A.
  // vtmpB holds mxH.

  // Move ahead one step
  AdjustState(stepsize,stepsize,*cstate,vtmpA,vtmpB,
              next_state_key.GetWriteReference());
  const Oxs_SimState* tstate = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;
  GetEnergyDensity(*tstate,temp_energy,&vtmpA,NULL,pE_pt);

  ZeroFixedSpinFields(&vtmpA,NULL);
  /// vtmpA holds mxH at point B

  // Update axis and compute average projected field
  for(i=0;i<size;++i) {
    ComputeAxisCorrectedH(stepsize,0.5,gamma[i],
                          rotation_limit,1e-4,
                          cspin[i],Torque_0[i],
                          tspin[i],vtmpA[i],
                          acH,acT0,acT1);
    vtmpA[i] = acH;
    vtmpB[i] = acT0;
  }
  // vtmpA is axis projected value of H at point B.
  // vtmpB holds mxH.

  tstate = NULL;  // Safety.
  /// NOTE: Neither tstate nor tspin should be used beyond this point.

  // Take step with new effective H
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  AdjustState(stepsize,stepsize,*cstate,vtmpA,vtmpB,*nstate);
  if(base_state != NULL) {
    UpdateTimeFields(*base_state,*nstate,effective_stepsize);
  }
  next_state_key.GetReadReference(); // Release write lock
}

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2c
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Takes a simple semi-analytic step
  // NOTE: It is the responsibility of the calling routine to
  //       handle the setting and updating of mxH_old_cached.

  // Do we have valid old torque?
  if(!mxH_old_cached_is_valid) {
    // No.  Fall back to Torque Step2b
    TakeSemiAnalyticTorqueStep2b(stepsize,current_state_key,
                                 H,mxH,next_state_key,
                                 error_estimate,global_error_order,
                                 new_energy_and_field_computed);
    return;
  }

  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    fprintf(stdout,"--- Iteration %5d ---\n",cstate->iteration_count);
    fprintf(stdout,"Timestep: %g\n",stepsize);
  }
#endif // DEEP_DEBUG

  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  next_state_key.GetReadReference().CloneHeader
    (temp_state_key.GetWriteReference());
  temp_state_key.GetReadReference(); // Release write lock

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  // Move ahead two half steps

  // First half-step
  SemiAnalyticAxisCorrectedSubstep2c(
               old_timestep_cached,m_old_cached,mxH_old_cached,
               stepsize/2,current_state_key,H,mxH,NULL,0,
               temp_state_key,0.25);
  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,&vtmpD,&vtmpE,pE_pt);
  ZeroFixedSpinFields(&vtmpD,&vtmpE);
  /// vtmpD holds mxH value at midpoint, vtmpE holds H value at midpoint

  // Second half-step
  SemiAnalyticAxisCorrectedSubstep2c(
               stepsize/2,cstate->spin,mxH,stepsize/2,
               temp_state_key,vtmpE,vtmpD,cstate,stepsize,
               next_state_key,0.25);

#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    fprintf(stdout,"m0     = (%18.15f,%18.15f,%18.15f)\n",
            cstate->spin[CHECK_INDEX].x,cstate->spin[CHECK_INDEX].y,
            cstate->spin[CHECK_INDEX].z);
    const Oxs_SimState* checkstate = &(temp_state_key.GetReadReference());
    fprintf(stdout,"m_mid  = (%18.15f,%18.15f,%18.15f)\n",
            checkstate->spin[CHECK_INDEX].x,checkstate->spin[CHECK_INDEX].y,
            checkstate->spin[CHECK_INDEX].z);
    checkstate = &(next_state_key.GetReadReference());
    fprintf(stdout,"m_end  = (%18.15f,%18.15f,%18.15f)\n",
            checkstate->spin[CHECK_INDEX].x,checkstate->spin[CHECK_INDEX].y,
            checkstate->spin[CHECK_INDEX].z);
  }
#endif // DEEP_DEBUG

  // For comparison, move forward one full step
  SemiAnalyticAxisCorrectedSubstep2c(
               old_timestep_cached,m_old_cached,mxH_old_cached,
               stepsize,current_state_key,H,mxH,NULL,0,
               temp_state_key,0.5);

#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    const Oxs_SimState* checkstate = &(temp_state_key.GetReadReference());
    fprintf(stdout,"m_comp = (%18.15f,%18.15f,%18.15f)\n",
            checkstate->spin[CHECK_INDEX].x,checkstate->spin[CHECK_INDEX].y,
            checkstate->spin[CHECK_INDEX].z);
    fprintf(stdout,"|gamma|*Hxm*t  = (%10.7g,%10.7g,%10.7g)\n",
            -2.211e5*mxH[CHECK_INDEX].x*stepsize,
            -2.211e5*mxH[CHECK_INDEX].y*stepsize,
            -2.211e5*mxH[CHECK_INDEX].z*stepsize);
    fflush(stdout);
    
  }
#endif // DEEP_DEBUG

  // Auxiliary pointers to endpoints
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& endspin = endstate->spin;

  const Oxs_SimState* tstate
    = &(temp_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;

  // Error estimate is max|endspin-tspin| + O(stepsize^4) (?)
#ifdef DEEP_DEBUG
  UINT4m bad_i = 0;
#endif
  REAL8m max_error_sq = 0.0;
  const UINT4m size = cstate->mesh->Size();
  if(!debug_step_error
     || (int)cstate->iteration_count < debug_step_error_start
     || (int)cstate->iteration_count >= debug_step_error_stop) {
    for(UINT4m i=0;i<size;++i) {
      ThreeVector tvec = tspin[i] - endspin[i];
      REAL8m error_sq = tvec.MagSq();
      if(error_sq>max_error_sq) {
        max_error_sq = error_sq;
#ifdef DEEP_DEBUG
        bad_i = i;
#endif // DEEP_DEBUG
      }
#ifdef DEEP_DEBUG
      if(CHECK_START <= cstate->iteration_count
         && cstate->iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        fprintf(stdout,"  Error[%u] = |(%18.15f, %18.15f, %18.15f)| = %#.3g\n",
                (unsigned int)i,tvec.x,tvec.y,tvec.z,sqrt(error_sq));
      }
#endif // DEEP_DEBUG
    }
  } else {
    vtmpF.AdjustSize(cstate->mesh);
    for(UINT4m i=0;i<size;++i) {
      ThreeVector tvec = tspin[i] - endspin[i];
      REAL8m error_sq = tvec.MagSq();
      if(error_sq>max_error_sq) {
        max_error_sq = error_sq;
#ifdef DEEP_DEBUG
        bad_i = i;
#endif // DEEP_DEBUG
      }
#ifdef DEEP_DEBUG
      if(CHECK_START <= cstate->iteration_count
         && cstate->iteration_count < CHECK_STOP
         && CHECK_INDEX == i) {
        fprintf(stdout,"  Error[%u] = |(%18.15f, %18.15f, %18.15f)| = %#.3g\n",
                (unsigned int)i,tvec.x,tvec.y,tvec.z,sqrt(error_sq));
      }
#endif // DEEP_DEBUG
      vtmpF[i] = tvec;
    }
    // Dump vtmpF to disk
    if(debug_step_error_iteration_count != (int)cstate->iteration_count) {
      debug_step_error_iteration_count = (int)cstate->iteration_count;
      debug_step_error_iteration_subcount = 0;
    } else {
      ++debug_step_error_iteration_subcount;
    }
    const int namebufsize = debug_step_error_basename.length()+256;
    char* namebuf;
    namebuf = new char[namebufsize];
    Oc_Snprintf(namebuf,namebufsize,
                "%s-%06d-%02d.ovf",
                debug_step_error_basename.c_str(),
                debug_step_error_iteration_count,
                debug_step_error_iteration_subcount);
    String title = "";
    String desc = "";
    String units = "";
    String datatype = "binary";
    String precision = "8";
    cstate->mesh->WriteOvf(namebuf,1,
                           title.c_str(),
                           desc.c_str(),
                           units.c_str(),
                           cstate->mesh->NaturalOutputType(),
                           datatype.c_str(),
                           precision.c_str(),
                           &vtmpF,NULL);
    delete[] namebuf;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 2.0;

#ifdef DEEP_DEBUG
  if(CHECK_START <= cstate->iteration_count
     && cstate->iteration_count < CHECK_STOP) {
    fprintf(stdout," Max Error[%4u] = %#.3g\n",
                  (unsigned int)bad_i,error_estimate);
  }
#endif // DEEP_DEBUG

  // Compute energy and field at *real* endpoint
  GetEnergyDensity(*endstate,
                   temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticAxisCorrectedStep2c:"
       " Programming error; data cache (pE/pt) already set.");
  }

  new_energy_and_field_computed = 1;

  // Note that end energy is in temp_energy, mxH is in vtmpA, and H is
  // in vtmpB as per contract.
}

void
Oxs_SemiAnalyticEvolve::SemiAnalyticTorqueSubstep2
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial state
 const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
 const Oxs_SimState* base_state,               // Used for updating state
 REAL8m effective_stepsize,                   /// time fields
 Oxs_Key<Oxs_SimState>& next_state_key         // Final state
)
{ // Computes one semi-analytic predictor-corrector step based on torque
  // values, of size stepsize, with no error control.  This routine is
  // intended as a support function for ::SemiAnalyticTorqueStep2.
  // In addition to modifying next_state_key, this routine also uses
  // as scratch space vtmpA and temp_energy.  Energy is
  // computed for the predictor step, but *not* at the corrected step,
  // so if that is desired the calling routine should call
  // GetEnergyDensity() on next_state upon return.

  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& cspin = cstate->spin;
  const UINT4m size = cstate->mesh->Size();

  UINT4m i=0;
  REAL8m pE_pt;  // Dummy value; not used.

  ThreeVector axis,tmpa,tmpb;
  vtmpA.AdjustSize(cstate->mesh);   // Axis 1
  for(i=0;i<size;++i) {
    vtmpA[i]  = Torque_0[i];
    vtmpA[i] ^= cspin[i];   // mxHxm
  }

  // Move ahead one step
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              next_state_key.GetWriteReference());
  const Oxs_SimState* tstate = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;
  GetEnergyDensity(*tstate,temp_energy,&vtmpA,NULL,pE_pt);

  ZeroFixedSpinFields(&vtmpA,NULL);
  /// vtmpA holds mxH at point B

  // Update axis and compute average projected field
  for(i=0;i<size;++i) {
    tmpa = Torque_0[i];
    tmpa ^= cspin[i];   // Projected torque at point A
    tmpb = vtmpA[i];
    tmpb ^= tspin[i];   // Projected torque at point B
    tmpa += tmpb;
    tmpa *= 0.5;
    vtmpA[i] = tmpa;
  }
  tstate = NULL;  // Safety.
  /// NOTE: Neither tstate nor tspin should be used beyond this point.

  // Take step with new effective H
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  AdjustState(stepsize,stepsize,*cstate,vtmpA,*nstate);
  if(base_state != NULL) {
    UpdateTimeFields(*base_state,*nstate,effective_stepsize);
  }
  next_state_key.GetReadReference(); // Release write lock
}

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticTorqueStep2
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& /* H */,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Takes a second-order semi-analytic step based on torques
  // NOTE: It is the responsibility of the calling routine to
  //       handle the setting and updating of mxH_old_cached.

  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  next_state_key.GetReadReference().CloneHeader
    (temp_state_key.GetWriteReference());
  temp_state_key.GetReadReference(); // Release write lock

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  // Move ahead two half steps

  // First half-step
  SemiAnalyticTorqueSubstep2(stepsize/2,
        current_state_key,mxH,NULL,0,temp_state_key);
  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,&vtmpD,NULL,pE_pt);
  ZeroFixedSpinFields(&vtmpD,NULL);
  /// vtmpD holds mxH value at midpoint

  // Second half-step
  SemiAnalyticTorqueSubstep2(stepsize/2,
      temp_state_key,vtmpD,cstate,stepsize,next_state_key);

  // For comparison, move forward one full step
  SemiAnalyticTorqueSubstep2(stepsize,
      current_state_key,mxH,NULL,0,temp_state_key);

  // Auxiliary pointers to endpoints
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& endspin = endstate->spin;

  const Oxs_SimState* tstate
    = &(temp_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;

  // Error estimate is max|endspin-tspin| + O(stepsize^4) (?)
  REAL8m max_error_sq = 0.0;
  const UINT4m size = cstate->mesh->Size();
  for(UINT4m i=0;i<size;++i) {
    ThreeVector tvec = endspin[i] - tspin[i];
    REAL8m error_sq = tvec.MagSq();
    if(error_sq>max_error_sq) max_error_sq = error_sq;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 2.0;

  // Compute energy and field at *real* endpoint
  GetEnergyDensity(*endstate,
                   temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticTorqueStep2:"
       " Programming error; data cache (pE/pt) already set.");
  }

  new_energy_and_field_computed = 1;

  // Note that end energy is in temp_energy, mxH is in vtmpA, and H is
  // in vtmpB as per contract.
}

void
Oxs_SemiAnalyticEvolve::SemiAnalyticTorqueSubstep2b
(REAL8m old_stepsize,             // timestep between prev and cstate
 const Oxs_MeshValue<ThreeVector>& old_mxHxm,  // mxHxm at prev state
 REAL8m stepsize,                              // Current timestep
 Oxs_ConstKey<Oxs_SimState> current_state_key, // Initial (current) state
 const Oxs_MeshValue<ThreeVector>& Torque_0,   // Torque at cstate
 const Oxs_SimState* base_state,               // Used for updating state
 REAL8m effective_stepsize,                   /// time fields
 Oxs_Key<Oxs_SimState>& next_state_key         // Final state
)
{ // Computes one semi-analytic predictor-corrector step based on torque
  // values, of size stepsize, with no error control.  This routine is
  // intended as a support function for ::SemiAnalyticTorqueStep2.
  // In addition to modifying next_state_key, this routine also uses
  // as scratch space vtmpA and temp_energy.  Energy is
  // computed for the predictor step, but *not* at the corrected step,
  // so if that is desired the calling routine should call
  // GetEnergyDensity() on next_state upon return.

  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& cspin = cstate->spin;
  const UINT4m size = cstate->mesh->Size();

  if(old_mxHxm.Size() != size) {
    throw Oxs_Ext::Error(this,
          "Oxs_SemiAnalyticEvolve::SemiAnalyticTorqueSubstep2b:"
          " Imports old_mxHxm and cstate are different sizes.");
  }

  UINT4m i=0;
  REAL8m pE_pt;  // Dummy value; not used.

  ThreeVector axis,tmpa,tmpb;
  vtmpA.AdjustSize(cstate->mesh);   // Axis 1
  REAL8m lambda = 0.0;
  if(old_stepsize != 0.0) {
    lambda = -0.5*stepsize/old_stepsize;
  }
  for(i=0;i<size;++i) {
    tmpa = Torque_0[i];
    tmpa ^= cspin[i];
    tmpa *= (1-lambda);
    tmpa.Accum(lambda,old_mxHxm[i]);
    vtmpA[i] = tmpa;
  }

  // Move ahead one step
  AdjustState(stepsize,stepsize,*cstate,vtmpA,
              next_state_key.GetWriteReference());
  const Oxs_SimState* tstate = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;
  GetEnergyDensity(*tstate,temp_energy,&vtmpA,NULL,pE_pt);

  ZeroFixedSpinFields(&vtmpA,NULL);
  /// vtmpA holds mxH at point B

  // Update axis and compute average projected field
  for(i=0;i<size;++i) {
    tmpa = Torque_0[i];
    tmpa ^= cspin[i];   // Projected torque at point A
    tmpb = vtmpA[i];
    tmpb ^= tspin[i];   // Projected torque at point B
    tmpa += tmpb;
    tmpa *= 0.5;
    vtmpA[i] = tmpa;
  }
  tstate = NULL;  // Safety.
  /// NOTE: Neither tstate nor tspin should be used beyond this point.

  // Take step with new effective H
  Oxs_SimState* nstate = &(next_state_key.GetWriteReference());
  AdjustState(stepsize,stepsize,*cstate,vtmpA,*nstate);
  if(base_state != NULL) {
    UpdateTimeFields(*base_state,*nstate,effective_stepsize);
  }
  next_state_key.GetReadReference(); // Release write lock
}

void Oxs_SemiAnalyticEvolve::TakeSemiAnalyticTorqueStep2b
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& H,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Takes a second-order semi-analytic step based on torques
  // NOTE: It is the responsibility of the calling routine to
  //       handle the setting and updating of mxH_old_cached.

  if(!mxH_old_cached_is_valid) {
    TakeSemiAnalyticTorqueStep2(stepsize,
                                current_state_key,H,mxH,next_state_key,
                                error_estimate,global_error_order,
                                new_energy_and_field_computed);
    return;
  }


  // Any locks arising from temp_state_key will be released when
  // temp_state_key is destroyed on method exit.
  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  Oxs_Key<Oxs_SimState> temp_state_key;
  director->GetNewSimulationState(temp_state_key);
  next_state_key.GetReadReference().CloneHeader
    (temp_state_key.GetWriteReference());
  temp_state_key.GetReadReference(); // Release write lock

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  REAL8m pE_pt;

  UINT4m i;
  const UINT4m size = cstate->mesh->Size();

  // Move ahead two half steps

  // First half-step
  SemiAnalyticTorqueSubstep2b(old_timestep_cached,mxHxm_old_cached,
                              stepsize/2,current_state_key,mxH,NULL,
                              0,temp_state_key);
  GetEnergyDensity(temp_state_key.GetReadReference(),
                   temp_energy,&vtmpD,NULL,pE_pt);
  ZeroFixedSpinFields(&vtmpD,NULL);
  /// vtmpD holds mxH value at midpoint

  // Second half-step
  vtmpE.AdjustSize(cstate->mesh);  // Scratch space for mxHxm at cstate
  for(i=0;i<size;++i) {
    vtmpE[i] = mxH[i];
    vtmpE[i] ^= cstate->spin[i];
  }
  SemiAnalyticTorqueSubstep2b(stepsize/2,vtmpE,
                             stepsize/2,temp_state_key,vtmpD,cstate,
                             stepsize,next_state_key);

  // For comparison, move forward one full step
  SemiAnalyticTorqueSubstep2b(old_timestep_cached,mxHxm_old_cached,
                              stepsize,current_state_key,mxH,NULL,
                              0,temp_state_key);

  // Auxiliary pointers to endpoints
  const Oxs_SimState* endstate
    = &(next_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& endspin = endstate->spin;

  const Oxs_SimState* tstate
    = &(temp_state_key.GetReadReference());
  const Oxs_MeshValue<ThreeVector>& tspin = tstate->spin;

  // Error estimate is max|endspin-tspin| + O(stepsize^4) (?)
  REAL8m max_error_sq = 0.0;
  for(i=0;i<size;++i) {
    ThreeVector tvec = endspin[i] - tspin[i];
    REAL8m error_sq = tvec.MagSq();
    if(error_sq>max_error_sq) max_error_sq = error_sq;
  }
  error_estimate = sqrt(max_error_sq);
  global_error_order = 2.0;

  // Compute energy and field at *real* endpoint
  GetEnergyDensity(*endstate,
                   temp_energy,&vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::TakeSemiAnalyticTorqueStep2:"
       " Programming error; data cache (pE/pt) already set.");
  }

  new_energy_and_field_computed = 1;

  // Note that end energy is in temp_energy, mxH is in vtmpA, and H is
  // in vtmpB as per contract.
}

void
Oxs_SemiAnalyticEvolve::RungeKuttaFehlbergBase54_Substep
(REAL8m stepsize,
 const Oxs_SimState* cstate,
 Oxs_Key<Oxs_SimState>& next_state_key,
 const Oxs_MeshValue<ThreeVector>* mxH_ptr,
 Oxs_MeshValue<ThreeVector>* H_ptr,
 int N,      // "N" is 1-based point number
 REAL8m aN,  // Actual timestep is aN*stepsize
 REAL8m* bN, // Array of length N (0-based)
 REAL8m& pE_pt)
{
  // On import, next_state should hold state mN in its spin array.
  // mxH_ptr should either point to the mxH data for mN, or else be
  // NULL.  *IF* (and only if) mxH_ptr is NULL, then GetEnergyDensity
  // is called using next_state with the following side effects:
  //   vtmpA is filled with mxH data.
  //   If H_ptr != NULL, then *H_ptr will be filled with H data.
  //   temp_energy is filled
  //   pE_pt is set.
  //
  // Next, mxH (or vtmpA) is used to compute dmN_dt, which is then
  // converted spherical coords and stored in rkf54_data.
  //
  // Finally, if N<7, then m{N+1} is computed, and the new state
  // is left in the next_state spin array.  If N==7, then next_state
  // is not touched.
  //
  // Note that this routine uses vtmpA and temp_energy iff
  // mxH_ptr == NULL.

  if(mxH_ptr == NULL) {
    GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                     &vtmpA,H_ptr,pE_pt);
    ZeroFixedSpinFields(&vtmpA,H_ptr);
    // vtmpA holds mxH at mN
    mxH_ptr = &vtmpA;
  }
  const Oxs_MeshValue<ThreeVector>& mxH = *mxH_ptr;

  const UINT4m size = cstate->mesh->Size();
  const Oxs_MeshValue<ThreeVector>& m1 = cstate->spin;
  const Oxs_MeshValue<REAL8m>& Ms = *(cstate->Ms);

  Oxs_SimState* nstate = NULL;
  const Oxs_MeshValue<ThreeVector>* nspin_ptr = NULL;
  if(N<7) {
    nstate = &(next_state_key.GetWriteReference());
    nspin_ptr = &(nstate->spin);
  } else {
    nspin_ptr = &(next_state_key.GetReadReference().spin);
  }
  const Oxs_MeshValue<ThreeVector>& mN_spin = *nspin_ptr;


  if(!(mN_spin.CheckMesh(cstate->mesh))) {
    // Safety
    throw Oxs_Ext::Error
      (this,
       "Oxs_Oxs_SemiAnalyticEvolve::RungeKuttaFehlbergBase54_Substep:"
       " Programming error; next_state as invalid spin array.");
  }

  // Fill data.dtheta[N-1] and data.dphi[N-1], and if
  // N<7 then fill nstate with m{N+1}
  REAL8m max_delta_phi = 0.0; /**/ // asdf
  UINT4m max_delta_phi_i = 0; /**/ // asdf
  for(UINT4m i=0;i<size;i++) {
    RKF54_dataobj& data = rkf54_data[i];
    if(Ms[i]==0.0) {
      data.dtheta[N-1] = data.dphi[N-1] = 0.0;
      if(N<7) nstate->spin[i] = m1[i];
      continue;
    }

    // Compute dmN_dt, and convert to spherical coords
    ThreeVector dmN_dt;
    const ThreeVector& mN = mN_spin[i];
    Compute_dm_dt(mN,mxH[i],gamma[i],alpha[i],dmN_dt);
    ThreeVector phivec = data.ez;
    phivec ^= mN;
    REAL8m sinthetasq = phivec.MagSq();
    if(sinthetasq == 0.0) {
      // m is aligned with ez.  What to do???
      data.dtheta[N-1] = sqrt(dmN_dt.MagSq());  // I guess?
      data.dphi[N-1] = 0.0;  // I don't know!
    } else {
      // Warning: overflow possible
      data.dtheta[N-1] = -1*data.ez*dmN_dt/sqrt(sinthetasq);
      data.dphi[N-1] = phivec*dmN_dt/sinthetasq;
    }

    if(N<7) {
      // Fill m{N+1} = m1+h*kN = m1 + b21*h*dm_dt1 + b22*h*dm_dt2
      // where kN = \sum_{M=1}^N dm_dtN*bNM
      REAL8m thetaNb = 0.0;
      REAL8m phiNb = 0.0;
      for(int j=0;j<N;j++) {
        thetaNb += data.dtheta[j]*bN[j];
        phiNb   += data.dphi[j]*bN[j];
      }
      phiNb *= stepsize;
      if(fabs(phiNb)>fabs(max_delta_phi)) { /**/ // asdf
        max_delta_phi = phiNb;
        max_delta_phi_i = i;
      }
      phiNb += data.phi1;
      thetaNb *= stepsize; thetaNb += data.theta1;
      // In use, the coords are set so that phi1 is 0.
      // However, we include += data.phi1 to handle the
      // development test case using fixed coords, in
      // which phi1 varies.

      REAL8m sinthetaNb = sin(thetaNb);
      REAL8m costhetaNb = cos(thetaNb);
      REAL8m sinphiNb = sin(phiNb);
      REAL8m cosphiNb = cos(phiNb);
      ThreeVector rmNb(sinthetaNb*cosphiNb,
                       sinthetaNb*sinphiNb,
                       costhetaNb);  // mNb in rotated coords
      nstate->spin[i].Set
        (data.ex.x*rmNb.x+data.ey.x*rmNb.y+data.ez.x*rmNb.z,
         data.ex.y*rmNb.x+data.ey.y*rmNb.y+data.ez.y*rmNb.z,
         data.ex.z*rmNb.x+data.ey.z*rmNb.y+data.ez.z*rmNb.z);
    }
  }

#if 0
  if(N==6) { /**/ // asdf
    // Actual step taken
    fprintf(stdout,"It %5u: Max delta phi = %#.5g deg at index %u\n",
            cstate->iteration_count,
            max_delta_phi*180./PI,
            max_delta_phi_i);
    fflush(stdout);
  }
#endif

  if(N<7) {
    UpdateTimeFields(*cstate,*nstate,aN*stepsize);
    // At this point nstate->spin holds mNb = m{N+1}.
  }


  next_state_key.GetReadReference(); // Release write lock
}

void Oxs_SemiAnalyticEvolve::RungeKuttaFehlbergBase54
(RKF_SubType method,
 REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& current_mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{ // Runge-Kutta-Fehlberg routine with combined 4th and 5th
  // order Runge-Kutta steps.  The difference between the
  // two results (4th vs. 5th) is used to estimate the error.
  // The largest detected error detected cellsize is returned
  // in export error_estimate.  The units on this are radians.
  // The export global_error_order is set to 4 by this routine.
  // (The local error order is one better, i.e., 5.)  The norm_error
  // export is set to the cellwise maximum deviation from unit norm
  // across all the spins in the final state, before renormalization.

  // The following coefficients appear in
  //
  //   J. R. Dormand and P. J. Prince, ``A family of embedded
  //   Runge-Kutta formulae,'' J. Comp. Appl. Math., 6, 19--26
  //   (1980).
  //
  // They are also listed in J. Stoer and R. Bulirsch's book,
  // ``Introduction to Numerical Analysis,'' Springer, 2nd edition,
  // Sec. 7.2.5, p 454, but there are a number of errors in the S&B
  // account; the reader is recommended to refer directly to the D&P
  // paper.  A follow-up paper,
  //
  //   J. R. Dormand and P. J. Prince, ``A reconsideration of some 
  //   embedded Runge-Kutta formulae,'' J. Comp. Appl. Math., 15,
  //   203--211 (1986)
  //
  // provides formulae with improved stability and higher order.
  // See also
  //
  //   J. H. Williamson, ``Low-Storage Runge-Kutta Schemes,''
  //   J. Comp. Phys., 35, 48--56 (1980).
  //
  // FORMULAE for RK5(4)7FM:
  //
  //     dm_dt1 = dm_dt(t1,m1)
  //     dm_dt2 = dm_dt(t1+ (1/5)*h, m1+h*k1);
  //     dm_dt3 = dm_dt(t1+(3/10)*h, m1+h*k2);
  //     dm_dt4 = dm_dt(t1+ (4/5)*h, m1+h*k3);
  //     dm_dt5 = dm_dt(t1+ (8/9)*h, m1+h*k4);
  //     dm_dt6 = dm_dt(t1+     1*h, m1+h*k5);
  //     dm_dt7 = dm_dt(t1+     1*h, m1+h*k6);
  //  where
  //     k1 = dm_dt1*1/5
  //     k2 = dm_dt1*3/40       + dm_dt2*9/40
  //     k3 = dm_dt1*44/45      - dm_dt2*56/15      + dm_dt3*32/9
  //     k4 = dm_dt1*19372/6561 - dm_dt2*25360/2187 + dm_dt3*64448/6561
  //               - dm_dt4*212/729
  //     k5 = dm_dt1*9017/3168  - dm_dt2*355/33     + dm_dt3*46732/5247
  //               + dm_dt4*49/176   - dm_dt5*5103/18656
  //     k6 = dm_dt1*35/384     + 0                 + dm_dt3*500/1113
  //               + dm_dt4*125/192  - dm_dt5*2187/6784  + dm_dt6*11/84
  // Then
  //     Da = dm_dt1*35/384     + 0 + dm_dt3*500/1113   + dm_dt4*125/192
  //              - dm_dt5*2187/6784      + dm_dt6*11/84
  //     Db = dm_dt1*5179/57600 + 0 + dm_dt3*7571/16695 + dm_dt4*393/640
  //              - dm_dt5*92097/339200   + dm_dt6*187/2100   + dm_dt7*1/40
  // and
  //     m2a = m1 + h*Da
  //     m2b = m1 + h*Db.
  //
  // where m2a is the 5th order estimate, which is the candidate
  // for the next state, and m2b is the 4th order estimate used
  // only for error estimation/stepsize control.  Note that the
  // 4th order estimate uses more dm/dt evaluations (7) than the
  // 5th order method (6).  This is intentional; the coefficients
  // are selected to minimize error (see D&P paper cited above).
  // The extra calculation costs are minimal, because the 7th dm_dt
  // evaluation is at the candidate next state, so it is re-used
  // on the next step unless the step rejected.
  //
  // The error estimate is
  // 
  //     E = |m2b-m2a| = h*|Db-Da| = C*h^5
  //
  // where C is a constant that can be estimated in terms of E.
  // Note that we don't need to know C in order to adjust the
  // stepsize, because stepsize adjustment involves only the
  // ratio of the old stepsize to the new, so C drops out.

  // Scratch space usage:
  //  The import next_state is used for intermediate state
  // storage for all dm_dt computations.  The final computation
  // is for dm_dt7 at m1+h*k6 = m1+h*Da, which is the candidate
  // next state.  (Da=k6; see FSAL note below.)

  // Four temporary arrays, A-D, are used:
  //
  // Step \  Temp
  // Index \ Array:  A         B         C         D
  // ------+---------------------------------------------
  //   1   |      dm_dt2       -         -         -
  //   2   |      dm_dt2      k2         -         -
  //   3   |      dm_dt2     dm_dt3      -         -
  //   4   |      dm_dt2     dm_dt3     k3         -
  //   5   |      dm_dt2     dm_dt3    dm_dt4      -
  //   6   |      dm_dt2     dm_dt3    dm_dt4     k4
  //   7   |      dm_dt2     dm_dt3    dm_dt4    dm_dt5
  //   8   |        k5       dm_dt3    dm_dt4    dm_dt5
  //   9   |      dm_dt6     dm_dt3    dm_dt4    dm_dt5
  //  10   |      k6(3,6)    dD(3,6)   dm_dt4    dm_dt5
  //  11   |      dm_dt7     dD(3,6)   dm_dt4    dm_dt5
  //  12   |      dm_dt7       dD      dm_dt4    dm_dt5
  //
  // Here dD is Db-Da.  We don't need to compute Db directly.
  // The parenthesized numbers, e.g., k6(3,6) indicates
  // a partially formed value.  The total value k6 depends
  // upon dm_dt1, dm_dt3, dm_dt4, dm_dt5, and dm_dt6.  But
  // if we form k6 directly at step 11 in array A, then we
  // lose dm_dt6 which is needed to compute dD.  Rather than
  // use an additional array, we accumulate partial results
  // into arrays A and B for k6 and dD as indicated.
  //   Note that Da = k6.  Further, note that dm_dt7 is
  // dm_dt for the next state candidate.  (This is the 'F',
  // short for 'FSAL' ("first same as last"?) in the method
  // name, RK5(4)7FM.  The 'M' means minimized error norm,
  // 7 is the number of stages, and 5(4) is the main/subsidiary
  // integration formula order.  See the D&P 1986 paper for
  // details and additional references.)

  // Coefficient arrays, a, b, dc, defined by:
  //
  //   dm_dt(N+1) = dm_dt(t1+aN*h,m1+h*kN)
  //           kN = \sum_{M=1}^{M=N} dm_dtM*bNM
  //      Db - Da = \sum dm_dtM*dcM
  //
  REAL8m a1,a2,a3,a4;
  REAL8m a5 = 1.0; // True for all cases
  REAL8m a6 = 1.0; // True for all cases

  // Arrays for use by substep routine.  The reference indices use
  // fortran-style (1-based) indices, as opposed to the c-style
  // (0-base) indices used by the arrays.  Just deal.
  REAL8m b1[1],b2[2],b3[3],b4[4],b5[5],b6[6];
  REAL8m& b11 = b1[0];

  REAL8m& b21 = b2[0];  REAL8m& b22 = b2[1];

  REAL8m& b31 = b3[0];  REAL8m& b32 = b3[1];  REAL8m& b33 = b3[2];

  REAL8m& b41 = b4[0];  REAL8m& b42 = b4[1];  REAL8m& b43 = b4[2];
  REAL8m& b44 = b4[3];

  REAL8m& b51 = b5[0];  REAL8m& b52 = b5[1];  REAL8m& b53 = b5[2];
  REAL8m& b54 = b5[3];  REAL8m& b55 = b5[4];

  REAL8m& b61 = b6[0];  REAL8m& b62 = b6[1];  REAL8m& b63 = b6[2];
  REAL8m& b64 = b6[3];  REAL8m& b65 = b6[4];  REAL8m& b66 = b6[5];

  b62 = 0.0;  // True for all cases

  // Array used for error estimation.
  // c[k] = b6[k], c^[2]=c[2]=0.0 (1-based), where
  // c are the coeffs for Da, c^ for Db, and dcM = c^[M]-c[M].
  REAL8m dc[7];
  REAL8m& dc1 = dc[0];  REAL8m& dc2 = dc[1];  REAL8m& dc3 = dc[2];
  REAL8m& dc4 = dc[3];  REAL8m& dc5 = dc[4];  REAL8m& dc6 = dc[5];
  REAL8m& dc7 = dc[6];
  dc2 = 0.0;  // True for all cases

  switch(method) {
  case RK547FC:
    /////////////////////////////////////////////////////////////////
    // Coefficients for Dormand & Prince RK5(4)7FC
    a1 = 1./5.;
    a2 = 3./10.;
    a3 = 6./13.;
    a4 = 2./3.;
    // a5 and a6 are 1.0

    b11 =      1./5.;
  
    b21 =      3./40.;
    b22 =      9./40.;
  
    b31 =    264./2197.;
    b32 =    -90./2197.;
    b33 =    840./2197.;
  
    b41 =    932./3645.;
    b42 =    -14./27.;
    b43 =   3256./5103.;
    b44 =   7436./25515.;
  
    b51 =   -367./513.;
    b52 =     30./19.;
    b53 =   9940./5643.;
    b54 = -29575./8208.;
    b55 =   6615./3344.;
  
    b61 =     35./432.;
    b63 =   8500./14553.;
    b64 = -28561./84672.;
    b65 =    405./704.;
    b66 =     19./196.;
    // b62 is 0.0

    // Coefs for error calculation (c^[k] - c[k]).
    // Note that c[k] = b6k, and c^[2]=c[2]=0.0
    dc1 =     11./108.    - b61;
    dc3 =   6250./14553.  - b63;
    dc4 =  -2197./21168.  - b64;
    dc5 =     81./176.    - b65;
    dc6 =    171./1960.   - b66;
    dc7 =      1./40.;
    break;

  case RK547FM:
    /////////////////////////////////////////////////////////////////
    // Coefficients for Dormand & Prince RK5(4)7FM
    a1 = 1./5.;
    a2 = 3./10.;
    a3 = 4./5.;
    a4 = 8./9.;
    // a5 and a6 are 1.0

    b11 =      1./5.;
  
    b21 =      3./40.;
    b22 =      9./40.;
  
    b31 =     44./45.;
    b32 =    -56./15.;
    b33 =     32./9.;
  
    b41 =  19372./6561.;
    b42 = -25360./2187.;
    b43 =  64448./6561.;
    b44 =   -212./729.;
  
    b51 =   9017./3168.;
    b52 =   -355./33.;
    b53 =  46732./5247.;
    b54 =     49./176.;
    b55 =  -5103./18656.;
  
    b61 =     35./384.;
    b63 =    500./1113.;
    b64 =    125./192.;
    b65 =  -2187./6784.;
    b66 =     11./84.;
    // b62 is 0.0

    // Coefs for error calculation (c^[k] - c[k]).
    // Note that c[k] = b6k, and c^[2]=c[2]=0.0
    dc1 =   5179./57600.  - b61;
    dc3 =   7571./16695.  - b63;
    dc4 =    393./640.    - b64;
    dc5 = -92097./339200. - b65;
    dc6 =    187./2100.   - b66;
    dc7 =      1./40.;
    break;
  case RK547FS:
    /////////////////////////////////////////////////////////////////
    // Coefficients for Dormand & Prince RK5(4)7FS
    a1 = 2./9.;
    a2 = 1./3.;
    a3 = 5./9.;
    a4 = 2./3.;
    // a5 and a6 are 1.0

    b11 =      2./9.;
  
    b21 =      1./12.;
    b22 =      1./4.;
  
    b31 =     55./324.;
    b32 =    -25./108.;
    b33 =     50./81.;
  
    b41 =     83./330.;
    b42 =    -13./22.;
    b43 =     61./66.;
    b44 =      9./110.;
  
    b51 =    -19./28.;
    b52 =      9./4.;
    b53 =      1./7.;
    b54 =    -27./7.;
    b55 =     22./7.;
  
    b61 =     19./200.;
    b63 =      3./5.;
    b64 =   -243./400.;
    b65 =     33./40.;
    b66 =      7./80.;
    // b62 is 0.0

    // Coefs for error calculation (c^[k] - c[k]).
    // Note that c[k] = b6k, and c^[2]=c[2]=0.0
    dc1 =    431./5000.  - b61;
    dc3 =    333./500.   - b63;
    dc4 =  -7857./10000. - b64;
    dc5 =    957./1000.  - b65;
    dc6 =    193./2000.  - b66;
    dc7 =     -1./50.;
    break;
  default:
    throw Oxs_Ext::Error(this,
                 "Oxs_Oxs_SemiAnalyticEvolve::RungeKuttaFehlbergBase54:"
                 " Programming error; Invalid sub-type.");
  }

#ifdef CODE_CHECKS
  // COEFFICIENT CHECKS ////////////////////////////////////////
  // Try to catch some simple typing errors
#define EPS (6*REAL8_EPSILON)
  if(fabs(b11-a1)>EPS ||
     fabs(b21+b22-a2)>EPS ||
     fabs(b31+b32+b33-a3)>EPS ||
     fabs(b41+b42+b43+b44-a4)>EPS ||
     fabs(b51+b52+b53+b54+b55-a5)>EPS ||
     fabs(b61+b63+b64+b65+b66-a6)>EPS) {
    char buf[512];
    Oc_Snprintf(buf,sizeof(buf),
                "Coefficient check failed:\n"
                "1: %g\n2: %g\n3: %g\n4: %g\n5: %g\n6: %g",
                static_cast<double>(b11-a1),
                static_cast<double>(b21+b22-a2),
                static_cast<double>(b31+b32+b33-a3),
                static_cast<double>(b41+b42+b43+b44-a4),
                static_cast<double>(b51+b52+b53+b54+b55-a5),
                static_cast<double>(b61+b63+b64+b65+b66-a6));
    throw Oxs_Ext::Error(this,buf);
  }
#endif // CODE_CHECKS

  const Oxs_SimState* cstate = &(current_state_key.GetReadReference());

  const UINT4m size = cstate->mesh->Size();
  rkf54_data.AdjustSize(cstate->mesh);

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate->mesh->Id()
     || !gamma.CheckMesh(cstate->mesh)
     || !alpha.CheckMesh(cstate->mesh)) {
    UpdateMeshArrays(cstate->mesh);
  }

  UINT4m i;

  // Step 1: Compute rotated coordinate frame and theta1.
  REAL8m lambda = 0.0;
  if(old_timestep_cached != 0.0) {
    lambda = -0.5*stepsize/old_timestep_cached;
  }

#if 0
  const Oxs_MeshValue<REAL8m>& Ms = *(cstate->Ms);
#endif
  for(i=0;i<size;i++) {
    const ThreeVector& m1 = cstate->spin[i];
    const ThreeVector& m1xH = current_mxH[i];
    RKF54_dataobj& data = rkf54_data[i];
#if 0
    if(Ms[i] ==0.0) {
      // Coords don't matter
      data.ex.Set(1,0,0);
      data.ey.Set(0,1,0);
      data.ez.Set(0,0,1);
      data.theta1 = 0.0;
      continue;
    }
#endif
//#define FIXED_COORDS
#ifdef FIXED_COORDS
#if 1
    data.ex.Set(1,0,0);
    data.ey.Set(0,-1,0);
    data.ez.Set(0,0,-1);
#else
    data.ex.Set(0,0,1);
    data.ey.Set(0,1,0);
    data.ez.Set(-1,0,0);
#endif
    REAL8m xdot = data.ex * m1;
    REAL8m ydot = data.ey * m1;
    REAL8m zdot = data.ez * m1;
    if(zdot>=1.0)       data.theta1 = 0.0;
    else if(zdot<=-1.0) data.theta1 = PI;
    else                data.theta1 = acos(zdot);
    data.phi1 = atan2(ydot,xdot);
#else // FIXED_COORDS
    BOOL use_mxHxm_for_axis = 1;
    if(mxH_old_cached_is_valid) {
      // Use data from last valid step and current
      // point to compute ez
      ThreeVector acH,acT0,acT1;
      ComputeAxisCorrectedH(stepsize,lambda,gamma[i],2.0,1e-4,
                            m_old_cached[i],mxH_old_cached[i],
                            m1,m1xH,
                            acH,acT0,acT1);
#if 0 // asdf
      if(470<cstate->iteration_count && cstate->iteration_count<473
         && (i == 1183)) {
        ThreeVector freebish = m1xH;  freebish ^= m1;
        fprintf(stdout,"Iteration %u, i=%u, stepsize=%#.5g ---\n",
                cstate->iteration_count,i,stepsize);
        fprintf(stdout,"    m1=(%8.5f,%8.5f,%8.5f)\n",m1.x,m1.y,m1.z);
        fprintf(stdout,"    T1=(%#8.5g,%#8.5g,%#8.5g)\n",m1xH.x,m1xH.y,m1xH.z);
        fprintf(stdout,"  acT1=(%#8.5g,%#8.5g,%#8.5g)\n",acT1.x,acT1.y,acT1.z);
        fprintf(stdout," T1xm1=(%#8.5g,%#8.5g,%#8.5g)\n",
                freebish.x,freebish.y,freebish.z);
        fprintf(stdout,"   acH=(%#8.5g,%#8.5g,%#8.5g)\n",acH.x,acH.y,acH.z);
        REAL8m angA = acos(acH*freebish/sqrt(acH.MagSq()*freebish.MagSq()));
        REAL8m angB = acos(acH*m1/sqrt(acH.MagSq()*m1.MagSq()));
        angA *= 180./PI;
        angB *= 180./PI;
        fprintf(stdout,"   ang acH , T1xm1 = %#.5g degrees\n",angA);
        fprintf(stdout,"   ang acH ,    m1 = %#.5g degrees\n",angB);
      }
#endif // asdf

      if(acH.MakeUnit()>0.0) {
#if 1
        data.ez = acH;
        REAL8m dot = data.ez * m1;
        data.ex = m1;
        data.ex.Accum(-1*dot,data.ez);
        if(data.ex.MakeUnit()>0.0) {
          data.ey = data.ez;
          data.ey ^= data.ex;
          data.ey.MakeUnit();
# if 1
          if(fabs(dot)<SQRT1_2) {
            data.theta1 = acos(dot);
          } else {
            ThreeVector cross = m1;  cross ^= data.ez;
            REAL8m sintheta = sqrt(cross.MagSq()/m1.MagSq());
            data.theta1 = asin(sintheta);
            if(dot<0) data.theta1 = PI - data.theta1;
          }
          data.phi1 = 0.0;

#  if 0
          // Why does the following blow chunks?
          data.ex *= -1;
          data.ey *= -1;
          data.phi1 -= PI;
#  endif

# else
          REAL8m xdot = data.ex * m1;
          REAL8m ydot = data.ey * m1;
          REAL8m zdot = data.ez * m1;
          if(zdot>=1.0)       data.theta1 = 0.0;
          else if(zdot<=-1.0) data.theta1 = PI;
          else                data.theta1 = acos(zdot);
          data.phi1 = atan2(ydot,xdot);
# endif
          use_mxHxm_for_axis = 0;
        }
#else
        data.ez = acH;
        data.ey = acH;
        data.ey ^= m1;
        while(data.ey.MakeUnit()==0.0) {
          data.ey ^= data.ez;
        }
        data.ex = data.ey;
        data.ex ^= data.ez;
        data.ex.MakeUnit(); // Safety;

        REAL8m xdot = data.ex * m1;
        REAL8m ydot = data.ey * m1;
        REAL8m zdot = data.ez * m1;
        if(zdot>=1.0)       data.theta1 = 0.0;
        else if(zdot<=-1.0) data.theta1 = PI;
        else                data.theta1 = acos(zdot);
        if(zdot>SQRT1_2) {
          ThreeVector foo = data.ez;
          foo ^= m1;
          data.theta1 = asin(sqrt(foo.MagSq()));
        } else if(zdot<=-1.0) {
          data.theta1 = PI;
        } else {
          data.theta1 = acos(zdot);
        }
        data.phi1 = atan2(ydot,xdot);

        use_mxHxm_for_axis = 0;
#endif
      }
    }
    if(use_mxHxm_for_axis) {
      // Ignore old m and mxH data (if any), and set
      // axis to mxHxm.
      ThreeVector ez = m1xH;
      ez ^= m1;
      if(ez.MakeUnit()<=0.0) {
        // If |ez|=0, then take ez parallel to m1, ex and ey don't matter
        data.ez = m1;
        data.ez.MakeUnit(); // Safety
        data.theta1 = 0.0;
        if(fabs(data.ez.x)<SQRT1_2) {
          data.ey.Set(-1,0,0);
        } else if(fabs(data.ez.y)<SQRT1_2) {
          data.ey.Set(0,-1,0);
        } else {
          data.ey.Set(0,0,-1);
        }
        data.ey ^= ez;  data.ey.MakeUnit();
        data.ex = data.ey;
        data.ex ^= ez;
        ez ^= m1;
      } else {
        data.ez = ez;
        data.ex = m1;
        data.ex.MakeUnit();  // Safety
        data.theta1 = 0.5*PI;
        data.ey = ez;
        data.ey ^= data.ex;
        data.ey.MakeUnit();  // Safety
      }
    }
#endif // FIXED_COORDS
  }

  // Step 2: Fill rkf54_data with dm_dtN data for N=1..7
  REAL8m pE_pt;
#if 1
  next_state_key.GetWriteReference().spin = cstate->spin; // Copy spin data
  next_state_key.GetReadReference(); // Release write lock

  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   &current_mxH,NULL,1,a1,b1,pE_pt);
  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   NULL,NULL,2,a2,b2,pE_pt);
  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   NULL,NULL,3,a3,b3,pE_pt);
  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   NULL,NULL,4,a4,b4,pE_pt);
  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   NULL,NULL,5,a5,b5,pE_pt);
  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   NULL,NULL,6,a6,b6,pE_pt);
  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   NULL,&vtmpB,7,0.0,NULL,pE_pt);
#else
  a1 = 1.0;  b1[0] = 1.0;
  a2 = 1.0;  b2[0] = b2[1] = 0.5;

  next_state_key.GetWriteReference().spin = cstate->spin; // Copy spin data
  next_state_key.GetReadReference(); // Release write lock

  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   &current_mxH,NULL,1,a1,b1,pE_pt);
  RungeKuttaFehlbergBase54_Substep(stepsize,cstate,next_state_key,
                                   NULL,NULL,2,a2,b2,pE_pt);
  GetEnergyDensity(next_state_key.GetReadReference(),temp_energy,
                   &vtmpA,&vtmpB,pE_pt);
  ZeroFixedSpinFields(&vtmpA,&vtmpB);
#endif
  const Oxs_SimState* endstate = &(next_state_key.GetReadReference());
  // At this point, next_state holds the final candidate,
  // vtmpA holds mxH at that candidate, vtmpB holds H,
  // and pE_pt is filled.

  // Step 3: Compute error estimate, in spherical coords.
  //   Instead, we could compute in Cartesian coords,
  //   but in that case we lose multiple wraps about
  //   the pole.  Another option would be to weight the
  //   phi error by sin(theta), so that small errors
  //   near the rotation poles are given the corresponding
  //   Cartesian weight.
  REAL8m max_error_sq = 0.0;
  for(i=0;i<size;i++) {
    RKF54_dataobj& data = rkf54_data[i];
    REAL8m theta_err=0.0,phi_err=0.0;
    for(int j=0;j<7;++j) {
      theta_err += data.dtheta[j] * dc[j];
      phi_err   += data.dphi[j]   * dc[j];
    }
    phi_err *= sin(data.theta1); // Weight phi error???
    REAL8m error_sq = theta_err*theta_err + phi_err*phi_err;
    if(error_sq>max_error_sq) max_error_sq = error_sq;
  }
  error_estimate = stepsize * sqrt(max_error_sq);
  global_error_order = 5.0;

  if(!endstate->AddDerivedData("pE/pt",pE_pt)) {
    throw Oxs_Ext::Error(this,
       "Oxs_SemiAnalyticEvolve::RungeKuttaFehlbergBase54:"
       " Programming error; data cache (pE/pt) already set.");
  }

  new_energy_and_field_computed = 1;

  // Note that end energy is in temp_energy, mxH is in vtmpA, and H is
  // in vtmpB as per contract.
}

void Oxs_SemiAnalyticEvolve::TakeRungeKuttaFehlbergStep54M
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& /* H */,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{
  RungeKuttaFehlbergBase54(RK547FC,stepsize,
     current_state_key,mxH,next_state_key,
     error_estimate,global_error_order,
     new_energy_and_field_computed);
}

void Oxs_SemiAnalyticEvolve::TakeRungeKuttaFehlbergStep54S
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& /* H */,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{
  RungeKuttaFehlbergBase54(RK547FM,stepsize,
     current_state_key,mxH,next_state_key,
     error_estimate,global_error_order,
     new_energy_and_field_computed);
}

void Oxs_SemiAnalyticEvolve::TakeRungeKuttaFehlbergStep54
(REAL8m stepsize,
 Oxs_ConstKey<Oxs_SimState> current_state_key,
 const Oxs_MeshValue<ThreeVector>& /* H */,
 const Oxs_MeshValue<ThreeVector>& mxH,
 Oxs_Key<Oxs_SimState>& next_state_key,
 REAL8m& error_estimate,
 REAL8m& global_error_order,
 BOOL& new_energy_and_field_computed)
{
  RungeKuttaFehlbergBase54(RK547FS,stepsize,
     current_state_key,mxH,next_state_key,
     error_estimate,global_error_order,
     new_energy_and_field_computed);
}

REAL8m Oxs_SemiAnalyticEvolve::MaxDiff
(const Oxs_MeshValue<ThreeVector>& vecA,
 const Oxs_MeshValue<ThreeVector>& vecB)
{
  UINT4m size = vecA.Size();
  if(vecB.Size()!=size) {
    throw Oxs_Ext::Error(this,
                 "Oxs_SemiAnalyticEvolve::MaxDiff:"
                 " Import MeshValues incompatible (different lengths).");
  }
  REAL8m max_magsq = 0.0;
  for(UINT4m i=0;i<size;i++) {
    ThreeVector vtemp = vecB[i] - vecA[i];
    REAL8m magsq = vtemp.MagSq();
    if(magsq>max_magsq) max_magsq = magsq;
  }
  return sqrt(max_magsq);
}

void Oxs_SemiAnalyticEvolve::AdjustStepHeadroom(INT4m step_reject)
{ // step_reject should be 0 or 1, reflecting whether the current
  // step was rejected or not.  This routine updates reject_ratio
  // and adjusts step_headroom appropriately.

  // First adjust reject_ratio, weighing mostly the last
  // thirty or so results.
  reject_ratio = (31*reject_ratio + step_reject)/32.;

  // Adjust step_headroom
  if(reject_ratio>reject_goal && step_reject>0) {
    // Reject ratio too high and getting worse
    step_headroom *= 0.925;
  }
  if(reject_ratio<reject_goal && step_reject<1) {
    // Reject ratio too small and getting smaller
    step_headroom *= 1.075;
  }

  if(step_headroom>max_step_headroom) step_headroom=max_step_headroom;
  if(step_headroom<min_step_headroom) step_headroom=min_step_headroom;
}

BOOL
Oxs_SemiAnalyticEvolve::Step(const Oxs_TimeDriver* driver,
                      Oxs_ConstKey<Oxs_SimState> current_state_key,
                      const Oxs_DriverStepInfo& step_info,
                      Oxs_Key<Oxs_SimState>& next_state_key)
{
  const REAL8m bad_energy_cut_ratio = 0.75;
  const REAL8m bad_energy_step_increase = 1.3;

  const REAL8m previous_next_timestep = next_timestep;

  const Oxs_SimState& cstate = current_state_key.GetReadReference();

  // Fill out alpha and gamma meshvalue arrays, as necessary.
  if(mesh_id != cstate.mesh->Id() || !gamma.CheckMesh(cstate.mesh)
     || !alpha.CheckMesh(cstate.mesh)) {
    UpdateMeshArrays(cstate.mesh);
  }

  CheckCache(cstate);

  UpdateFixedSpinList(cstate.mesh);

  // Note if start_dm is being used
  BOOL start_dm_active=0;
  if(next_timestep<=0.0 ||
     (cstate.stage_iteration_count<1
      && step_info.current_attempt_count==0)) {
    if(cstate.stage_number==0
       || stage_init_step_control == SISC_START_DM) {
      start_dm_active = 1;
    } else if(stage_init_step_control == SISC_CONTINUOUS) {
      start_dm_active = 0;  // Safety
    } else if(stage_init_step_control == SISC_AUTO) {
      // Automatic detection based on energy values across
      // stage boundary.  For now just assume true:
      start_dm_active = 0;
    } else {
      throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::Step; Programming error:"
           " unrecognized stage_init_step_control value");
    }
  }

  // Disable use of mxH_old_cached in start_dm is being used
  if(start_dm_active) mxH_old_cached_is_valid = 0;

  // Negotiate timestep, and also initialize both next_state and
  // temp_state structures.
  Oxs_SimState* work_state = &(next_state_key.GetWriteReference());
  BOOL force_step=0,driver_set_step=0;
  NegotiateTimeStep(driver,cstate,*work_state,next_timestep,
                    start_dm_active,force_step,driver_set_step);
  REAL8m stepsize = work_state->last_timestep;
  work_state=NULL; // Safety: disable pointer

  // Step
  REAL8m error_estimate;
  REAL8m global_error_order;
  BOOL new_energy_and_field_computed;
  BOOL reject_step=0;
  (this->*sastep_ptr)(stepsize,current_state_key,
                      H_cached,
                      mxH_output.cache.value,
                      next_state_key,
                      error_estimate,global_error_order,
                      new_energy_and_field_computed);
  const Oxs_SimState& nstate = next_state_key.GetReadReference();

  REAL8m max_dm_dt;
  cstate.GetDerivedData("Max dm/dt",max_dm_dt);
  REAL8m reference_stepsize = stepsize;
  if(driver_set_step) reference_stepsize = previous_next_timestep;
  BOOL good_step = CheckError(global_error_order,error_estimate,
                              stepsize,reference_stepsize,
                              max_dm_dt,next_timestep);
  /// Note: Might want to use average or larger of max_dm_dt
  /// and new_max_dm_dt (computed below.)

  mxH_old_cached_is_valid = 0;  // Default; only change this if step
  /// is found to be good and we copy new state into cache.

  if(!good_step && !force_step) {
    // Bad step; The only justfication to do energy and dm_dt
    // computation would be to get an energy-based stepsize
    // adjustment estimate, which we only need to try if 
    // next_timestep is larger than cut applied by energy
    // rejection code (i.e., bad_energy_cut_ratio).
    if(next_timestep<=stepsize*bad_energy_cut_ratio) {
      AdjustStepHeadroom(1);
      return 0; // Don't bother with energy calculation
    }
    reject_step=1; // Otherwise, mark step rejected and see what energy
    /// info suggests for next stepsize
  }

  if(start_dm_active && !force_step) {
    // Check that no spin has moved by more than start_dm
    REAL8m diff = MaxDiff(cstate.spin,nstate.spin);
    if(diff>start_dm) {
      next_timestep = step_headroom * stepsize * (start_dm/diff);
      if(next_timestep<=stepsize*bad_energy_cut_ratio) {
        AdjustStepHeadroom(1);
        return 0; // Don't bother with energy calculation
      }
      reject_step=1; // Otherwise, mark step rejected and see what energy
      /// info suggests for next stepsize
    }
  }

  // Energy timestep control:
  //   The relationship between energy error and stepsize appears to be
  // highly nonlinear, so that estimating appropriate stepsize from energy
  // increase is difficult.  Perhaps it is possible to include energy
  // interpolation into RK step routines, but for the present we just
  // reduce the step by a fixed ratio if we detect energy increase beyond
  // that which can be attributed to numerical errors.  Of course, this
  // doesn't take into account the expected energy decrease (which depends
  // on the damping ratio alpha), which is another reason to try to build
  // it into the high order RK step routines.
  REAL8m pE_pt,new_pE_pt=0.;
  cstate.GetDerivedData("pE/pt",pE_pt);
  if(new_energy_and_field_computed) {
    nstate.GetDerivedData("pE/pt",new_pE_pt);
  } else {
    GetEnergyDensity(nstate,temp_energy,
                     &vtmpA,&vtmpB,new_pE_pt);
    ZeroFixedSpinFields(&vtmpA,&vtmpB);
    if(!nstate.AddDerivedData("pE/pt",new_pE_pt)) {
      throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::Step:"
           " Programming error; data cache (pE/pt) already set.");
    }
  }

  REAL8m dE=0.0;
  REAL8m var_dE=0.0;
  REAL8m total_E=0.0;
  UINT4m i;
  const UINT4m size = nstate.mesh->Size();
  for(i=0;i<size;++i) {
    REAL8m vol = nstate.mesh->Volume(i);
    REAL8m e = energy[i];
    total_E += e * vol;
    REAL8m new_e = temp_energy[i];
    dE += (new_e - e) * vol;
    var_dE += (new_e*new_e + e*e)*vol*vol;
  }
  if(!nstate.AddDerivedData("Delta E",dE)) {
    throw Oxs_Ext::Error(this,
         "Oxs_SemiAnalyticEvolve::Step:"
         " Programming error; data cache (Delta E) already set.");
  }

  if(expected_energy_precision>=0.) {
    var_dE *= expected_energy_precision * expected_energy_precision;
    /// Variance, assuming error in each energy[i] term is independent,
    /// uniformly distributed, 0-mean, with range
    ///        +/- expected_energy_precision*energy[i].
    /// It would probably be better to get an error estimate directly
    /// from each energy term.
    REAL8m E_numerror = OC_MAX(fabs(total_E)*expected_energy_precision,
                               2*sqrt(var_dE));
    REAL8m pE_pt_max = OC_MAX(pE_pt,new_pE_pt); // Might want to
    /// change this from a constant to a linear function in timestep.

    REAL8m reject_dE = 2*E_numerror + pE_pt_max * stepsize;
    if(dE>reject_dE) {
      REAL8m teststep = bad_energy_cut_ratio*stepsize;
      if(teststep<next_timestep) {
        next_timestep=teststep;
        max_step_increase = bad_energy_step_increase;
        // Damp the enthusiasm of the RK stepper routine for
        // stepsize growth, for a step or two.
      }
      if(!force_step) reject_step=1;
    }
  }


  if(!force_step && reject_step) {
    AdjustStepHeadroom(1);
    return 0;
  }

  // Otherwise, we are accepting the new step.

  // Fill in cache.
  energy.Swap(temp_energy);
  mxH_old_cached.Swap(mxH_output.cache.value);
  m_old_cached.AdjustSize(mxH_old_cached);
  mxHxm_old_cached.AdjustSize(mxH_old_cached);
  for(i=0;i<size;++i) {
    m_old_cached[i] = cstate.spin[i];
    mxHxm_old_cached[i] = mxH_old_cached[i];
    mxHxm_old_cached[i] ^= m_old_cached[i];
  }
  old_timestep_cached = stepsize;
  mxH_old_cached_is_valid = 1;
  mxH_output.cache.value.Swap(vtmpA);
  H_cached.Swap(vtmpB);
  mxH_output.cache.state_id = energy_state_id = nstate.Id();
  
  REAL8m new_max_dm_dt,new_dE_dt,new_timestep_lower_bound;
  ComputeMiscValues(nstate,mxH_output.cache.value,new_pE_pt,
                    new_max_dm_dt,new_dE_dt,new_timestep_lower_bound);

  if(!nstate.AddDerivedData("Timestep lower bound",
                            new_timestep_lower_bound) ||
     !nstate.AddDerivedData("Max dm/dt",new_max_dm_dt) ||
     !nstate.AddDerivedData("dE/dt",new_dE_dt)) {
    throw Oxs_Ext::Error(this,
                         "Oxs_SemiAnalyticEvolve::Step:"
                         " Programming error; data cache already set.");
  }

  AdjustStepHeadroom(0);
  if(!force_step && max_step_increase<max_step_increase_limit) {
    max_step_increase *= max_step_increase_adj_ratio;
  }
  if(max_step_increase>max_step_increase_limit) {
    max_step_increase = max_step_increase_limit;
  }
  return 1; // Accept step
}

void Oxs_SemiAnalyticEvolve::UpdateDerivedOutputs(const Oxs_SimState& state)
{ // This routine fills all the Oxs_SemiAnalyticEvolve Oxs_ScalarOutput's to
  // the appropriate value based on the import "state", and any of
  // Oxs_VectorOutput's that have CacheRequest enabled are filled.
  // It also makes sure all the expected WOO objects in state are
  // filled.
  max_dm_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = delta_E_output.cache.state_id
    = 0;  // Mark change in progress

  REAL8m dummy_value;
  if(!state.GetDerivedData("Max dm/dt",max_dm_dt_output.cache.value) ||
     !state.GetDerivedData("dE/dt",dE_dt_output.cache.value) ||
     !state.GetDerivedData("Delta E",delta_E_output.cache.value) ||
     !state.GetDerivedData("pE/pt",dummy_value) ||
     !state.GetDerivedData("Timestep lower bound",dummy_value) ||
     (mxH_output.GetCacheRequestCount()>0
      && mxH_output.cache.state_id != state.Id())) {

    // Missing at least some data, so calculate from scratch

    // Fill out alpha and gamma meshvalue arrays, as necessary.
    if(mesh_id != state.mesh->Id() || !gamma.CheckMesh(state.mesh)
       || !alpha.CheckMesh(state.mesh)) {
      UpdateMeshArrays(state.mesh);
    }

    // Calculate H and mxH outputs
    Oxs_MeshValue<ThreeVector>& mxH = mxH_output.cache.value;
    REAL8m pE_pt;
    GetEnergyDensity(state,energy,&mxH,&H_cached,pE_pt);
    ZeroFixedSpinFields(&mxH,&H_cached);
    energy_state_id=state.Id();
    mxH_output.cache.state_id=state.Id();
    if(!state.GetDerivedData("pE/pt",dummy_value)) {
      state.AddDerivedData("pE/pt",pE_pt);
    }

    // Calculate Max dm/dt and dE/dt
    REAL8m timestep_lower_bound;
    ComputeMiscValues(state,mxH,pE_pt,
                      max_dm_dt_output.cache.value,
                      dE_dt_output.cache.value,timestep_lower_bound);
    if(!state.GetDerivedData("Max dm/dt",dummy_value)) {
      state.AddDerivedData("Max dm/dt",max_dm_dt_output.cache.value);
    }
    if(!state.GetDerivedData("dE/dt",dummy_value)) {
      state.AddDerivedData("dE/dt",dE_dt_output.cache.value);
    }
    if(!state.GetDerivedData("Timestep lower bound",dummy_value)) {
      state.AddDerivedData("Timestep lower bound",
                           timestep_lower_bound);
    }

    if(!state.GetDerivedData("Delta E",dummy_value)) {
      if(state.previous_state_id!=0 && state.stage_iteration_count>0) {
        // Strictly speaking, we should be able to create dE for
        // stage_iteration_count==0 for stages>0, but as a practical
        // matter we can't at present.  Should give this more thought.
        // -mjd, 27-July-2001
        throw Oxs_Ext::Error(this,
           "Oxs_SemiAnalyticEvolve::UpdateDerivedOutputs:"
           " Can't derive Delta E from single state.");
      }
      state.AddDerivedData("Delta E",0.0);
      dummy_value = 0.;
    }
    delta_E_output.cache.value=dummy_value;
  }

  max_dm_dt_output.cache.value*=(180e-9/PI);
  /// Convert from radians/second to deg/ns

  max_dm_dt_output.cache.state_id
    = dE_dt_output.cache.state_id
    = delta_E_output.cache.state_id
    = state.Id();
}
