# darwin.tcl
#
# Configuration feature definitions for the configuration 'darwin'
#
# Editing instructions begin at "START EDIT HERE" below.

set config [Oc_Config RunPlatform]

if {![string match [string tolower [file rootname [file tail [info script]]]] \
        [$config GetValue platform_name]]} {
    error "Configuration cache file '[info script]' 
sourced by '[$config GetValue platform_name]'"
}

########################################################################
# START EDIT HERE
# OOMMF software cannot classify your computing platform type.  See
# the Installation section of the OOMMF User Manual for instructions on 
# how to add a new platform type to the collection of types recognized 
# by OOMMF software.
#
# Say you add the new platform type 'foo' to describe your computing
# platform.  Then copy this file to ./foo.tcl , and edit it to
# describe your computing platform.  In its initial state, this file
# describes a rather generic Linux system, so there shouldn't be much
# editing required for any Unix-based platform.  Other platforms may
# be more difficult or impossible.  In any event, please e-mail the
# OOMMF developers for assistance setting up OOMMF for your particular
# circumstances.
########################################################################
# In order to properly build, install, and run on your computing 
# platform, the OOMMF software must know certain features of your
# computing environment.  In this file are lines which set the value of 
# certain features of your computing environment.  Each line looks like:
#
# $config SetValue <feature> {<value>}
#
# where each <feature> is the name of some feature of interest,
# and <value> is the value which is assigned to that feature in a
# description of your computing environment.  Your task is to edit
# the values as necessary to properly describe your computing 
# environment.
#
# The character '#' at the beginning of a line is a comment character.
# It causes the contents of that line to be ignored.  To select
# among lines providing alternative values for a feature, uncomment the
# line containing the proper value.
#
# The features in this file are divided into three sections.  The first
# section (REQUIRED CONFIGURATION) includes features which require you 
# to provide a value.  The second section (OPTIONAL CONFIGURATION)
# includes features which have usable default values, but which you
# may wish to customize.  The third section (ADVANCED CONFIGURATION)
# contains features which you probably do not need or want to change
# without a good reason.
########################################################################
# REQUIRED CONFIGURATION

########################################################################
# OPTIONAL CONFIGURATION

# Set the feature 'program_compiler_c++' to the program to run on this 
# platform to compile source code files written in the language C++ into 
# object files.  Select from the choices below.  If the compiler is not 
# in your path, be sure to use the whole pathname.  Also include any 
# options required to instruct your compiler to only compile, not link.  
#
# If your compiler is not listed below, additional features will
# have to be added in the ADVANCED CONFIGURATION section below to 
# describe to the OOMMF software how to operate your compiler.  Send
# e-mail to the OOMMF developers for assistance.
#
# The GNU C++ compiler 'g++'
# <URL:http://www.gnu.org/software/gcc/gcc.html>
# <URL:http://egcs.cygnus.com/>
$config SetValue program_compiler_c++ {g++ -c}

# Set the feature 'path_directory_temporary' to the name of an existing 
# directory on your computer in which OOMMF software should write 
# temporary files.  All OOMMF users must have write access to this 
# directory.
#
# $config SetValue path_directory_temporary {/tmp}

########################################################################
# SUPPORT PROCEDURES
#
# Routines to guess the processor type
source [file join [file dirname [Oc_DirectPathname [info script]]]  \
         cpuguess-darwin.tcl]

# Routine to guess the gcc version.  Note: Apple's version of gcc
# mungs the version string (at least in Leopard).
proc GuessGccVersion { gcc } {
    set guess {}
    catch {
        set fptr [open "|$gcc --version" r]
        set verstr [read $fptr]
        close $fptr
        regexp -- {(^| )([0-9][0-9.]+)( |$)} $verstr dummy dumA guess dumB
    }
    return [split $guess "."]
}

########################################################################
# ADVANCED CONFIGURATION

# Compile in thread support?
set oommf_threads [$config GetValue oommf_threads]

# Compiler option processing...
if {[string match g++ [file tail [lindex \
        [$config GetValue program_compiler_c++] 0]]]} {
   # ...for GNU g++ C++ compiler
   set opts {}               ;# Use optimization defined below
   # set opts [list -O0]     ;# Override opts: No optimization
   # set opts [list -O%s]    ;# Override opts: Compiler standard optimization
   # Note: opts setting is complicated a bit by the Apple-specific "-fast"
   #  flag; this flag specifies a variety of general optimization flags, but
   #  is also arch specific on the PowerPC arch, but not on Intel targets.

    # Architecture specific compile?
   if {[catch {$config GetValue cpu_arch} cpu_arch]} {
      set cpu_arch generic
   }

   if {![string match {} $opts] || \
          [string match generic [string tolower $cpu_arch]]} {
       # Generic build
       set cpuopts {}
    } else {
       # Arch specific build.  If cpu_arch is "host", then try to
       # guess.  Otherwise, assume user knows what he is doing and has
       # inserted an appropriate cpu_arch string, i.e., one that
       # matches the format and known types as returned from GuessCPU.
       if {[string match host $cpu_arch]} {
          set cpu_arch [GuessCPU]
       }
       foreach {cpubrand cputype} [string tolower $cpu_arch] { break }
       set gccversion [GuessGccVersion \
                          [lindex [$config GetValue program_compiler_c++] 0]]
       if {[string match "ppc" $cpubrand]} {
          # PowerPC CPU
          if {$gccversion>=3.3} {
             # Processor specific optimization
             if {[string match 7?? $cputype]} {
                # G3; use default opts string
             } elseif {[string match 9?? $cputype]} {
                # G5
                if {[string match {} $opts]} {
                   set opts -fast
                   ## -fast is an Apple specific option
                }
             } elseif {[string match 74?? $cputype]} {
                # G4
                if {$cputype>=7450} {
                   set cpuopts -mcpu=7450
                } else {
                   set cpuopts -mcpu=7400
                }
                if {[string match {} $opts]} {
                   set opts -fast
                   ## -fast is an Apple specific option
                }
             }
          }
       } else {
          # Intel CPU.  Compare to corresponding block in lintel.tcl
          set verA [lindex $gccversion 0]
          set verB [lindex $gccversion 1]
          set cpuopts {}
          if {$verA==2 && $verB>=95} {
             switch -glob -- $cputype {
                k6*      { set cpuopts [list -march=k6] }
                pentium  { set cpuopts [list -march=pentium] }
                athlon*  -
                opteron  -
                prescott -
                nocona   -
                core2    -
                pentium* { set cpuopts [list -march=pentiumpro] }
             }
          } elseif {$verA==3 && $verB<1} {
             switch -glob -- $cputype {
                pentium  { set cpuopts [list -march=pentium] }
                prescott -
                nocona   -
                core2    -
                pentium* { set cpuopts [list -march=pentiumpro] }
                k6*      { set cpuopts [list -march=k6] }
                opteron  -
                athlon*  { set cpuopts [list -march=athlon] }
             }
          } elseif {$verA==3 || ($verA==4 && $verB<2)} {
             switch -exact -- $cputype {
                pentium    { set cpuopts [list -march=pentium] }
                pentiumpro { set cpuopts [list -march=pentiumpro] }
                pentium2   { set cpuopts [list -march=pentium2] }
                pentium3   { set cpuopts [list -march=pentium3] }
                pentium-m  { set cpuopts \
                                [list -march=pentium-m -msse2 -mfpmath=sse] }
                pentium4   { set cpuopts \
                                [list -march=pentium4 -msse2 -mfpmath=sse] }
                prescott   { set cpuopts \
                                [list -march=prescott -msse3 -mfpmath=sse] }
                nocona     -
                core2      { set cpuopts \
                                [list -march=nocona -msse3 -mfpmath=sse] }
                k6         { set cpuopts [list -march=k6] }
                k6-2       { set cpuopts [list -march=k6-2] }
                k6-3       { set cpuopts [list -march=k6-3] }
                athlon     { set cpuopts [list -march=athlon] }
                athlon-tbird { set cpuopts [list -march=athlon-tbird] }
                athlon-4   { set cpuopts [list -march=athlon-4] }
                athlon-xp  { set cpuopts [list -march=athlon-xp] }
                athlon-mp  { set cpuopts [list -march=athlon-mp] }
                athlon-fx  -
                athlon64   -
                opteron    -
                k8         { set cpuopts \
                                [list -march=k8 -msse2 -mfpmath=sse] }
             }
             # Note 1: Replace -march=k8 with -march=k6 if compiler
             #         doesn't know k8.
             # Note 2: One can try -msse3 for nocona entry, but it
             #         may not be supported by older compilers.
             # Note 3: core 2 is a rather different architecture than
             #         nocona.  Might want to experiment with other settings.
          } elseif {$verA>4 || ($verA==4 && $verB>=2)} {
             set cpuopts [list -march=native]
             # -march/-mtune=native setting introduced with gcc 4.2
             switch -exact -- $cputype {
                pentium-m  -
                pentium4   { lappend cpuopts -msse2 -mfpmath=sse }
                prescott   -
                nocona     -
                core2      { lappend cpuopts -msse3 -mfpmath=sse }
                athlon-fx  -
                athlon64   -
                opteron    -
                k8         { lappend cpuopts -msse2 -mfpmath=sse }
             }
          }
          if {[string match {} $opts] && \
                 ![string match {} $cpuopts] && $gccversion>=3.3} {
             set opts -fast
             ## -fast is an Apple specific option
          }
       }
    }

    if {[string match {} $opts]} {
       # Default opts
       set opts [list -O2 -ffast-math -fstrict-aliasing]
    }

    if {[llength $cpuopts]>0} {
       set opts [concat $opts $cpuopts]
    }

    # 32 or 64 bit?
    switch -exact $tcl_platform(wordSize) {
       4 { lappend opts -m32 }
       8 { lappend opts -m64 }
    }

    # Compile in thread support?
    if {$oommf_threads} {
        lappend opts -DTCL_THREADS -DOOMMF_THREADS=1
    } else {
        lappend opts -DOOMMF_THREADS=0
    }

    $config SetValue program_compiler_c++_option_opt "format \"$opts\""

    $config SetValue program_compiler_c++_option_out {format "-o \"%s\""}
    $config SetValue program_compiler_c++_option_src {format \"%s\"}
    $config SetValue program_compiler_c++_option_inc {format "\"-I%s\""}
    $config SetValue program_compiler_c++_option_def {format "\"-D%s\""}

    # Widest natively support floating point type
    $config SetValue program_compiler_c++_typedef_realwide "double"

    # Directories to exclude from explicit include search path, i.e.,
    # the -I list.  Some versions of gcc complain if "system" directories
    # appear in the -I list.
    $config SetValue \
        program_compiler_c++_system_include_path [list /usr/include]
}

# The program to run on this platform to link together object files and
# library files to create an executable binary.
# (Use the selected compiler to control the linking.)
#
# Linker and linker option processing...
if {[string match g++ [file tail [lindex \
        [$config GetValue program_compiler_c++] 0]]]} {
   # ...for GNU g++ as linker
   set linkcmdline [lindex \
                       [$config GetValue program_compiler_c++] 0]
   # 32 or 64 bit?
   switch -exact $tcl_platform(wordSize) {
      4 { lappend linkcmdline -m32 }
      8 { lappend linkcmdline -m64 }
   }
   $config SetValue program_linker $linkcmdline
   unset linkcmdline
   $config SetValue program_linker_option_obj {format \"%s\"}
   $config SetValue program_linker_option_out {format "-o \"%s\""}
   $config SetValue program_linker_option_lib {format \"%s\"}
   $config SetValue program_linker_uses_-L-l {1}
}

# The program to run on this platform to create a single library file out
# of many object files.
$config SetValue program_libmaker {ar cr}

if {[string match ar [file tail [lindex \
        [$config GetValue program_libmaker] 0]]]} {
    # Option processing for ar
    $config SetValue program_libmaker_option_obj {format \"%s\"}
    $config SetValue program_libmaker_option_out {format \"%s\"}
}

# The absolute, native filename of the null device
$config SetValue path_device_null {/dev/null}

# A partial Tcl command (or script) which when completed by lappending
# a file name stem and evaluated returns the corresponding file name for
# an executable on this platform
$config SetValue script_filename_executable {format %s}

# A partial Tcl command (or script) which when completed by lappending
# a file name stem and evaluated returns the corresponding file name for
# an object file on this platform
$config SetValue script_filename_object {format %s.o}

# A partial Tcl command (or script) which when completed by lappending
# a file name stem and evaluated returns the corresponding file name for
# a static library on this platform
$config SetValue script_filename_static_library {format lib%s.a}

########################################################################
unset config
